{*****************************************************************************
* UNIT DESCRIPTION:  Base class for UI input values used by charting functions
* SPEC REF: (if used)
* DESIGN REF: (if used)
*
* Copyright (c) 2009 Schweitzer Engineering Laboratories, Pullman, Washington
******************************************************************************}
unit ClassChartInput;

interface
uses
  Windows, Messages, SysUtils, StrUtils, Variants, Classes, DateUtils,
  Contnrs, TimeRoutines, TimeSeriesConstants;
type
  TIntegerArray  = array of integer;
  TBooleanArray = array of boolean;
  TTimeSeriesChartRenderType = (sctUnknown = -1, sctLine, sctArea);

  TTimeSeriesInputSettings = class(TObject)
  private
    FChartAxisDateTime      : boolean;
    FChartStyle             : TTimeSeriesChartRenderType;
    FChartTransparency      : integer;
    FChartQuery             : boolean;
    FChartLabels            : boolean;
    FChart3D                : boolean;
    FChartSecondaryUnits    : boolean;
    FChartScale             : TTimeSeriesScaleType;

    FElevation              : integer;
    FRotation               : integer;
    FZoom                   : integer;
    FAspectZoom             : integer;
    FHorizOffset            : integer;
    FVertOffset             : integer;
    FPercent3D              : integer;

    FActiveChannels : TBooleanArray;
    FActiveEvents   : TBooleanArray;
    FWindowX        : integer;
    FWindowY        : integer;
    FWindowHeight   : integer;
    FWindowWidth    : integer;
  protected
  public
    function GetActiveChannels(index: integer) : boolean;
    procedure SetActiveChannels(index: integer; state : boolean);
    procedure SetActiveChannelLength(size : integer);
    function GetActiveEvents(index: integer) : boolean;
    procedure SetActiveEvents(index: integer; state : boolean);
    procedure SetActiveEventLength(size : integer);
    constructor Create;
    procedure Initialize;

    property ActiveChannelArray : TBooleanArray read FActiveChannels;
    property ActiveEventArray : TBooleanArray read FActiveEvents;
    property ChartAxisDateTime : boolean read FChartAxisDateTime write FChartAxisDateTime;
    property ChartStyle         : TTimeSeriesChartRenderType read FChartStyle write FChartStyle;
    property ChartTransparency : integer read FChartTransparency write FChartTransparency;
    property ChartQuery        : boolean read FChartQuery write FChartQuery;
    property ChartLabels       : boolean read FChartLabels write FChartLabels;
    property ChartSecondaryUnits    : boolean read FChartSecondaryUnits write FChartSecondaryUnits;
    property ChartScale        : TTimeSeriesScaleType read FChartScale write FChartScale;
    property Chart3D           : boolean read FChart3D write FChart3D;
    property Elevation         : integer read FElevation write FElevation;
    property Rotation          : integer read FRotation write FRotation;
    property Zoom              : integer read FZoom write FZoom;
    property AspectZoom        : integer read FAspectZoom write FAspectZoom;
    property HorizOffset       : integer read FHorizOffset write FHorizOffset;
    property VertOffset        : integer read FVertOffset write FVertOffset;
    property Percent3D         : integer read FPercent3D write FPercent3D;
    property WindowX           : integer read FWindowX write FWindowX;
    property WindowY           : integer read FWindowY write FWindowY;
    property WindowHeight      : integer read FWindowHeight write FWindowHeight;
    property WindowWidth       : integer read FWindowWidth write FWindowWidth;
    property ActiveChannels[index: integer] : boolean read GetActiveChannels write SetActiveChannels;
    property ActiveEvents[index: integer] : boolean read GetActiveEvents write SetActiveEvents;
  end;

implementation
{!******************************************************************************
* MODULE NAME ...  Create
* SYSTEM NAME ...  Acselerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:  Create the TTimeSeriesInputSettings object.
*
* INPUT PARAMETERS:
*                    None:
*
* OUTPUT PARAMETERS:
*                    None:
*
* RETURN VALUE:
*                    None:
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.12.03  Kent Myers        Initial implementation
*******************************************************************************}
constructor TTimeSeriesInputSettings.Create;
begin
  inherited Create;
  Initialize;
end;
{!******************************************************************************
* MODULE NAME ...  Initialize
* SYSTEM NAME ...  Acselerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:  Initialize Input setttings.
*
* INPUT PARAMETERS:
*                    None:
*
* OUTPUT PARAMETERS:
*                    None:
*
* RETURN VALUE:
*                    None:
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.12.03  Kent Myers        Initial implementation
*******************************************************************************}
procedure TTimeSeriesInputSettings.Initialize;
begin
  // Initialize input settings
  FChartAxisDateTime := false;     // Date/time labels disabled
  FChartStyle := sctLine;           // Use line chart
  FChartLabels := true;            // Chart labels active by default
  FChartSecondaryUnits := true;    // Chart primary/secondary units (secondary = true)
  FChartScale := stUnknown;        // Chart scale
  FChartTransparency := 0;         // Percentage transparent (opaque by default)
  FChartQuery := false;            // Chart query function disabled
  FChart3D := false;               // Chart is 2D by default
  FElevation := 0;                 // Percent elevation (none)
  FRotation := 0;                  // Percent rotation (none)
  FZoom := 100;                    // Percent zoom (fully zoomed)
  FHorizOffset := 0;               // Pixel offset in horizontal direction
  FVertOffset := 0;                // Pixel offset in vertical direction
  FPercent3D := 100;               // Percentage 3D
  FWindowX := 150;
  FWindowY := 50;
  FWindowWidth := 750;             // default to initial window size
  FWindowHeight := 641;
end;

{!******************************************************************************
* MODULE NAME ...  GetActiveChannels
* SYSTEM NAME ...  Acselerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:  Getter for ActiveChannels dynamic array.
*
* INPUT PARAMETERS:
*                    index: integer  index into array
*
* OUTPUT PARAMETERS:
*                    None:
*
* RETURN VALUE:
*                    result:  boolean  state of array element
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.12.04  Kent Myers        Initial implementation
*******************************************************************************}
function TTimeSeriesInputSettings.GetActiveChannels(index : integer) : boolean;
begin
  if((index >= Low(FActiveChannels)) and
     (index <= High(FActiveChannels))) then
  begin
    result := FActiveChannels[index];
  end else
  begin
    result := false;
  end;
end;
{!******************************************************************************
* MODULE NAME ...  SetActiveChannels
* SYSTEM NAME ...  Acselerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:  Setter for ActiveChannels dynamic array.
*
* INPUT PARAMETERS:
*                    index: integer  index into array
*                    state : boolean  state of array element
*
* OUTPUT PARAMETERS:
*                    None:
*
* RETURN VALUE:
*                    None:
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.12.04  Kent Myers        Initial implementation
*******************************************************************************}
procedure TTimeSeriesInputSettings.SetActiveChannels(index : integer; state : boolean);
begin
  if((index >= Low(FActiveChannels)) and
     (index <= High(FActiveChannels))) then
  begin
    FActiveChannels[index] := state;
  end;
end;
{!******************************************************************************
* MODULE NAME ...  SetActiveChannelLength
* SYSTEM NAME ...  Acselerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:  Allocate ActiveChannels dynamic array.
*
* INPUT PARAMETERS:
*                    size : integer size of array
*
* OUTPUT PARAMETERS:
*                    None:
*
* RETURN VALUE:
*                    None:
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.12.04  Kent Myers        Initial implementation
*******************************************************************************}
procedure TTimeSeriesInputSettings.SetActiveChannelLength(size : integer);
begin
  SetLength(FActiveChannels,size);
end;
{!******************************************************************************
* MODULE NAME ...  GetActiveEvents
* SYSTEM NAME ...  Acselerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:  Getter for ActiveEvents dynamic array.
*
* INPUT PARAMETERS:
*                    index: integer  index into array
*
* OUTPUT PARAMETERS:
*                    None:
*
* RETURN VALUE:
*                    result : boolean  state of array element
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.12.04  Kent Myers        Initial implementation
*******************************************************************************}
function TTimeSeriesInputSettings.GetActiveEvents(index : integer) : boolean;
begin
  if((index >= Low(FActiveEvents)) and
     (index <= High(FActiveEvents))) then
  begin
    result := FActiveEvents[index];
  end else
  begin
    result := false;
  end;
end;
{!******************************************************************************
* MODULE NAME ...  SetActiveEvents
* SYSTEM NAME ...  Acselerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:  Setter for ActiveEvents dynamic array.
*
* INPUT PARAMETERS:
*                    index: integer  index into array
*                    state : boolean  state of array element
*
* OUTPUT PARAMETERS:
*                    None:
*
* RETURN VALUE:
*                    None:
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.12.04  Kent Myers        Initial implementation
*******************************************************************************}
procedure TTimeSeriesInputSettings.SetActiveEvents(index : integer; state : boolean);
begin
  if((index >= Low(FActiveEvents)) and
     (index <= High(FActiveEvents))) then
  begin
    FActiveEvents[index] := state;
  end;
end;
{!******************************************************************************
* MODULE NAME ...  SetActiveEventLength
* SYSTEM NAME ...  Acselerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:  Allocate ActiveEvents dynamic array.
*
* INPUT PARAMETERS:
*                    size : integer size of array
*
* OUTPUT PARAMETERS:
*                    None:
*
* RETURN VALUE:
*                    None:
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.12.04  Kent Myers        Initial implementation
*******************************************************************************}
procedure TTimeSeriesInputSettings.SetActiveEventLength(size : integer);
begin
  SetLength(FActiveEvents,size);
end;

end.
