{*****************************************************************************
* UNIT DESCRIPTION:  Base class for input values and methods to load time series
*                    data from a file
* SPEC REF: (if used)
* DESIGN REF: (if used)
*
* Copyright (c) 2009 Schweitzer Engineering Laboratories, Pullman, Washington
******************************************************************************}
unit ClassTimeSeries;

interface

uses
  Windows, Messages, SysUtils, StrUtils, Variants, Classes, DateUtils,
  Contnrs, TimeRoutines, ClassChartInput, TimeSeriesConstants;
const
  PHASE_A = 1;
  PHASE_C = 3;
  COL_RECORD = 0;
  COL_DATE = 1;
  COL_TIME = 2;
  COL_PHASE = 3;
  COL_STATUS = 4;
  SEL351RS_COUNT = 8;
  RELAY_3XX_OFFSET = 1;   // one more column
  RELAY_3XX_START = 7;  // skip Vs
  RELAY_351RS_OFFSET = (-6);  // 6 less columns
  RELAY_4XX_OFFSET = (-5);  // 5 less columns
  CHART_SUBTYPE_LIST : array[0..7] of string = ('Invalid','Unknown',
               'HRM_PIV','HRM_MIV','HRM_MAI','HRM_MAV','HRM_POW','HRM_INT');
type
  THMIChannelPhase = (cpA=0, cpB, cpC, cpOther);
  THMIChannelType = (ctCurrent=0, ctVoltage, ctPower, ctOther);
  THMIEventPhase = (epA=0, epB, epC, epAll);

  TTimeSeriesDeviceType = (dtSEL3XX, dtSEL4XX, dtSEL7XX, dtSEL351RS);
  TTimeSeriesColumnType = (scRecord, scDate, scTime, scPhase, scStatus);
//  TTimeSeriesFileType = (ftUnknown = -1, ftReport, ftCSV, ftText);
  TTimeSeriesCBEMAViolationType = (cvtNone = 0, cvt90, cvt80, cvt70, cvt0, cvt110, cvt120, cvt200);
  TTimeSeriesEventType = (evNone=-1, evSag=0, evSwell, evInterrupt, evAll, evMulti);
  TTimeSeriesChartType = (ctInvalid=-1,ctUndefined=0, ctSSI, ctLDP, ctHarmonic);
  TTimeSeriesChartSubType = (cstInvalid=-1,cstUnknown=0, cstHRM_PIV, cstHRM_MIV,
                             cstHRM_MAI, cstHRM_MAV, cstHRM_POW, cstHRM_INT);
  TTimeSeriesHarmonicType = (htUnknown=-1,htSpectral=0, htComponents, htGrouping, htInterGrouping);
  TTimeSeriesHarmonicDisplayType = (hdtUnknown=-1,hdtAll=0, hdtOdd, hdtEven, hdtInterval);
  TTimeSeriesCompleteType = (ctComplete=0, ctIncompleteStart, ctIncompleteStop, ctIncompleteBoth);

  TTimeSeriesSettingRecord = class(TObject)
  private
    FSettingName            : string;
    FSettingValue           : double;
  protected
  public
    property SettingName  : string read FSettingName write FSettingName;
    property SettingValue : double read FSettingValue write FSettingValue;
  end;

  TTimeSeriesRecord = class(TObject)
  private
  protected
  public
    RecordNumber           : integer;
    RecordValue            : double;  // decimal value of record number (harmonics require this)
    DateTime               : TDateTime;
    Value                  : array of double;
    Phase                  : string;
    Status                 : string;
  end;

  TTimeSeriesIndexRecord = class(TObject)
  private
  protected
  public
    DateTime               : TDateTime;
    FirstRecord            : integer;    // first record of event
    LastRecord             : integer;    // last record of event
    StartEvent             : integer;    // end of pre-disturbance
    StopEvent              : integer;    // start of post-disturbance
    EventName              : string;    // Event number
    Description            : string;
    Duration               : TDateTime;   // span of time
    EventType              : TTimeSeriesEventType;
    Magnitude              : double;
    VA_Vbase               : double;
    VB_Vbase               : double;
    VC_Vbase               : double;
    VAMin                  : double;
    VAMax                  : double;
    VBMin                  : double;
    VBMax                  : double;
    VCMin                  : double;
    VCMax                  : double;
    HasVBase               : boolean;
    Checked                : boolean;    // user selected for display
    Completion             : TTimeSeriesCompleteType;    // state of completion
    Violation              : TTimeSeriesCBEMAViolationType;
    EventTypeOverride      : boolean;  // if true, don't base event type on first record
  end;

  TTimeSeriesLDPEventRecord = class(TObject)
  private
  protected
  public
    DateTime               : TDateTime;
    EventName              : string;
    EventCode              : string;
    Description            : string;
  end;

  TTimeSeriesLDPChannelRecord = class(TObject)
  private
  protected
  public
    ChannelName             : string;
    ChannelFunction         : string;
    ChannelUnits            : string;
    ChannelRules            : string;
    ChannelMult             : double;
    ChannelScale            : double;
  end;

  TTimeSeriesDataTable = class(TObject)
  private
    FRawDataLabels         : TStringList; // list of variable names
    FRawDataTable          : TObjectList;   // list of raw data records
    FIndexTable            : TObjectList;  // list of event index records
    FLDPEventTable         : TObjectList;  // list of LDP event records
    FLDPChannelTable       : TObjectList;  // list of LDP channel records
    FSettingTable          : TObjectList;  // List of setting names/values
    FInputSettings         : TTimeSeriesInputSettings;   // input settings from GUI

    FChartName             : string;  // Name of chart
    FFileType              : TTimeSeriesFileType;  // type of file (CSV, Report)
    FChartType             : TTimeSeriesChartType;  // Type of chart (based on name)
    FChartSubType          : TTimeSeriesChartSubType;  // Sub-type of chart
    FDeviceType            : TTimeSeriesDeviceType;  // type of device (dt734XX)
    FChartVersion          : string; // Chart version
    FChartDateTime         : TDateTime; // Chart Date/Time
    FChartField            : string; // name of first column of charted table

    FMeterID               : string;  // MID of meter
    FMeterFID              : string;  // FID of meter
    FMeterTID              : string;  // TID of meter
    FTimeSource            : string;  // Time source

    FHeaderColumns         : TIntegerArray; // list of header columns in charted table
    FDataColumns           : TIntegerArray;   // list of data columns in charted table
    FDateFormat            : string;   // Date format setting string (MDY,...)
    FVoltageIndices        : TIntegerArray;  // indices of voltages in record
    FHasEventSummary       : boolean; // True if event summary is part of the input

    FLDPUnits              : TTimeSeriesUnitsType;  // LDP Units
    FLDPUnitsBase          : TTimeSeriesUnitsType;  // LDP Base Units : PRI_SCA
    FLDPScale              : TTimeSeriesScaleType;  // LDP Scale
    FLDPRecorder           : word;   // LDP recorder number
  protected

  public
    procedure Clear();
    procedure InitDateTimeFormat(formatType : TDateFormatType; var myFormat : TFormatSettings);
    function GetFileType(fileData : TStringList; var fileDelimiter: System.char) : TTimeSeriesFileType;  overload;
    function GetFileType() : TTimeSeriesFileType; overload;
    function ReadFileType(filePath : string) : TTimeSeriesFileType;
    function ReadChartType(filePath: string;
                var fileData : TStringList;
                var fileType : TTimeSeriesFileType;
                var aChartSubType : TTimeSeriesChartSubType) : TTimeSeriesChartType;  overload;
    function ReadChartType(filePath : string;
                var aChartSubType : TTimeSeriesChartSubType) : TTimeSeriesChartType;  overload;
    function GetChartType(name : string) : TTimeSeriesChartType;
    function GetChartSubType(name : string) : TTimeSeriesChartSubType;
    function GetReportType(fileData,recordData : TStringList;
             var myChartSubType : TTimeSeriesChartSubType): TTimeSeriesChartType;
    function LoadData(fileData : TStringList; xColumns, yColumns : array of integer) : boolean;
    procedure LoadVSSISummaryData(fileData : TStringList; irecord : integer); virtual;
    function ConvertTimeToDuration(timeString : String; aFormat : TFormatSettings) : TDateTime;
    procedure ReadFileCSVVSSIEventData(var iRecord : integer; fileData : TStringList);
    procedure FindEventRecords( indexRecord : TTimeSeriesIndexRecord);
    function DetermineDeviceType(fileType : TTimeSeriesFileType; columnCount, lastColumn : integer): TTimeSeriesDeviceType;
    function IsValidDateTime(myDateTime:string): boolean;
    function GetHeaderColumn(Index: integer) : integer;
    function GetDataColumn(Index: integer) : integer;
    procedure SetHeaderColumn(Index: Integer; Value: integer);
    procedure SetDataColumn(Index: Integer; Value: integer);
    procedure SetLengthHeaderColumns(size : integer);
    procedure SetLengthDataColumns(size : integer);
    function CBEMAViolation( indexRecord : TTimeSeriesIndexRecord) : TTimeSeriesCBEMAViolationType;

    function ReadFileData( filePath : string; var fileData, recordData : TStringList;
                          var fileType: TTimeSeriesFileType; var myHeader,
                          myData : array of integer): boolean;
    procedure ReadFileCSVHeaderData(fileData, recordData : TStringList);
    procedure ReadFileCSVDeviceData(var iRecord : integer; fileData,
                                    recordData : TStringList);
    procedure ReadFileReportDeviceData(var iRecord : integer; fileData,
                                       recordData : TStringList);
    procedure ReadFileCSVChannelData(var iRecord : integer; fileData,
                                     recordData : TStringList; eventField : string);
    procedure ReadFileCSVEventData(var iRecord : integer; fileData,
                                     recordData : TStringList);

    function ReadChartRecord(
                   fileType : TTimeSeriesFileType;
                   recordData : TStringList;
                   myFormat : TFormatSettings;
                   myFormatEnglish : TFormatSettings;
                   xColumns, yColumns : array of integer;
                   columnCount : integer;
                   dataCount : integer;
                   var myStatus : string) : boolean;
    function ReadChartRecordNumber(
                   recordData : TStringList;
                   xColumns : array of integer;
                   dataRecord : TTimeSeriesRecord) : boolean;
    function ReadChartRecordDate(
                   recordData : TStringList;
                   xColumns : array of integer;
                   fileType : TTimeSeriesFileType;
                   myFormat : TFormatSettings;
                   myFormatEnglish : TFormatSettings;
                   dataRecord : TTimeSeriesRecord) : boolean;
    function ReadChartRecordPhase(
                   recordData : TStringList;
                   xColumns : array of integer;
                   dataRecord : TTimeSeriesRecord) : boolean;
    function ReadChartRecordStatus(
                   recordData : TStringList;
                   xColumns : array of integer;
                   dataRecord : TTimeSeriesRecord) : boolean;
    function ReadChartRecordData(
                   recordData : TStringList;
                   dataCount : integer;
                   yColumns : array of integer;
                   fileType   : TTimeSeriesFileType;
                   dataRecord : TTimeSeriesRecord) : boolean;
    function ReportFileHasStatusColumn(fileData : TStringList;
                   var count : integer) : boolean;
    function ReportType(filePath : string) : TTimeSeriesChartType;
    function IsInterruptRecord(dataRecord : TTimeSeriesRecord) : boolean;
    function NonLocaleStrToFloatDef(value : String; defaultValue : single) : single;
    function TryNonLocaleStrToFloat(svalue : String; var value : double) : boolean;

    property RawDataLabels     : TStringList read FRawDataLabels write FRawDataLabels;
    property RawDataTable      : TObjectList read FRawDataTable write FRawDataTable;
    property IndexTable        : TObjectList read FIndexTable write FIndexTable;
    property LDPEventTable     : TObjectList read FLDPEventTable write FLDPEventTable;
    property LDPChannelTable   : TObjectList read FLDPChannelTable write FLDPChannelTable;
    property SettingTable      : TObjectList read FSettingTable;
    property InputSettings     : TTimeSeriesInputSettings read FInputSettings write FInputSettings;
    property ChartField        : string read FChartField write FChartField;
    property ChartName         : string read FChartName write FChartName;
    property ChartType         : TTimeSeriesChartType read FChartType write FChartType;
    property ChartSubType      : TTimeSeriesChartSubType read FChartSubType write FChartSubType;
    property DeviceType        : TTimeSeriesDeviceType read FDeviceType;
    property FileType          : TTimeSeriesFileType read FFileType;
    property ChartVersion      : string read FChartVersion write FChartVersion;
    property ChartDateTime     : TDateTime read FChartDateTime write FChartDateTime;
    property MeterID           : string read FMeterID write FMeterID;  // MID of meter
    property MeterFID          : string read FMeterFID write FMeterFID;  // FID of meter
    property MeterTID          : string read FMeterTID write FMeterTID;  // TID of meter
    property TimeSource        : string read FTimeSource write FTimeSource;  // TimeSource of meter
    property DataColumnArray   : TIntegerArray read FDataColumns write FDataColumns;
    property HeaderColumnArray : TIntegerArray read FHeaderColumns write FHeaderColumns;
    property HeaderColumns[Index: integer]: integer read GetHeaderColumn write SetHeaderColumn;
    property DataColumns[Index: integer]: integer read GetDataColumn write SetDataColumn;
    property DateFormat        : string read FDateFormat write FDateFormat;  // date format string
    property LDPUnits          : TTimeSeriesUnitsType read FLDPUnits write FLDPUnits;
    property LDPUnitsBase      : TTimeSeriesUnitsType read FLDPUnitsBase write FLDPUnitsBase;
    property LDPScale          : TTimeSeriesScaleType read FLDPScale write FLDPScale;
    property LDPRecorder       : word read FLDPRecorder write FLDPRecorder;
    property VoltageIndices    : TIntegerArray read FVoltageIndices write FVoltageIndices;
    property HasEventSummary   : boolean read FHasEventSummary write FHasEventSummary;
    constructor Create;
    destructor  Destroy; override;

  end;

  function CompareRecords(item1,item2: Pointer): Integer;

implementation

//uses
//  SELUtilitiesFuncs;

{!******************************************************************************
* MODULE NAME ...  Create
* SYSTEM NAME ...  Acselerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:  TimeSeriesDataTable is used to hold channel data obtained
*                          from the device. The data may be used to generated a
*                          graphical chart or to export as CSV data file.
*
* INPUT PARAMETERS:
*                    None:
*
* OUTPUT PARAMETERS:
*                    None:
*
* RETURN VALUE:
*                    None:
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.09.24  Kent Myers        Initial implementation
*******************************************************************************}
constructor TTimeSeriesDataTable.Create;
begin
  inherited Create;

  FRawDataTable := TObjectList.Create;    // raw meter data
  FRawDataLabels := TStringList.Create;   // channel names
  FIndexTable := TObjectList.Create;      // event index (used for SSI)
  FLDPEventTable := TObjectList.Create;      // event index (used for SSI)
  FLDPChannelTable := TObjectList.Create;      // event index (used for SSI)
  FSettingTable := TObjectList.Create;      // event index (used for SSI)
  FChartField := '#';               // default column name for charted data
  FChartName := 'None';
  FChartDateTime := Now;
  FChartSubType := cstUnknown;
  FChartVersion := '1.0';
  FTimeSource := 'Unknown';
  FMeterID := 'SEL-734';  // MID of meter
  FMeterFID := 'FID';  // FID of meter
  FMeterTID := 'TID';  // TID of meter
  FDateFormat := 'mm/dd/yyyy';
  FLDPScale := stUnknown;
  FLDPUnits := utSecondary;
  FLDPRecorder := 0;
  SetLength(FVoltageIndices,4);

end;
{!******************************************************************************
* MODULE NAME ...  Destroy
* SYSTEM NAME ...  Acselerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:
* destruction!!!!
*
* INPUT PARAMETERS:
*                    None:
* 
* OUTPUT PARAMETERS:
*                    None:                        
*
* RETURN VALUE:
*                    None:                        
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.09.24  Kent Myers        Initial implementation
*******************************************************************************}
destructor TTimeSeriesDataTable.Destroy;
begin
  FreeAndNil(FRawDataTable);
  FreeAndNil(FRawDataLabels);
  FreeAndNil(FIndexTable);
  FreeAndNil(FLDPEventTable);
  FreeAndNil(FLDPChannelTable);
  FreeAndNil(FSettingTable);
  inherited Destroy;
end;
{************************************************************************
* NAME:       Clear
* FUNCTIONAL DESCRIPTION:  Clears content of each table (without destroying the table)
* INPUT PARAMETERS: none
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************}
procedure TTimeSeriesDataTable.Clear();
begin
  FRawDataTable.Clear;
  FRawDataLabels.Clear;
  FIndexTable.Clear;
  FLDPEventTable.Clear;
  FLDPChannelTable.Clear;
  FSettingTable.Clear;
end;
{************************************************************************
* NAME:       SetLengthHeaderColumns
* FUNCTIONAL DESCRIPTION:    Set length of header columns
* INPUT PARAMETERS: size : integer array size
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************}
procedure TTimeSeriesDataTable.SetLengthHeaderColumns(size: Integer);
begin
  SetLength(FHeaderColumns,size);
end;
{************************************************************************
* NAME:       SetLengthDataColumns
* FUNCTIONAL DESCRIPTION:    Set length of data columns
* INPUT PARAMETERS: size : integer array size
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************}
procedure TTimeSeriesDataTable.SetLengthDataColumns(size: Integer);
begin
  SetLength(FDataColumns,size);
end;
{************************************************************************
* NAME:       GetHeaderColumn
* FUNCTIONAL DESCRIPTION:    Get a column from the header column array
* INPUT PARAMETERS: Index : integer column number
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************}
function  TTimeSeriesDataTable.GetHeaderColumn(Index: Integer) : integer;
begin
  result := FHeaderColumns[Index];
end;
{************************************************************************
* NAME:       SetHeaderColumn
* FUNCTIONAL DESCRIPTION:    Set a column from the data column array
* INPUT PARAMETERS: Index : integer  column number
*                   Value : integer  value
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************}
procedure TTimeSeriesDataTable.SetHeaderColumn(Index: Integer; Value: integer);
begin
  FHeaderColumns[Index] := Value;
end;
{************************************************************************
* NAME:       GetDataColumn
* FUNCTIONAL DESCRIPTION:    Get a column from the data column array
* INPUT PARAMETERS: Index : integer  column number
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************}
function  TTimeSeriesDataTable.GetDataColumn(Index: Integer) : integer;
begin
  result := FDataColumns[Index];
end;
{************************************************************************
* NAME:       SetDataColumn
* FUNCTIONAL DESCRIPTION:    Set a column from the data column array
* INPUT PARAMETERS: Index : integer  column number
*                   Value : integer  value
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************}
procedure TTimeSeriesDataTable.SetDataColumn(Index: Integer; Value: integer);
begin
  FDataColumns[Index] := Value;
end;
{!******************************************************************************
* MODULE NAME ...  GetFileType
* SYSTEM NAME ...  Acselerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:
* Determine file type and delimiter by examining record
*
* INPUT PARAMETERS:
*                    fileData: string                sample data record
*
* OUTPUT PARAMETERS:
*                    fileDelimiter: string   delimiter to use when parsing file
* RETURN VALUE:     fileType :  TTimeSeriesFileType  type of file (ftCSV or ftReport)
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.12.07  Kent Myers        Initial implementation
*******************************************************************************}
function TTimeSeriesDataTable.GetFileType(fileData : TStringList;
         var fileDelimiter: System.char) : TTimeSeriesFileType;
var
  fileType : TTimeSeriesFileType;
begin
  if(fileData.Count > 0) then
  begin
    if(AnsiContainsStr(fileData[0],',')) then  // it is a CSV file
    begin
      fileType := ftCSV;
      fileDelimiter := ',';
    end else
    begin
      fileType := ftReport;
      fileDelimiter := ' ';
    end;
  end else
  begin
    fileType := ftUnknown;
    fileDelimiter := ' ';
  end;
  FFileType := fileType;
  result := fileType;
end;

{************************************************************************
* NAME:       GetFileType
* FUNCTIONAL DESCRIPTION:    Get file type
* INPUT PARAMETERS: none
* OUTPUT PARAMETERS: none
* RETURN: result : TTimeSeriesFileType  type of file (ftCSV, ftReport)
************************************************************************}
function TTimeSeriesDataTable.GetFileType() : TTimeSeriesFileType;
begin
  result := FFileType;
end;

{************************************************************************
* NAME:       ReadFileType
* FUNCTIONAL DESCRIPTION:    Read file type from file (if found)
* INPUT PARAMETERS: filePath : string  file name
* OUTPUT PARAMETERS: none
* RETURN: result : TTimeSeriesFileType  type of file (ftCSV, ftReport)
************************************************************************}
function TTimeSeriesDataTable.ReadFileType(filePath : string) : TTimeSeriesFileType;
var
  fileData, recordData : TStringList;
  delim : System.char;
  fileType : TTimeSeriesFileType;
begin

  if(FileExists(filePath)) then
  begin
    fileData := TStringList.Create;
    recordData := TStringList.Create;
    try
      fileData.LoadFromFile(filePath);
      fileType := GetFileType(fileData,delim);
    finally
      fileData.Free;
      recordData.Free;
    end;
  end else
  begin
    fileType := ftUnknown;
  end;
  result := fileType;
end;
{************************************************************************
* NAME:       ReadChartType
* FUNCTIONAL DESCRIPTION:    Read chart type from file (if found)
* INPUT PARAMETERS: filePath : string  file name
* OUTPUT PARAMETERS: aChartSubType : TTimeSeriesChartSubType
* RETURN: result : TTimeSeriesChartType  type of chart data found in file
************************************************************************}
function TTimeSeriesDataTable.ReadChartType(filePath : string;
         var aChartSubType : TTimeSeriesChartSubType) : TTimeSeriesChartType;
var
  fileType : TTimeSeriesFileType;
  fileData : TStringList;
begin
  fileData := TStringList.Create;
  try
    result := ReadChartType(filePath,fileData,fileType,aChartSubType);
  finally
    fileData.Free;
  end;
end;
{************************************************************************
* NAME:       ReadChartType
* FUNCTIONAL DESCRIPTION:    Read chart type from file (if found)
* INPUT PARAMETERS: filePath : string  file name
* OUTPUT PARAMETERS: fileType : TTimeSeriesFileType  type of file (ftReport,ftCSV)
*                  aChartSubType : TTimeSeriesChartSubType
* RETURN: result : TTimeSeriesChartType  type of chart data found in file
************************************************************************}
function TTimeSeriesDataTable.ReadChartType(filePath : string;
                var fileData : TStringList;
                var fileType : TTimeSeriesFileType;
                var aChartSubType : TTimeSeriesChartSubType) : TTimeSeriesChartType;
const
  RECORD_VERSION = 1;
  FIELD_CHARTNAME = 0;
  FIELD_CHARTSUBTYPE = 9;
var
  recordData : TStringList;
  delim : System.char;
  myChartType : TTimeSeriesChartType;
begin

  if(FileExists(filePath)) then
  begin
    recordData := TStringList.Create;
    try
      fileData.LoadFromFile(filePath);

      fileType := GetFileType(fileData,delim);
      if(fileType = ftUnknown) then
      begin
        result := ctUndefined;         // invalid file
        exit;
      end;
      recordData.Delimiter := delim;

      if((fileType = ftCSV) and (fileData.Count > RECORD_VERSION)) then   // get standard columns and data columns
      begin
        recordData.CommaText := fileData[RECORD_VERSION];    // get Chart record name
        FChartName := recordData[FIELD_CHARTNAME];
        myChartType := GetChartType(ChartName);
        if(recordData.Count > FIELD_CHARTSUBTYPE) then
        begin
          aChartSubType := TTimeSeriesChartSubType(
             StringInArray(recordData[FIELD_CHARTSUBTYPE],CHART_SUBTYPE_LIST)-1);
        end else
        begin
          aChartSubType := cstUnknown;
        end;
      end else
      if(fileType = ftReport) then
      begin
        ChartField := '#';  // report file data delimiter
        myChartType := GetReportType(fileData,recordData,aChartSubType);
        if(myChartType = ctUndefined) then
        begin
          ChartField := 'Harmonic';  // report file data delimiter
          myChartType := GetReportType(fileData,recordData,aChartSubType);
          if(myChartType = ctUndefined) then
          begin
            ChartField := 'Frequency';  // report file data delimiter
            myChartType := GetReportType(fileData,recordData,aChartSubType);
          end;
        end;
      end else
      begin
        myChartType := ctUndefined;
      end;

    finally
      recordData.Free;
    end;
  end else
  begin
    myChartType := ctUndefined;
  end;
  result := myChartType;
end;

{************************************************************************
* NAME:       GetChartType
* FUNCTIONAL DESCRIPTION:    Determine Chart type from chart name
* INPUT PARAMETERS:  name : string chart name
* OUTPUT PARAMETERS:  none
* RETURN: result : TTimeSeriesChartType  type of chart data found in file
************************************************************************}
function TTimeSeriesDataTable.GetChartType(name : string) : TTimeSeriesChartType;
const
  CHART_TYPES : array [0..4] of string  = ('None','SSI','VSSI','LDP','HRM');
var
  i,found : Integer;
  myChartType : TTimeSeriesChartType;
begin
  found := -1;
  for i := Low(CHART_TYPES) to High(CHART_TYPES) do
  begin
    if(name = CHART_TYPES[i]) then
    begin
      found := i;
      break;
    end;
  end;

  case found of
    1,2: // SSI,VSSI
      begin
        myChartType := ctSSI;
      end;
    3:
      begin
        myChartType := ctLDP;
      end;
    4:
      begin
        myChartType := ctHarmonic;
      end;
    else
    begin
      myChartType := ctUndefined;
    end;
  end;
  result := myChartType;
end;
{************************************************************************
* NAME:       GetChartSubType
* FUNCTIONAL DESCRIPTION:    Determine Chart sub-type from chart sub-type name
* INPUT PARAMETERS:  name : string chart sub-type
* OUTPUT PARAMETERS:  none
* RETURN: result : TTimeSeriesChartSubType  type of chart data found in file
************************************************************************}
function TTimeSeriesDataTable.GetChartSubType(name : string) : TTimeSeriesChartSubType;
begin
  result := TTimeSeriesChartSubType(StringInArray(name,CHART_SUBTYPE_LIST)-1);
end;
{!******************************************************************************
* MODULE NAME ...  ReadFileData
* SYSTEM NAME ...  Acselerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:
* Reads the file until the label record is found
*
* INPUT PARAMETERS:
*                    filePath: string                the stream to load from
*
* OUTPUT PARAMETERS:
*                    fileData: TStringList containing data content of file
*                    recordData: TStringList containing parsed header record (channel names)
*                    fileType:  TTimeSeriesFileType  ftCSV or ftReport
* RETURN VALUE:
*                    found : header record found true/false
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.09.24  Kent Myers        Initial implementation
* 2008.12.04  Kent Myers        Add test for file type
*******************************************************************************}
function TTimeSeriesDataTable.ReadFileData(filePath : string;
         var fileData, recordData  : TStringList;
         var fileType: TTimeSeriesFileType;
         var myHeader, myData : array of integer) : boolean;
const
  RECORD_DATA = 6;
  VSSI_EVENT_FIELD = 'Event';
  LDP_EVENT_FIELD = 'EventDate';
var
  recordCount, iRecord : Integer;
  chartFound, deviceFound, eventFound, channelFound : boolean;
  deviceField, channelField : string;
  aChartSubType : TTimeSeriesChartSubType;
begin

  chartFound := false;       // assume valid header record not found
  HasEventSummary := false;  // assume no event summary in the input

  // Each of these flags corresponds to a table in a CSV file

  deviceFound := false;
  eventFound := false;
  channelFound := false;

  if(FileExists(filePath)) then
  begin
    try
      FChartType := ReadChartType(filePath,fileData,fileType,aChartSubType);
      FChartSubType := aChartSubType;
      if(fileType = ftReport) then
      begin
        recordData.Delimiter := ' ';
      end else
      begin
        recordData.Delimiter := ',';
      end;

      iRecord := 0;  // index of current record in file
      recordCount := fileData.Count;

      if(recordCount <= 2) then    // cannot have valid data
      begin
        FChartType := ctUndefined;
        result := false;;
        exit;
      end;

      // keywords used to identify each table in file

      deviceField := 'DeviceID';  // column name of device information table
      channelField := 'ChannelName';  // column name of channel table

      if(fileType = ftCSV) then   // get standard X and Y data columns
      begin
        ReadFileCSVHeaderData(fileData, recordData);
        iRecord := RECORD_DATA;   // record where non-header CSV tables start
      end;

      // Read tables that are not chart data

      while iRecord < recordCount  do
      begin
        recordData.DelimitedText := fileData[iRecord];
        if(recordData.Count > 0) then
        begin
          if(not deviceFound) then  // no device information found yet
          begin
            if(fileType = ftCSV) then
            begin
              if(AnsiCompareText(recordData[0],deviceField) = 0) then
              begin
                ReadFileCSVDeviceData(iRecord,fileData,recordData);
                deviceFound := true;
              end;
            end else
            if(fileType = ftReport) then
            begin
              ReadFileReportDeviceData(iRecord,fileData,recordData);
              deviceFound := true;
            end;
          end else
          if(not channelFound and (fileType = ftCSV) and (ChartType = ctLDP) and
            (AnsiCompareText(recordData[0],ChannelField) = 0)) then
          begin
            ReadFileCSVChannelData(iRecord,fileData,recordData,LDP_EVENT_FIELD);
            channelFound := true;
          end else
          if(not eventFound and (fileType = ftCSV) and (ChartType = ctLDP) and
            (AnsiCompareText(recordData[0],LDP_EVENT_FIELD) = 0)) then
          begin
            ReadFileCSVEventData(iRecord,fileData,recordData);
            eventFound := true;
            Dec(iRecord);  // set pointer back to re-read this record
          end else
          if(not eventFound and (fileType = ftCSV) and (ChartType = ctSSI) and
            (AnsiCompareText(recordData[0],VSSI_EVENT_FIELD) = 0)) then
          begin
            ReadFileCSVVSSIEventData(iRecord,fileData);
            eventFound := true;
            Dec(iRecord);  // set pointer back to re-read this record
          end else
          if(not chartFound and
            (AnsiCompareText(recordData[0],ChartField) = 0)) then
          begin
             chartFound := true;  // found start of chart data so stop reading
             break;
          end;
        end;
        Inc(iRecord);
      end;
    except
      chartFound := false;
      FChartType := ctUndefined;
    end;
  end;
  result := chartFound;
end;
{************************************************************************
* NAME:       ReportFileHasStatusColumn
* FUNCTIONAL DESCRIPTION:   determine whether file has a status column
* INPUT PARAMETERS: fileData : TStringList   contains content of file
* OUTPUT PARAMETERS:
*                   count : integer   number of columns of data
* RETURN: result : boolean   true if file has Status column
************************************************************************}
function TTimeSeriesDataTable.ReportFileHasStatusColumn(fileData : TStringList; var count : integer) : boolean;
var
  iRecord, recordCount : integer;
  recordData : TStringList;
  hasStatus : boolean;
begin
  iRecord := 0;  // index of current record in file
  recordCount := fileData.Count;
  hasStatus := false;

  recordData := TStringList.Create;
  try
    recordData.Delimiter := ' ';
    while iRecord < recordCount  do
    begin
      recordData.DelimitedText := fileData[iRecord];
      if(recordData.Count > 0) then
      begin
        if(recordData[0] = ChartField) then
        begin
          count := recordData.Count;
          hasStatus := (UpperCase(recordData[count-1]) = 'STATUS');
          break;
        end;
      end;
      Inc(iRecord);
    end;

  finally
    recordData.Free;
  end;
  result := hasStatus;
end;
{************************************************************************
* NAME:       ReportType
* FUNCTIONAL DESCRIPTION:   return type of report file
* INPUT PARAMETERS: fileData : TStringList   contains content of file
* OUTPUT PARAMETERS:
* RETURN: result : result : TTimeSeriesChartType  (ctSSI, ctLDP, ctUnknown)
************************************************************************}
function TTimeSeriesDataTable.ReportType(filePath : string) : TTimeSeriesChartType;
var
  recordData : TStringList;
  fileData : TStringList;
  myChartType : TTimeSeriesChartType;
  myChartSubType : TTimeSeriesChartSubType;
begin
  myChartType := ctUndefined;

  if(FileExists(filePath)) then
  begin
    fileData := TStringList.Create;
    recordData := TStringList.Create;
    try
      fileData.LoadFromFile(filePath);
      myChartType := GetReportType(fileData,recordData,myChartSubType);
    finally
      recordData.Free;
      fileData.Free;
    end;
  end;
  result := myChartType;
end;
{************************************************************************
* NAME:       GetReportType
* FUNCTIONAL DESCRIPTION:   determine whether file has a SSI report content
* INPUT PARAMETERS: fileData : TStringList   contains content of file
*                   recordData : TstringList  use for parsing file
* OUTPUT PARAMETERS: myChartSubType : TTimeSeriesChartSubType
* RETURN: result :  fileType : TTimeSeriesChartType
************************************************************************}
function TTimeSeriesDataTable.GetReportType(fileData,
            recordData : TStringList;
            var myChartSubType : TTimeSeriesChartSubType) : TTimeSeriesChartType;
var
  iRecord : integer;
  myChartType : TTimeSeriesChartType;
  delim : System.char;
begin
  iRecord := 0;  // index of current record in file
  myChartType := ctUndefined;  // assume unknown type

  GetFileType(fileData,delim);
  recordData.Delimiter := delim;
  myChartSubType := cstUnknown;

  while iRecord < fileData.Count  do
  begin
    recordData.DelimitedText := fileData[iRecord];
    if(recordData.Count > 0) then
    begin
      if(recordData[0] = ChartField) then  // reached label header (can't be SSI)
      begin
        if((ChartField = 'Harmonic') or (ChartField = 'Frequency')) then
        begin
          myChartType := ctHarmonic;
          if(ChartField = 'Frequency') then
          begin
            myChartSubType := cstHRM_INT;     // Spectral Analysis
          end else
          begin
            if(recordData[1] = 'Mag') then  // it is a mag/angle report
            begin
              recordData.DelimitedText := fileData[iRecord-1];
              if(recordData[0] = 'IA') then
              begin
                myChartSubType := cstHRM_MAI;   // Mag/Angle currents
              end else
              begin
                myChartSubType := cstHRM_MAV;    // Mag/Angle Volts
              end;
            end else
            if((recordData[1] = 'PA') or (recordData[1] = 'P3')) then
            begin
              myChartSubType := cstHRM_POW;  // Power
            end else
            begin
              if(AnsiContainsStr(fileData[iRecord+1],'%')) then
              begin
                myChartSubType := cstHRM_PIV; // percent IV
              end else
              begin
                myChartSubType := cstHRM_MIV; // magnitude IV
              end;
            end;

          end;
        end else
        begin
          myChartType := ctLDP;  // must be LDP
        end;
        break;
      end else
      if((AnsiContainsText(fileData[iRecord],'Voltage(%')) or
       (AnsiContainsText(fileData[iRecord],'Current(%')) or
       (AnsiContainsText(fileData[iRecord],'Current I (%'))) then
      begin
        myChartType := ctSSI;  // SSI type
        break;
      end else
      if(iRecord > 20) then
      begin
        break;  // no need to keep looking
      end;
    end;
    Inc(iRecord);
  end;
  result := myChartType;
end;
{************************************************************************
* NAME:       ReadFileCSVDeviceData
* FUNCTIONAL DESCRIPTION:    Read data describing the device (MID, FID, TID, settings)
* INPUT PARAMETERS: iRecord : current record number
*                   fileData : TStringList   contains content of file
*                   recordData : TStringList  contains content of record
* OUTPUT PARAMETERS: none
* RETURN: result : TTimeSeriesChartType  type of chart data found in file
************************************************************************}
procedure TTimeSeriesDataTable.ReadFileCSVHeaderData(fileData, recordData : TStringList);
const
  RECORD_VERSION = 1;
  RECORD_STD_COL = 3;
  RECORD_DATA_COL = 5;
  FIELD_CHARTNAME = 0;
  FIELD_CHARTVERSION = 1;
  FIELD_CHARTDATE = 2;
  FIELD_CHARTTIME = 3;
  FIELD_CHARTFIELD = 4;
  FIELD_TIMESOURCE = 5;
  FIELD_RECORDER = 6;
  FIELD_SCALE = 7;
  FIELD_UNITS = 8;
  FIELD_CHART_SUBTYPE = 9;
  FIELD_BASE_UNITS = 10;
  FIELD_CHARTCOUNT = 5;  // TIMESOURCE,UNITS,SCALE were added so they aren't counted
  HEADER_RECORD_MSG = 'Insufficient header records in CSV file.';
  HEADER_DATA_MSG = 'Insufficient header data in CSV file.';
  HEADER_CHEADER_MSG = 'Invalid header column.';
  HEADER_CDATA_MSG = 'Invalid data column.';
  HEADER_CHART_TYPE_MSG = 'Unknown chart type in data file';
var
  i, iValue : integer;
  myDate, myTime : string;
  myFormat : TFormatSettings;
begin
  InitDateTimeFormat(ftEnglish_US,myFormat);
  myFormat.ShortDateFormat := 'MM/DD/YY';   // must be standard English

  // are there enough records?
  if(RECORD_VERSION > (fileData.Count - 1)) then
  begin
    raise Exception.Create(HEADER_RECORD_MSG);
  end;
  recordData.CommaText := fileData[RECORD_VERSION];    // get Chart record name

  // are there enough fields in the record?
  if(recordData.Count < FIELD_CHARTCOUNT) then   // original field count
  begin
    raise Exception.Create(HEADER_DATA_MSG);
  end;

  FChartName := recordData[FIELD_CHARTNAME];

  myDate := recordData[FIELD_CHARTDATE];
  myTime := recordData[FIELD_CHARTTIME];

  FChartDateTime := StrToDateTime(myDate + ' ' + myTime, myFormat);

  FChartField := recordData[FIELD_CHARTFIELD];

  {Timesource is a new field, so test to see if it is available}
  if(recordData.Count > FIELD_CHARTCOUNT) then   // added timesource to header
  begin
    FTimeSource := recordData[FIELD_TIMESOURCE];
  end else
  begin
    FTimeSource := 'Unknown';      // In case timesource is not present
  end;

  if(recordData.Count > FIELD_RECORDER) then  // test for LDP recorder number
  begin
    if(not TryStrToInt(recordData[FIELD_RECORDER],iValue)) then
    begin
      raise Exception.Create(HEADER_CHEADER_MSG);
    end else
    begin
      FLDPRecorder := iValue;
    end;
  end;

  if(recordData.Count > FIELD_SCALE) then   // Test for LDP Scale
  begin
    if(recordData[FIELD_SCALE] = 'Unity') then
    begin
      FLDPScale := stUnity;
    end else
    if(recordData[FIELD_SCALE] = 'Kilo') then
    begin
      FLDPScale := stKilo;
    end else
    if(recordData[FIELD_SCALE] = 'Mega') then
    begin
      FLDPScale := stMega;
    end else
    begin
      FLDPScale := stUnknown;
    end;
  end else
  begin
    FLDPScale := stUnity;
  end;

  if(recordData.Count > FIELD_UNITS) then  // Test for LDP Units
  begin
    if(recordData[FIELD_UNITS] = 'Primary') then
    begin
      FLDPUnits := utPrimary;
    end else
    if(recordData[FIELD_UNITS] = 'Secondary') then
    begin
      FLDPUnits := utSecondary;
    end else
    begin
      FLDPUnits := utUnknown;
    end;
  end else
  begin
    FLDPUnits := utSecondary;
  end;

  if(recordData.Count > FIELD_CHART_SUBTYPE) then  // Test for LDP Units
  begin
    FChartSubType := GetChartSubType(recordData[FIELD_CHART_SUBTYPE]);
  end else
  begin
    FChartSubType := cstInvalid;
  end;

  if(recordData.Count > FIELD_BASE_UNITS) then  // Test for LDP Units
  begin
    if(recordData[FIELD_BASE_UNITS] = 'Primary') then
    begin
      FLDPUnitsBase := utPrimary;
    end else
    if(recordData[FIELD_BASE_UNITS] = 'Secondary') then
    begin
      FLDPUnitsBase := utSecondary;
    end else
    begin
      FLDPUnitsBase := utUnknown;
    end;
  end else
  begin
    FLDPUnitsBase := utSecondary;
  end;

  FChartType := GetChartType(FChartName);

  if(ChartType = ctUndefined) then   // is there a valid chart type in file?
  begin
    raise Exception.Create(HEADER_CHART_TYPE_MSG);
  end;

  recordData.DelimitedText := fileData[RECORD_STD_COL];  // Indep. var. record

  // get column indices for chart table that are indep. vars (rec, Date, Time, etc.)

  SetLengthHeaderColumns(recordData.Count);
  for i := 0 to recordData.Count-1 do
  begin
    if(not TryStrToInt(recordData[i],iValue)) then
    begin
      raise Exception.Create(HEADER_CHEADER_MSG);
    end else
    begin
      FHeaderColumns[i] := iValue;
    end;
  end;

  // get column indices for chart table that are dependent vars (channels)

  recordData.DelimitedText := fileData[RECORD_DATA_COL]; // Dependent var. data columns

  SetLengthDataColumns(recordData.count);
  for i := 0 to recordData.Count-1 do
  begin
    if(not TryStrToInt(recordData[i],iValue)) then
    begin
      raise Exception.Create(HEADER_CDATA_MSG);
    end else
    begin
      FDataColumns[i] := iValue;
    end;
  end;
end;
{************************************************************************
* NAME:       ReadFileCSVDeviceData
* FUNCTIONAL DESCRIPTION:    Read data describing the device (MID, FID, TID, settings)
* INPUT PARAMETERS: iRecord : current record number
*                   fileData : TStringList   contains content of file
*                   recordData : TStringList  contains content of record
* OUTPUT PARAMETERS: none
* RETURN: result : TTimeSeriesChartType  type of chart data found in file
************************************************************************}
procedure TTimeSeriesDataTable.ReadFileCSVDeviceData(var iRecord : integer; fileData, recordData : TStringList);
const
  DEVICE_ID = 0;
  DEVICE_FID = 1;
  DEVICE_TID = 2;
  DEVICE_ID_COUNT = 3;
var
  i : integer;
  settingList : TStringList;
  settingRecord : TTimeSeriesSettingRecord;
begin
  SettingTable.Clear;
  settingList := TStringList.Create;
  try
    // read setting names from incoming recordData
    for i := DEVICE_ID_COUNT to recordData.Count - 1 do    // skip ID's
    begin
      settingList.Add(recordData[i]);
    end;
    Inc(iRecord);

    recordData.CommaText := fileData[iRecord];  // next record
    if(recordData.Count < DEVICE_ID_COUNT) then
    begin
      raise Exception.Create('Insufficient device data in CSV file.');
    end;

    FMeterID := recordData[DEVICE_ID];     // get device ID's
    FMeterFID := recordData[DEVICE_FID];
    FMeterTID := recordData[DEVICE_TID];

    // read setting values
    for i := DEVICE_ID_COUNT to recordData.Count-1 do  // skip ID's
    begin
      settingRecord := TTimeSeriesSettingRecord.Create;
      FSettingTable.Add(settingRecord);
      settingRecord.FSettingName := settingList[i-DEVICE_ID_COUNT];
      settingRecord.FSettingValue := StrToFloat(recordData[i]);
    end;

  finally
    settingList.Free;
  end;
end;

{************************************************************************
* NAME:       ReadFileReportDeviceData
* FUNCTIONAL DESCRIPTION:    Read data describing the device (MID, FID, TID, settings)
* INPUT PARAMETERS: iRecord : current record number
*                   fileData : TStringList   contains content of file
*                   recordData : TStringList  contains content of record
* OUTPUT PARAMETERS: none
* RETURN: result : TTimeSeriesChartType  type of chart data found in file
************************************************************************}
procedure TTimeSeriesDataTable.ReadFileReportDeviceData(var iRecord : integer; fileData, recordData : TStringList);
const
  DATA_COLUMN_MSG = 'Insufficient data columns in file.';
  DATA_RECORD_MSG = 'Insufficient data records in file.';
  INVALID_FILE_MSG = 'Invalid file.';
  CHART_TYPE_MSG = 'Unknown chart type.';
  DATE_STRING = 'Date:';
  TIME_STRING = 'Time:';
  TIMESOURCE_STRING = 'Time Source:';
  VBASE_STRING = 'VBASE';
  FUND_STRING = 'Fundamental';
  VBASE_OFFSET = 8;
  MAX_TID_STRING = 40;
  VBASE_OFFSET_HMI_V101 = 4;
  VBASE_OFFSET_HMI_V100 = 3;
  QUOTE = #39;
  NOMINAL_ABCG_STRING = 'NOMINAL_ABCG';
  NOMINAL_N_STRING = 'NOMINAL_N';
  NOMINAL_ABCG_ITEM = 7;
  NOMINAL_N_ITEM = 11;
  SKIP_OVER_FID_LABEL = 5;
var
  pos1,pos2 : integer;
  settingRecord : TTimeSeriesSettingRecord;
  i, iValue, vBasePos : Integer;
  myDate, myTime, myString, trimRecord : string;
  myFormat : TFormatSettings;
  nomABCG, nomN : double;
begin
  Dec(iRecord);   // start one record back
  pos1 := 0;
  while(pos1 = 0) do  // scan for record containing date
  begin
    Inc(iRecord);
    if(iRecord >= fileData.Count) then
    begin
      raise Exception.Create(INVALID_FILE_MSG);
      exit;
    end;
    pos1 := AnsiPos(DATE_STRING,fileData[iRecord]); // find date
  end;
  // Make sure no embedded double-quotes - they interfere with CSV parsing.
  FMeterID := AnsiReplaceStr(Trim(AnsiLeftStr(fileData[iRecord],pos1-1)),'"',QUOTE);
  pos2 := AnsiPos(TIME_STRING,fileData[iRecord]);  // find time

  if(pos2 > 0) then // found time
  begin
    myDate := Trim(AnsiMidStr(fileData[iRecord],
                      pos1+Length(DATE_STRING),
                      pos2-pos1-Length(DATE_STRING)));
    myTime := Trim(AnsiRightStr(fileData[iRecord],
                      Length(fileData[iRecord])-pos2-Length(TIME_STRING)));
  end else          // Missing time
  begin
    myDate := Trim(AnsiRightStr(fileData[iRecord],Length(fileData[iRecord])
                  - pos1 - Length(DATE_STRING)));
    myTime := '';
  end;

  case FChartType of
    ctSSI:     // Read English date/time
    begin
      InitDateTimeFormat(ftEnglish_US, myFormat);
      FChartDateTime := StrToDateTime(myDate + ' ' + myTime,myFormat);
    end;
    ctLDP,ctHarmonic:     // Read native date/time
    begin
      if(AnsiContainsStr(myDate,'/')) then  // English date convention
      begin
        InitDateTimeFormat(ftEnglish_US, myFormat);
      end else                             // International date convention
      begin
        InitDateTimeFormat(ftInternational, myFormat);
      end;
      myDate := myDate + ' ' + myTime;
      FChartDateTime := StrToDateTime(myDate, myFormat);
    end;
    else
    begin
      Assert(false,'Invalid chart type');
    end;
  end;
  Inc(iRecord);
  if(iRecord >= fileData.Count) then
  begin
    raise Exception.Create(DATA_RECORD_MSG);
  end;

  pos1 := AnsiPos(TIMESOURCE_STRING,fileData[iRecord]);  // find Time source
  if(pos1 > 0) then   // Found a time source
  begin
    FTimeSource := Trim(AnsiMidStr(fileData[iRecord],pos1 + 13,8));
    if(UpperCase(FTimeSource[1]) = 'I') then
    begin
      FTimeSource := 'Internal';
    end else
    begin
      FTimeSource := 'External';
    end;
    // Make sure no embedded double-quotes - they interfere with CSV parsing.
    FMeterTID := AnsiReplaceStr(Trim(AnsiLeftStr(fileData[iRecord],pos1-1)),'"',QUOTE);
  end else            // No time source, just use entire record as TID
  begin
    FTimeSource := 'Unknown';
    // Make sure no embedded double-quotes - they interfere with CSV parsing.
    FMeterTID := AnsiReplaceStr(Trim(AnsiLeftStr(fileData[iRecord],MAX_TID_STRING)),'"',QUOTE);
  end;

  Inc(iRecord,2);
  if(iRecord >= fileData.Count) then
  begin
    raise Exception.Create(DATA_RECORD_MSG);
  end;

  case FChartType of
    ctSSI:
      begin
        FChartName := 'VSSI';
        trimRecord := Trim(fileData[iRecord]);
        pos1 := AnsiPos(' ',trimRecord);
        if(pos1 > 0) then    // there is a separator
        begin
          FMeterFID := Trim(AnsiMidStr(trimRecord,SKIP_OVER_FID_LABEL,pos1));
        end else            // just use entire record
        begin
          FMeterFID := trimRecord;
        end;
        if(Length(FMeterFID) = 0) then
        begin
          FMeterFID := 'Unknown';
          vBasePos := VBASE_OFFSET_HMI_V100;
        end else
        begin
          if(IsAdvanced73x(FMeterFID)) then
          begin
             vBasePos := VBASE_OFFSET_HMI_V101;
          end else
          begin
             vBasePos := VBASE_OFFSET_HMI_V100;
          end;
        end;

        // Look ahead for INom values
        for i := 0 to VBASE_OFFSET do
        begin
          recordData.DelimitedText := fileData[iRecord+i];
          if((recordData.Count > NOMINAL_N_ITEM) and (recordData[0] = 'I')) then
          begin
            if(not TryStrToFloat(recordData[NOMINAL_ABCG_ITEM],nomABCG)) then
            begin
              nomABCG := 0.0;
            end;
            if(not TryStrToFloat(recordData[NOMINAL_N_ITEM],nomN)) then
            begin
              nomN := 0.0;
            end;

            settingRecord := TTimeSeriesSettingRecord.Create;
            FSettingTable.Add(settingRecord);
            settingRecord.FSettingName := NOMINAL_ABCG_STRING;
            settingRecord.FSettingValue := nomABCG;

            settingRecord := TTimeSeriesSettingRecord.Create;
            FSettingTable.Add(settingRecord);
            settingRecord.FSettingName := NOMINAL_N_STRING;
            settingRecord.FSettingValue := nomN;
            break;
          end;
        end;

        // Look ahead to get VBASE
        for i := 0 to VBASE_OFFSET do
        begin
          recordData.DelimitedText := fileData[iRecord+i];
          if((recordData.Count > 0) and (recordData[0] = '#')) then
          begin
            recordData.DelimitedText := fileData[iRecord+i+2]; // skip blank record
            break;
          end;
        end;
        settingRecord := TTimeSeriesSettingRecord.Create;
        FSettingTable.Add(settingRecord);
        settingRecord.FSettingName := VBASE_STRING;
        if(recordData.Count < 3) then
        begin
          raise Exception.Create(DATA_COLUMN_MSG);
        end;
        settingRecord.FSettingValue := StrToFloat(recordData[recordData.Count-vBasePos]);
      end;

    ctLDP:
      begin
        FChartName := 'LDP';
        pos1 := AnsiPos('Function:',fileData[iRecord]);
        if(pos1 > 0) then
        begin
          FMeterFID := Trim(AnsiLeftStr(filedata[iRecord],pos1-1));
          pos2 := AnsiPos('Recorder:',fileData[iRecord]);  // Test for LDP Recorder #
          if(pos2 > 0) then     // Found LDP recorder
          begin
            myString := Trim(AnsiMidStr(filedata[iRecord],pos1+9,4));
            if(TryStrToInt(myString,iValue)) then
            begin
              FLDPRecorder := iValue;
            end else
            begin
              FLDPRecorder := 0;
            end;
          end else
          begin
            FLDPRecorder := 0;
          end;
        end else
        begin
          if(AnsiPos('#',fileData[iRecord]) > 0) then  // missing FID record
          begin
            FMeterFID := 'Unknown';
            Dec(iRecord);  // record contains data column labels
          end else
          begin
            FMeterFID := Trim(filedata[iRecord]);
          end;
        end;
      end;

    ctHarmonic:
      begin
        FChartName := 'Harmonic';
        FLDPRecorder := 0;
        FMeterFID := 'Unknown';
        Dec(iRecord);

        // Look ahead to get VBASE
        for i := 0 to VBASE_OFFSET do
        begin
          recordData.DelimitedText := fileData[iRecord+i];
          if(recordData.Count > 0) then
          begin
            if((recordData[0] = 'Harmonic') or (recordData[0] = 'Frequency')) then
            begin
              exit;  // did not find harmonic heading
            end else
            if(recordData[0] = FUND_STRING) then
            begin
              settingRecord := TTimeSeriesSettingRecord.Create;
              FSettingTable.Add(settingRecord);
              settingRecord.FSettingName := FUND_STRING;
              if(recordData.Count < 4) then
              begin
                raise Exception.Create(DATA_COLUMN_MSG);
              end;
              settingRecord.FSettingValue := StrToFloat(recordData[3]);
              break;
            end;
          end;
        end;
      end
    else
    begin
      raise Exception.Create(CHART_TYPE_MSG);
    end;
  end;
end;
{************************************************************************
* NAME:       ReadFileCSVChannelData
* FUNCTIONAL DESCRIPTION:    Read data describing the channels
* INPUT PARAMETERS: iRecord : current record number
*                   fileData : TStringList   contains content of file
*                   recordData : TStringList  contains content of record
*                   eventField : string   name that indicates start of event data
* OUTPUT PARAMETERS: none
* RETURN: result : TTimeSeriesChartType  type of chart data found in file
************************************************************************}
procedure TTimeSeriesDataTable.ReadFileCSVChannelData(var iRecord : integer; fileData, recordData : TStringList; eventField : string);
const
  CHANNEL_FIELDS = 5;
  CHANNEL_NAME = 0;
  CHANNEL_FUNCTION = 1;
  CHANNEL_UNITS = 2;
  CHANNEL_RULES = 3;
  CHANNEL_MULTIPLIER = 4;
  CHANNEL_SCALE = 6;
  CHANNEL_MSG ='Insufficient fields in channel table record.';
var
  channelRecord : TTimeSeriesLDPChannelRecord;
begin
  Inc(iRecord);
  recordData.CommaText := fileData[iRecord];
  LDPChannelTable.Clear;

  // read records until the Event Table or the Chart Table is found

  while ((recordData[CHANNEL_NAME] <> eventField) and
         (recordData[CHANNEL_NAME] <> ChartField)) do
  begin
    if(recordData.Count < CHANNEL_FIELDS) then   // test for valid channel record
    begin
      raise Exception.Create(CHANNEL_MSG);
    end;
    channelRecord := TTimeSeriesLDPChannelRecord.Create;
    try
      LDPChannelTable.Add(channelRecord);
      channelRecord.ChannelName := recordData[CHANNEL_NAME];
      channelRecord.ChannelFunction := recordData[CHANNEL_FUNCTION];
      channelRecord.ChannelUnits := recordData[CHANNEL_UNITS];
      channelRecord.ChannelRules := recordData[CHANNEL_RULES];
      channelRecord.ChannelMult :=  StrToFloat(recordData[CHANNEL_MULTIPLIER]);
      if(recordData.Count > CHANNEL_SCALE) then
      begin
        channelRecord.ChannelScale :=  StrToFloat(recordData[CHANNEL_SCALE]);
      end else
      begin
        channelRecord.ChannelScale :=  1.0;
      end;
      Inc(iRecord);
      recordData.CommaText := fileData[iRecord];    // get next record
    finally
    end;
  end;
  // Decrement record index so that this record will be tested again
  // since the next function does not know which table was found
  Dec(iRecord);
end;
{************************************************************************
* NAME:       ReadFileCSVEventData
* FUNCTIONAL DESCRIPTION:    Read data describing events
* INPUT PARAMETERS: iRecord : current record number
*                   fileData : TStringList   contains content of file
*                   recordData : TStringList  contains content of record
* OUTPUT PARAMETERS: none
* RETURN: result : TTimeSeriesChartType  type of chart data found in file
************************************************************************}
procedure TTimeSeriesDataTable.ReadFileCSVEventData(var iRecord : integer; fileData, recordData : TStringList);
const
  EVENT_FIELDS = 4;
  EVENT_MSG ='Insufficient fields in event table record.';
var
  eventRecord : TTimeSeriesLDPEventRecord;
  myDateTime : string;
  myFormat : TFormatSettings;
begin

  InitDateTimeFormat(ftEnglish_US,myFormat);
  myFormat.ShortDateFormat := 'MM/DD/YY';   // must be standard English

  Inc(iRecord);
  recordData.CommaText := fileData[iRecord];
  LDPEventTable.Clear;
  while ((recordData[0] <> ChartField) and (iRecord < fileData.Count)) do
  begin
    if(recordData.Count < EVENT_FIELDS) then   // test for valid event record
    begin
      raise Exception.Create(EVENT_MSG);
    end;
    eventRecord := TTimeSeriesLDPEventRecord.Create;
    try
      LDPEventTable.Add(eventRecord);
      eventRecord.EventName := Format('#%.4d',[LDPEventTable.Count]);
      myDateTime := recordData[0] + ' ' + recordData[1];
      eventRecord.DateTime := StrToDateTime(myDateTime,myFormat);
      eventRecord.EventCode := recordData[2];
      eventRecord.Description := recordData[3];
      Inc(iRecord);
      if(iRecord < fileData.Count) then
      begin
        recordData.CommaText := fileData[iRecord];  // get next record
      end;
    finally
    end;
  end;
end;
{************************************************************************
* NAME:       InitDateTimeFormat
* FUNCTIONAL DESCRIPTION:    Initialize date/time format record
* INPUT PARAMETERS: formatType : TDateFormatType  (ftEnglish_US, ftInternational)
*                   myFormat : TFormatSettings
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************}
procedure TTimeSeriesDataTable.InitDateTimeFormat(formatType : TDateFormatType;
                             var myFormat : TFormatSettings);
begin
  InitTFormatSettings(formatType,FDateFormat,myFormat);
end;
{************************************************************************
* NAME:       LoadData
* FUNCTIONAL DESCRIPTION:  loads the data portion of the record from the
*                          given stream using the given record size
* INPUT PARAMETERS:
*                 fileData: TStringList           list of data records
*                 xColumns: array of integer      list of independent var data colums
*                                                 (record, date, time, phase, status)
*                 yColumns:  array of integer      list of dependent data columns
* OUTPUT PARAMETERS: none
* RETURN:
*                 result: boolean   true if data ok, false if data bad
************************************************************************}
function TTimeSeriesDataTable.LoadData(fileData : TStringList; xColumns, yColumns : array of integer) : boolean;
const
  DATA_OVERWRITE = 'Command Aborted, Data overwrite occurred.';
  DATA_INTERRUPT = 'No Voltage Sag/Swell/Interruption Data';
  DATA_INVALID = 'Invalid Data';
  DATA_FILE_INVALID = 'Invalid file content.';
  DATA_TRANSFER_ABORT = 'Transfer Abort';
  DATA_PROCESSING = 'Processing.';
  DATA_PROCESSING_ROW_LIMIT = 5;
  CURRENT_START = 0;
  CURRENT_END = 4;
  VOLT_START = 5;
  VOLT_END = 7;
  VSSI_VOLT_COUNT = 6;
  LDP_UNITS = 'Units:';
  MINIMUM_COLUMNS_SSI = 8;
  MINIMUM_COLUMNS_LDP = 4;
  MINIMUM_COLUMNS_HARMONIC = 2;
  VSSI_SKIP_CURRENT = 5;
  VSSI_VOLT_LIST_F9 : array[0..5] of string =
                   ('Vbase-a','Va','Vbase-b','Vb','Vbase-c','Vc');
  VSSI_VOLT_LIST_F5 : array[0..5] of string =
                   ('Vbase-ab','Vab','Vbase-bc','Vbc','Vbase-ca','Vca');
  SEL351RS_CURRENT_INDEX = 3;
  SEL351RS_VOLTAGE_INDEX = 4;
var
  recordData,labelData : TStringList;
  i,j, iRecord : Integer;
  chartFound,status,my73x,myAdvanced73x : boolean;
  columnCount, dataCount : integer;
  fileType : TTimeSeriesFileType;
  myStatus,myLabel : string;
  myFormat : TFormatSettings;
  myFormatEnglish : TFormatSettings;
  deviceField,deviceID,myRecord,myScale,myUnits : string;
  delim : System.char;
  pos1,pos2,pos3 : word;
  channelRecord : TTimeSeriesLDPChannelRecord;
begin

  Assert(fileData <> nil, DATA_FILE_INVALID);

  recordData := TStringList.Create;
  labelData := TStringList.Create;
  deviceID := GetHMIDevice(MeterFID);
  my73x := isDevice73x(MeterFID);
  myAdvanced73x := isAdvanced73x(MeterFID);

  try
    deviceField := 'DeviceID';  // column name of device information table
    fileType := GetFileType(fileData,delim);
    recordData.Delimiter := delim;

    chartFound := false;
    columnCount := 0;
    dataCount := 0;
    FDeviceType := dtSEL7XX;   // in case it is a relay SSI report
    
    if(fileType = ftCSV) then
    begin
      InitDateTimeFormat(ftEnglish_US,myFormat);
      myFormat.ShortDateFormat := 'MM/DD/YY';   // must be standard English
    end else
    begin
      if(ChartType = ctSSI) then
      begin
        InitDateTimeFormat(ftEnglish_US,myFormat);
      end else
      begin
        InitDateTimeFormat(ftInternational,myFormat);
      end;
    end;

    InitDateTimeFormat(ftEnglish_US,myFormatEnglish);
    if(fileType = ftCSV) then
    begin
      myFormatEnglish.ShortDateFormat := 'mm/dd/yy';   // must be standard English
    end;

    // extract data from each data record
    iRecord := 0;
    while iRecord < fileData.Count  do
    begin
      // test for error messages in record
      if(AnsiContainsText(fileData[iRecord],DATA_OVERWRITE)) then
      begin
        Inc(iRecord);
        continue; // skip record
      end;
      if(AnsiContainsText(fileData[iRecord],DATA_INVALID)) then
      begin
        Inc(iRecord);
        continue; // skip record
      end;
      if(AnsiContainsText(fileData[iRecord],DATA_PROCESSING)) then
      begin
        Inc(iRecord);
        if(iRecord > DATA_PROCESSING_ROW_LIMIT) then
        begin
           LoadVSSISummaryData(fileData, iRecord);
        end else
        begin
          continue; // skip record
        end;
      end;
      if(AnsiContainsText(fileData[iRecord],DATA_INTERRUPT)) then
      begin
        result := false;
        exit;
      end;
      if(AnsiContainsText(fileData[iRecord],DATA_TRANSFER_ABORT)) then
      begin
        result := false;
        exit;
      end;

      if(AnsiContainsStr(fileData[iRecord],'*')) then
      begin
        myStatus := '*';
        fileData[iRecord] := AnsiReplaceStr(fileData[iRecord],'*',' ');  // replace "trigger" flag
      end else
      begin
        myStatus := ' ';
      end;

      if(fileType = ftReport) then
      begin
        recordData.DelimitedText := fileData[iRecord];  // Report
      end else
      begin
        recordData.CommaText := fileData[iRecord];   // CSV
      end;

      if(recordData.Count > 0) then   // something is in the record
      begin
        if(not chartFound and
          (AnsiCompareText(recordData[0],ChartField) = 0)) then
        begin
          if(((ChartType = ctSSI) and (recordData.Count < MINIMUM_COLUMNS_SSI)) or
             ((ChartType = ctLDP) and (recordData.Count < MINIMUM_COLUMNS_LDP)) or // invalid column count
             ((ChartType = ctHarmonic) and (recordData.Count < MINIMUM_COLUMNS_HARMONIC))) then // invalid column count
          begin
            result := false;
            exit;
          end;
          // 73x Meter and 3xx/4xx relays have different number of columns in SSI
          if((ChartType = ctSSI) and (not my73x) and
             ((recordData.Count-COL_TIME-1) < yColumns[Length(yColumns)-1])) then
          begin
            if(recordData.Count <= SEL351RS_COUNT) then     // SEL351RS
            begin
              yColumns[0] := SEL351RS_CURRENT_INDEX;
              yColumns[1] := SEL351RS_VOLTAGE_INDEX;
              for i := 2 to Length(yColumns)-1 do
              begin
                yColumns[i] := -1;  // deactivate non-current/volt columns
              end;
            end else                                          // SEL4XX
            begin
              for i := CURRENT_START to CURRENT_END do
              begin
                yColumns[i] := -1;  // deactivate current columns
              end;
              for i := VOLT_START to VOLT_END do
              begin
                Dec(yColumns[i],VOLT_START);  // decrement voltage column indices
              end;
            end;
          end;
          if((fileType = ftReport) and (ChartType = ctHarmonic) and (Length(fileData[iRecord-1]) > 1)) then
          begin
            labelData.DelimitedText := fileData[iRecord-1];
            for j := Low(yColumns) to High(yColumns) do
            begin
              if(ycolumns[j] < 0) then continue;  // inactive column, skip it
              myLabel := labelData[(yColumns[j]-1) div 2] + '-' + recordData[yColumns[j]];
              RawDataLabels.Add(myLabel)
            end;
          end else
          begin
            for j := Low(yColumns) to High(yColumns) do
            begin
              if(ycolumns[j] < 0) then continue;  // inactive column, skip it
              RawDataLabels.Add(recordData[yColumns[j]])
            end;
            if((ChartType = ctSSI) and (fileType = ftReport) and
               myAdvanced73x)  then  // new VSSI report format
            begin
              recordData.DelimitedText := fileData[iRecord-3];  // Phase record
              if(AnsiContainsText(recordData[0],'Ph-AB')) then  // form 5
              begin
                for j := 0 to VSSI_VOLT_COUNT - 1 do
                begin
                  RawDataLabels[j+VSSI_SKIP_CURRENT] := VSSI_VOLT_LIST_F5[j];
                end;
              end else                                         // form 9
              begin
                for j := 0 to VSSI_VOLT_COUNT - 1 do
                begin
                  RawDataLabels[j+VSSI_SKIP_CURRENT] := VSSI_VOLT_LIST_F9[j];
                end;
              end;
            end;
          end;
          if((ChartType = ctLDP) and (fileType = ftReport)) then  // init channel table
          begin
            LDPChannelTable.Clear;
            for j := 0 to RawDataLabels.Count - 1 do
            begin
              channelRecord := TTimeSeriesLDPChannelRecord.Create;
              try
                LDPChannelTable.Add(channelRecord);
                channelRecord.ChannelName := RawDataLabels[j];
                channelRecord.ChannelFunction :='Unknown';
                channelRecord.ChannelUnits := 'Unknown';
                channelRecord.ChannelRules := Format('%.2f',[1.0]);  // regional language
                channelRecord.ChannelMult :=  1.0;
                channelRecord.ChannelScale :=  1.0;
              finally
              end;
            end;
          end;
          chartFound := true;      // found header, remaining records are data
          // determine number of data columns
          dataCount := 0;
          for j := Low(yColumns) to High(yColumns) do
          begin
            if(yColumns[j] > 0) then
            begin
              Inc(dataCount);
            end;
          end;
        end else
        begin
          if(chartFound) then      // we are reading data records from now on
          begin
            // set column count based on first data record
            if(columnCount = 0) then
            begin
              columnCount := recordData.Count;  // remember column count
              FDeviceType := DetermineDeviceType(fileType,columnCount,xColumns[High(xColumns)]);
            end;
            try
              status := ReadChartRecord(fileType,recordData,myFormat,myFormatEnglish,
                                xColumns,yColumns,columnCount,dataCount,myStatus);
            except
              status := false;   // exception means bad record format
            end;
            if(not status) then
            begin
              Inc(iRecord);
              continue;   // bad record, skip and keep reading
            end;
          end;
        end;
      end else
      if(chartFound and (RawDataTable.Count > 0)) then  // encountered blank record in data, stop now
      begin
        if((iRecord < fileData.Count) and
           (fileType = ftReport)) then
        begin
          if(ChartType = ctLDP) then // test for LDP units/Scale
          begin
            LDPUnits := utSecondary;   // default for LDP in case Units/Scale not found
            LDPUnitsBase := utSecondary;   // default for LDP in case Units/Scale not found
            LDPScale := stUnity;
            if(fileData.Count > (iRecord + 1)) then
            begin
              myRecord := fileData[iRecord+1];  // Get Units record if present
              pos1 := AnsiPos(LDP_UNITS,myRecord);
              if(pos1 > 0) then
              begin
                pos1 := AnsiPos(':',myRecord);     // Start of units
                pos2 := AnsiPos('(',myRecord);     // End units, start scale
                pos3 := AnsiPos(')',myRecord);     // End scale
                myUnits := Trim(AnsiMidStr(myRecord,pos1+1,pos2-pos1-1));
                myScale := Trim(AnsiMidStr(myRecord,pos2+1,pos3-pos2-1));
                LDPUnits := TTimeSeriesUnitsType(StringInArray(myUnits,LDP_UNITS_LIST)-1);
                LDPScale := TTimeSeriesScaleType(StringInArray(myScale,LDP_SCALE_LIST)-1);
              end;
            end;
          end else
          if(ChartType = ctSSI) then
          begin
            LoadVSSISummaryData(fileData, iRecord);
          end;
        end;
        result := true;
        exit;
      end;
      Inc(iRecord);
    end;

  finally
    FreeAndNil(recordData);
  end;
  result := chartFound;
end;

{************************************************************************
* NAME:    LoadVSSISummaryData
* FUNCTIONAL DESCRIPTION: load VSSI summary data from report file into index object
* INPUT PARAMETERS:       fileData : TStringList input records
*                         iRecord : integer record number
* OUTPUT PARAMETERS:      none
* RETURN:                 result : boolean   true if data parsed, otherwise false
************************************************************************}
procedure TTimeSeriesDataTable.LoadVSSISummaryData(fileData : TStringList; iRecord : integer);
const
  COL_EVENT_NUMBER = 0;
  COL_EVENT_TYPE = 1;
  COL_EVENT_DATE = 2;
  COL_EVENT_TIME = 3;
  COL_EVENT_DURATION = 4;
  COL_EVENT_DEPTH = 5;
  COL_VBASE_A = 6;
  COL_VA_MIN = 7;
  COL_VA_MAX = 8;
  COL_VBASE_B = 9;
  COL_VB_MIN = 10;
  COL_VB_MAX = 11;
  COL_VBASE_C = 12;
  COL_VC_MIN = 13;
  COL_VC_MAX = 14;
  COL_ITIC_REGION = 15;

  EVENT_SAG = 'SAG';
  EVENT_TRIG = 'TRIG';
  EVENT_INT = 'INT';

var
  more : boolean;
  recordData : TStringList;
  indexRecord : TTimeSeriesIndexRecord;
  eventDateTime : string;
  eventNumber : integer;
  foundEvent : boolean;
  eventType : string;
  myFormat : TFormatSettings;
begin

  { Parse the event summary from the input report found in fileData }

  InitDateTimeFormat(ftEnglish_US,myFormat);
  // device generated report must be standard English
  myFormat.ShortDateFormat := 'MM/DD/YY';
  IndexTable.Clear();

  more := true;
  foundEvent := false;
  eventNumber := 1;

  recordData := TStringList.Create();
  try
    while (more) do
    begin
      // Stop if end of file reached
      if(iRecord < fileData.Count) then
      begin
        recordData.DelimitedText := fileData[iRecord];
        if((recordData.Count = 0) or
           (not foundEvent and not AnsiContainsStr(recordData[COL_EVENT_NUMBER],'#'))) then
        begin
          // Skip empty and non-event header records
          Inc(iRecord);
          continue;
        end;

        // check for event header record
        if(not foundEvent and AnsiContainsStr(recordData[COL_EVENT_NUMBER],'#')) then
        begin
          Inc(iRecord);
          foundEvent := true;
          continue;
        end;

        if((recordData.Count <= COL_ITIC_REGION) or (recordData[COL_EVENT_TYPE] = '***')) then
        begin
          // Skip invalid event record
          Inc(iRecord);
          continue;
        end;

        // This is an event record, so parse it into the index
        indexRecord := TTimeSeriesIndexRecord.Create();
        try
          IndexTable.Add(indexRecord);

          indexRecord.EventName := Format('#%.4d',[eventNumber]);
          indexRecord.Description := recordData[COL_EVENT_TYPE];
          eventType := recordData[COL_EVENT_TYPE];
          if(eventType = EVENT_SAG)  then
          begin
            indexRecord.EventType := evSag;
          end else
          if(eventType = EVENT_TRIG) then
          begin
            indexRecord.EventType := evNone;
          end else
          if(eventType = EVENT_INT) then
          begin
            indexRecord.EventType := evInterrupt;
          end else
          begin
            indexRecord.EventType := evSwell;
          end;

          eventDateTime := recordData[COL_EVENT_DATE] + ' ' + recordData[COL_EVENT_TIME];
          try
            indexRecord.DateTime := StrToDateTime(eventDateTime,myFormat);
          except
            indexRecord.DateTime := Now;
          end;

          eventDateTime := recordData[COL_EVENT_DURATION];
          try
            indexRecord.Duration := ConvertTimeToDuration(eventDateTime,myFormat);
          except
            indexRecord.Duration := 0;
          end;

          indexRecord.Magnitude := StrToFloatDef(recordData[COL_EVENT_DEPTH],0.0);
          indexRecord.VA_Vbase := StrToFloatDef(recordData[COL_VBASE_A],0.0);
          indexRecord.VAMin := StrToFloatDef(recordData[COL_VA_MIN],0.0);
          indexRecord.VAMax := NonLocaleStrToFloatDef(recordData[COL_VA_MAX],0.0);
          indexRecord.VB_Vbase := NonLocaleStrToFloatDef(recordData[COL_VBASE_B],0.0);
          indexRecord.VBMin := NonLocaleStrToFloatDef(recordData[COL_VB_MIN],0.0);
          indexRecord.VBMax := NonLocaleStrToFloatDef(recordData[COL_VB_MAX],0.0);
          indexRecord.VC_Vbase := NonLocaleStrToFloatDef(recordData[COL_VBASE_C],0.0);
          indexRecord.VCMin := NonLocaleStrToFloatDef(recordData[COL_VC_MIN],0.0);
          indexRecord.VCMax := NonLocaleStrToFloatDef(recordData[COL_VC_MAX],0.0);

          indexRecord.Violation := CBEMAViolation(indexRecord);

          // Find the record interval of each event
          FindEventRecords(indexRecord);
          Inc(iRecord);
          Inc(eventNumber);
        finally
        end;
      end else
      begin
        more := false;
      end;
    end;
  finally
    recordData.Free();
  end;
end;

function TTimeSeriesDataTable.TryNonLocaleStrToFloat(svalue : String; var value : double) : boolean;
begin
  result := TryStrToFloat(svalue, value);
end;

function TTimeSeriesDataTable.NonLocaleStrToFloatDef(value : String; defaultValue : single) : single;
begin
  result := StrToFloatDef(value,defaultValue);
end;
{************************************************************************
* NAME:    ConvertTimeToDuration
* FUNCTIONAL DESCRIPTION: convert duration string to TDateTime
* INPUT PARAMETERS:       timeString : string containing time
*                         aFormat : TFormatSettings - format used for string conversion
* OUTPUT PARAMETERS:      none
* RETURN:                 TDateTime : duration
************************************************************************}
function TTimeSeriesDataTable.ConvertTimeToDuration(timeString : String; aFormat : TFormatSettings) : TDateTime;
var
  index : integer;
  hours : String;
  duration : TDateTime;
  hourCount : integer;
begin
// The standard DateTimeToStr() method does not support durations > 24 hours
// so this method addresses that limitation
  index := Pos(':', timeString);
  hours := LeftStr(timeString,index-1);
  if(not TryStrToInt(hours,hourCount)) then
  begin
     hourCount := 0;
  end;
  // replace hours with 0 then increment the duration by the number of hours
  timeString := '0' + RightStr(timeString,Length(timeString) - index + 1);
  duration := StrToDateTime(timeString);
  duration := IncHour(duration, hourCount);
  result := duration;
end;
{************************************************************************
* NAME:    ReadFileCSVVSSIEventData
* FUNCTIONAL DESCRIPTION: read VSSI event data from CSV file into index
* INPUT PARAMETERS:       iRecord : integer record number
*                         fileData : TStringList input records
* OUTPUT PARAMETERS:      iRecord : updated record number returned
* RETURN:                 none
************************************************************************}
procedure TTimeSeriesDataTable.ReadFileCSVVSSIEventData(var iRecord : integer; fileData : TStringList);
const
  COL_EVENT_NUMBER = 0;
  COL_EVENT_DATE = 1;
  COL_EVENT_TIME = 2;
  COL_EVENT_DURATION = 3;
  COL_EVENT_DESCRIPTION = 4;
  COL_EVENT_DEPTH = 5;
  COL_VIOLATION = 6;
  COL_FIRST = 7;
  COL_LAST = 8;
  COL_START = 9;
  COL_STOP = 10;
  COL_VBASE_A = 11;
  COL_VA_MIN = 12;
  COL_VA_MAX = 13;
  COL_VBASE_B = 14;
  COL_VB_MIN = 15;
  COL_VB_MAX = 16;
  COL_VBASE_C = 17;
  COL_VC_MIN = 18;
  COL_VC_MAX = 19;

  EVENT_SAG = 'SAG';
  EVENT_TRIG = 'TRIG';
  EVENT_INT = 'INT';

var
  more : boolean;
  recordData : TStringList;
  indexRecord : TTimeSeriesIndexRecord;
  eventDateTime : string;
  eventNumber : integer;
  eventType : string;
  myFormat : TFormatSettings;
begin

  InitDateTimeFormat(ftEnglish_US,myFormat);
  // CSV file must be standard English
  myFormat.ShortDateFormat := 'MM/DD/YY';
  IndexTable.Clear();

  more := true;
  eventNumber := 1;
  Inc(iRecord);

  // If event record is truncated (legacy file) or has invalid data,
  // set flag so that the event index will be generated from the raw VSSI records
  // later.  By default, assume that the event records will be generated later.
  HasEventSummary := false;

  recordData := TStringList.Create();
  try
    while (more) do
    begin
      // Stop if end of file reached
      if(iRecord < fileData.Count) then
      begin
        recordData.DelimitedText := fileData[iRecord];
        // Event records must start with a # sign, if not stop
        if((recordData.Count <= COL_VC_MAX) or
           not AnsiContainsStr(recordData[COL_EVENT_NUMBER],'#')) then
        begin
          break;
        end;

        indexRecord := TTimeSeriesIndexRecord.Create();
        try
          IndexTable.Add(indexRecord);

          // parse event data into event index table
          indexRecord.EventName := Format('#%.4d',[eventNumber]);
          indexRecord.Description := recordData[COL_EVENT_DESCRIPTION];
          eventType := indexRecord.Description;
          if(eventType = EVENT_SAG)  then
          begin
            indexRecord.EventType := evSag;
          end else
          if(eventType = EVENT_TRIG) then
          begin
            indexRecord.EventType := evNone;
          end else
          if(eventType = EVENT_INT) then
          begin
            indexRecord.EventType := evInterrupt;
          end else
          begin
            indexRecord.EventType := evSwell;
          end;

          eventDateTime := recordData[COL_EVENT_DATE] + ' ' + recordData[COL_EVENT_TIME];
          try
            indexRecord.DateTime := StrToDateTime(eventDateTime,myFormat);
          except
            indexRecord.DateTime := Now;
          end;

          eventDateTime := recordData[COL_EVENT_DURATION];
          try
            indexRecord.Duration := ConvertTimeToDuration(eventDateTime,myFormat);
          except
            indexRecord.Duration := 0;
          end;

          if(not TryNonLocaleStrToFloat(recordData[COL_EVENT_DEPTH],indexRecord.Magnitude)) then
          begin
            indexRecord.Magnitude := 0.0;
          end;

          if(not TryStrToInt(recordData[COL_FIRST],indexRecord.FirstRecord)) then
          begin
            indexRecord.FirstRecord := 0;
          end;
          if(not TryStrToInt(recordData[COL_LAST],indexRecord.LastRecord)) then
          begin
            indexRecord.LastRecord := 0;
          end;
          if(not TryStrToInt(recordData[COL_START],indexRecord.StartEvent)) then
          begin
            indexRecord.StartEvent := 0;
          end;
          if(not TryStrToInt(recordData[COL_STOP],indexRecord.StopEvent)) then
          begin
            indexRecord.StopEvent := 0;
          end;

          // This data was added to the record with z105
          if(recordData.Count > COL_VBASE_A) then
          begin
            HasEventSummary := true;
            if(not TryNonLocaleStrToFloat(recordData[COL_VBASE_A],indexRecord.VA_Vbase)) then
            begin
              indexRecord.VA_Vbase := 0.0;
              HasEventSummary := false;
            end;
            if(not TryNonLocaleStrToFloat(recordData[COL_VA_MIN],indexRecord.VAMin)) then
            begin
              indexRecord.VAMin := 0.0;
              HasEventSummary := false;
            end;
            if(not TryNonLocaleStrToFloat(recordData[COL_VA_MAX],indexRecord.VAMax)) then
            begin
              indexRecord.VAMax := 0.0;
              HasEventSummary := false;
            end;
            if(not TryNonLocaleStrToFloat(recordData[COL_VBASE_B],indexRecord.VB_Vbase)) then
            begin
              indexRecord.VB_Vbase := 0.0;
              HasEventSummary := false;
            end;
            if(not TryNonLocaleStrToFloat(recordData[COL_VB_MIN],indexRecord.VBMin)) then
            begin
              indexRecord.VBMin := 0.0;
              HasEventSummary := false;
            end;
            if(not TryNonLocaleStrToFloat(recordData[COL_VB_MAX],indexRecord.VBMax)) then
            begin
              indexRecord.VBMax := 0.0;
              HasEventSummary := false;
            end;
            if(not TryNonLocaleStrToFloat(recordData[COL_VBASE_C],indexRecord.VC_Vbase)) then
            begin
              indexRecord.VC_Vbase := 0.0;
              HasEventSummary := false;
            end;
            if(not TryNonLocaleStrToFloat(recordData[COL_VC_MIN],indexRecord.VCMin)) then
            begin
              indexRecord.VCMin := 0.0;
              HasEventSummary := false;
            end;
            if(not TryNonLocaleStrToFloat(recordData[COL_VC_MAX],indexRecord.VCMax)) then
            begin
              indexRecord.VCMax := 0.0;
              HasEventSummary := false;
            end;
          end;

          indexRecord.Violation := CBEMAViolation(indexRecord);

          Inc(iRecord);
          Inc(eventNumber);
        finally
        end;
      end else
      begin
        more := false;
      end;
    end;
  finally
    recordData.Free();
  end;
end;

{************************************************************************
* NAME: FindEventRecords
* FUNCTIONAL DESCRIPTION: find record indices associated with time interval of this event
* INPUT PARAMETERS: indexRecord : TTimeSeriesIndexRecord
* OUTPUT PARAMETERS: none
************************************************************************}
procedure TTimeSeriesDataTable.FindEventRecords( indexRecord : TTimeSeriesIndexRecord);
var
  dataRecord : TTimeSeriesRecord;
  recordDate, endDateTime : TDateTime;
  foundStart : boolean;
  i : integer;
begin

  // Assume event spans entire record set by default
  indexRecord.StartEvent := 0;
  indexRecord.StopEvent := RawDataTable.Count - 1;
  foundStart := false;
  endDateTime := Now;

  for i := 0 to RawDataTable.Count - 1 do
  begin
    dataRecord := (RawDataTable[i] as TTimeSeriesRecord);   // get indexed record from list
    recordDate := dataRecord.DateTime;

    if(not foundStart and (indexRecord.DateTime <= recordDate)) then
    begin
      // found matching raw VSSI record to match start time
      indexRecord.StartEvent := i;
      endDateTime := indexRecord.DateTime + indexRecord.Duration;
      foundStart := true;
    end;

    if(foundStart and (recordDate >= endDateTime)) then
    begin
      // found matching raw VSSI record to match start + duration (stop time)
      indexRecord.StopEvent := i;
      break;
    end;
  end;

  // assume pre-disturbance and post-disturbance coincide with event start/stop
  indexRecord.FirstRecord := indexRecord.StartEvent;
  indexRecord.LastRecord := indexRecord.StopEvent;

  // find pre- and post-disturbance intervals if any
  for i := indexRecord.StartEvent - 1 downto 0 do
  begin
    dataRecord := (RawDataTable[i] as TTimeSeriesRecord);   // get indexed record from list
    if(dataRecord.Status = 'P') then
    begin
      // found start of pre-disturbance
      indexRecord.FirstRecord := i;
    end else
    begin
      break;
    end;
  end;

  for i := indexRecord.StopEvent + 1 to RawDataTable.Count - 1 do
  begin
    dataRecord := (RawDataTable[i] as TTimeSeriesRecord);   // get indexed record from list
    if(dataRecord.Status = 'E') then
    begin
      // found end of post-disturbance
      indexRecord.LastRecord := i;
    end else
    begin
      break;
    end;
  end;

end;

{************************************************************************
* NAME: CBEMAViolation
* FUNCTIONAL DESCRIPTION: return CBEMA violation code based on event duration/magnitude
* INPUT PARAMETERS: indexRecord : TTimeSeriesIndexRecord
* OUTPUT PARAMETERS: none
* RETURN: result : TTimeSeriesCBEMAViolationType
************************************************************************}
function TTimeSeriesDataTable.CBEMAViolation( indexRecord : TTimeSeriesIndexRecord) : TTimeSeriesCBEMAViolationType;
const
  DURATION70 = 20;     // 20ms
  DURATION80 = 500;    // 0.5s
  DURATION90 = 10000;   // 10s
  DURATION110 = 500;   //  0.5s
  DURATION120 = 3;     // 3ms
  DURATION200 = 1;      // 1ms
  MILLISEC_PER_SEC = 1000;
  SECONDS_PER_HOUR = 3600;
  SECONDS_PER_MINUTE = 60;
var
  duration : integer;
  violation : TTimeSeriesCBEMAViolationType;
  hour, min, sec, mSec : word;
begin
  DecodeTime(indexRecord.Duration,hour,min,sec,mSec);
  duration := (hour * SECONDS_PER_HOUR + min * SECONDS_PER_MINUTE + sec) *
               MILLISEC_PER_SEC + mSec;

  if((indexRecord.Magnitude < 70.0) and (duration > DURATION70)) then
  begin
    violation := cvt70;
  end else
  if((indexRecord.Magnitude < 80.0) and (duration > DURATION80)) then
  begin
    violation := cvt80;
  end else
  if((indexRecord.Magnitude < 90.0) and (duration > DURATION90)) then
  begin
    violation := cvt90;
  end else
  if((indexRecord.Magnitude > 200.0) and (duration > DURATION200)) then
  begin
    violation := cvt200;
  end else
  if((indexRecord.Magnitude > 120.0) and (duration > DURATION120)) then
  begin
    Violation := cvt120;
  end else
  if((indexRecord.Magnitude > 110.0) and (duration > DURATION110)) then
  begin
    violation := cvt110;
  end else
  begin
    violation := cvtnone;
  end;
  result := violation;
end;

{************************************************************************
* NAME:       ReadChartRecord
* FUNCTIONAL DESCRIPTION:    Read and parse a chart record and add to collection
* INPUT PARAMETERS:  fileType : TTimeSeriesFileType   type of file
*                    recordData : TStringList  content of present record
*                    myFormat : TFormatSettings  used to parse date/time strings
*                    myFormatEnglish : TFormatSettings  used to parse date/time strings
*                    standard : array of integer  column position of standard values
*                    columns : array of integer column position of data columns
*                    columnCount : expected number of columns
*                    dataCount : number of active data columns
* OUTPUT PARAMETERS:
*                    deviceType : TTimeSeriesDeviceType (dtSEL3xx,...)
*                    myStatus : string  status field
* RETURN:            result : boolean  record is valid (true/false)
************************************************************************}
function TTimeSeriesDataTable.ReadChartRecord(
                   fileType : TTimeSeriesFileType;
                   recordData : TStringList;
                   myFormat : TFormatSettings;
                   myFormatEnglish : TFormatSettings;
                   xColumns, yColumns : array of integer;
                   columnCount : integer;
                   dataCount : integer;
                   var myStatus : string) : boolean;
const
  FIELD_COUNT_MSG = 'Insufficient fields in record.';
  INVALID_RECORD_MSG = 'Invalid record number.';
  INVALID_DATE_MSG = 'Invalid date.';
  INVALID_PHASE_MSG = 'Invalid phase.';
  INVALID_STATUS_MSG = 'Invalid status.';
  INVALID_DATA_MSG = 'Invalid channel data.';
var
  dataRecord : TTimeSeriesRecord;
  valid : boolean;
  status : boolean;
begin
  status := true;  // Assume the record is valid
  if(recordData.Count < 1) then    // make sure something in record
  begin
    raise Exception.Create(FIELD_COUNT_MSG);
  end;

  if(recordData[0] = '#') then   // found comment in data, stop scanning
  begin
    result := false;
    exit;
  end else
  begin
    if((columnCount <> recordData.Count) or       // inconsistent column count
       (recordData.Count < yColumns[High(yColumns)])) then  // not enough data columns
    begin
      result := false;
      exit;
    end;
    dataRecord := TTimeSeriesRecord.Create;

    try
      // Get Record number
      try
        valid := ReadChartRecordNumber(recordData, xColumns, dataRecord);
        if(not valid) then
        begin
          status := false;
        end;
      except
        status := false;
      end;

      // Get Date
      try
        valid := ReadChartRecordDate(recordData, xColumns, fileType,
                                     myFormat, myFormatEnglish, dataRecord);
        if(not valid) then
        begin
          status := false;
        end;
      except
        status := false;
      end;

      // Get Phase
      try
        valid := ReadChartRecordPhase(recordData, xColumns, dataRecord);
        if(not valid) then
        begin
          status := false;
        end;
      except
        dataRecord.Phase := '...';
        status := false;
      end;

      // Get Status
      try
        valid := ReadChartRecordStatus(recordData, xColumns, dataRecord);
        if(not valid) then
        begin
          status := false;
        end;
        myStatus := dataRecord.Status;     // return the status that was found
      except
        dataRecord.Status := 'R';
        status := false;
      end;

      // Get Data
      try
        valid := ReadChartRecordData(recordData, dataCount, yColumns, fileType, dataRecord);
        if(not valid or (myStatus = 'X')) then  // X is invalid record marker
        begin
          status := false;
        end;
      except
        status := false;
      end;
    except    // dataRecord was already added to the RawDataTable
      status := false;
    end;
  end;

  if(status = true) then     // record is valid
  begin
    RawDataTable.Add(dataRecord);
  end else
  begin
    dataRecord.Free;
  end;

  result := status;
end;
{************************************************************************
* NAME:       ReadChartRecordNumber
* FUNCTIONAL DESCRIPTION:    Read the RECORD Number column from the data record
* INPUT PARAMETERS:
*                    recordData : TStringList  content of present record
*                    xColumns : array of integer  column position of standard values
*                    fileType : TTimeSeriesFileType   ftCSV, ftReport...
* OUTPUT PARAMETERS:
*                    dataRecord : TTimeSeriesRecord   record of data found
* RETURN:            result : boolean  record is valid (true/false)
************************************************************************}
function TTimeSeriesDataTable.ReadChartRecordNumber(
                   recordData : TStringList;
                   xColumns : array of integer;
                   dataRecord : TTimeSeriesRecord) : boolean;
begin
  if(xColumns[COL_RECORD] >= 0) then    // extracting actual record number
  begin
    if(ChartSubType = cstHRM_INT) then
    begin
      if(not TryNonLocaleStrToFloat(recordData[xColumns[COL_RECORD]],dataRecord.RecordValue)) then
      begin
        result := false;
        exit;
      end else
      begin
        dataRecord.RecordNumber := Round(dataRecord.RecordValue);
      end;
    end else
    begin
      if(not TryStrToInt(recordData[xColumns[COL_RECORD]], dataRecord.RecordNumber)) then
      begin
        result := false;
        exit;
      end;
      dataRecord.RecordValue := dataRecord.RecordNumber;  // float representation
    end;
  end else                     // construct record number automatically
  begin
    dataRecord.RecordNumber := RawDataTable.Count;
  end;
  result := true;
end;
{************************************************************************
* NAME:       ReadChartRecordDate
* FUNCTIONAL DESCRIPTION:    Read the DATE column from the data record
* INPUT PARAMETERS:
*                    recordData : TStringList  content of present record
*                    xColumns : array of integer  column position of standard values
*                    fileType : TTimeSeriesFileType   ftCSV, ftReport...
*                    myFormat : TFormatSettings
*                    myFormatEnglish : TFormatSettings
* OUTPUT PARAMETERS:
*                    dataRecord : TTimeSeriesRecord   record of data found
* RETURN:            result : boolean  record is valid (true/false)
************************************************************************}
function TTimeSeriesDataTable.ReadChartRecordDate(
                   recordData : TStringList;
                   xColumns : array of integer;
                   fileType : TTimeSeriesFileType;
                   myFormat : TFormatSettings;
                   myFormatEnglish : TFormatSettings;
                   dataRecord : TTimeSeriesRecord) : boolean;
var
  myDateTime : string;
  status : boolean;
begin
  if(Length(xColumns) < COL_DATE) then  // invalid column index array
  begin
    result := false;
    exit;
  end;

  status := true;  // assume date is valid
  if(xColumns[COL_DATE] >= 0) then   // extract and record date/time
  begin
    try
      myDateTime := recordData[xColumns[COL_DATE]] + ' ' + recordData[xColumns[COL_TIME]];
      if(AnsiContainsStr(myDateTime,'/')) then  // English date
      begin
        dataRecord.DateTime := StrToDateTime(myDateTime,myFormatEnglish);
      end else
      begin                                     // International date
        dataRecord.DateTime := StrToDateTime(myDateTime,myFormat);
      end;
    except                 // probably bad date format
      status := false;
    end;
  end else
  begin                       // use current date/time
    dataRecord.DateTime := Now;
  end;
  result := status;
end;

{************************************************************************
* NAME:       ReadChartRecordPhase
* FUNCTIONAL DESCRIPTION:    Read the PHASE column from the data record
* INPUT PARAMETERS:
*                    recordData : TStringList  content of present record
*                    xColumns : array of integer  column position of standard values
* OUTPUT PARAMETERS:
*                    dataRecord : TTimeSeriesRecord   record of data found
* RETURN:            result : boolean  record is valid (true/false)
************************************************************************}
function TTimeSeriesDataTable.ReadChartRecordPhase(
                   recordData : TStringList;
                   xColumns : array of integer;
                   dataRecord : TTimeSeriesRecord) : boolean;
const
  PHASE_A = 1;
  PHASE_C = 3;
var
  j,iColumn,maxPhase : integer;
begin
  maxPhase := PHASE_C;
  if(xColumns[COL_PHASE] >= 0) then    // extract Phase descriptor
  begin
    iColumn := xColumns[COL_PHASE];    // usual column for PHASE
    case FDeviceType of
      dtSEL3XX:
        begin
          Inc(iColumn,RELAY_3XX_OFFSET);
        end;
      dtSEL351RS:
        begin
          Inc(iColumn,RELAY_351RS_OFFSET);
          maxPhase := PHASE_A;
        end;
      dtSEL4XX:
        begin
          Inc(iColumn,RELAY_4XX_OFFSET);
        end;
    end;
    dataRecord.Phase := recordData[iColumn];

    for j := PHASE_A to maxPhase do
    begin
      if(not CharInSet(dataRecord.Phase[j], ['I','O','U','.'])) then
      begin
        result := false;
        exit;
      end;
    end;
  end else
  begin
    dataRecord.Phase := '';    // empty string
  end;
  result := true;
end;
{************************************************************************
* NAME:       ReadChartRecordStatus
* FUNCTIONAL DESCRIPTION:    Read the STATUS column from the data record
* INPUT PARAMETERS:
*                    recordData : TStringList  content of present record
*                    xColumns : array of integer  column position of standard values
* OUTPUT PARAMETERS:
*                    dataRecord : TTimeSeriesRecord   record of data found
* RETURN:            result : boolean  record is valid (true/false)
************************************************************************}
function TTimeSeriesDataTable.ReadChartRecordStatus(
                   recordData : TStringList;
                   xColumns : array of integer;
                   dataRecord : TTimeSeriesRecord) : boolean;
var
  iColumn, index : integer;
  myStatus : string;
begin
  if(xColumns[COL_STATUS] >= 0) then    // extract status descriptor
  begin
    iColumn := xColumns[COL_STATUS];     // usual column for STATUS
    case FDeviceType of
      dtSEL3XX:
        begin
          Inc(iColumn,RELAY_3XX_OFFSET);
        end;
      dtSEL351RS:
        begin
          Inc(iColumn,RELAY_351RS_OFFSET);
        end;
      dtSEL4XX:
        begin
          Inc(iColumn,RELAY_4XX_OFFSET);
        end;
    end;
    myStatus := recordData[iColumn];
    if(Length(myStatus) > 0) then
    begin
      if(FChartType = ctSSI) then   // extract SSI Phase descriptor
      begin
        index := AnsiIndexStr(myStatus,['R','P','F','E','M','S','D','X']);
        if(index < 0) then
        begin
          result := false;
          exit;
        end;
      end;
    end else
    begin
      myStatus := '...';
    end;
  end else
  begin
    myStatus := '...'
  end;
  dataRecord.Status := myStatus;
  result := true;
end;
{************************************************************************
* NAME:       ReadChartRecordData
* FUNCTIONAL DESCRIPTION:    Read the DATA columns from the data record
* INPUT PARAMETERS:
*                    recordData : TStringList  content of present record
*                    dataCount : integer number of active data columns
*                    yColumns : array of integer  column position of dep. data values
*                    fileType : TTimeSeriesFileType (ftCSV, ftReport,...)
* OUTPUT PARAMETERS:
*                    dataRecord : TTimeSeriesRecord   record of data found
* RETURN:            result : boolean  record is valid (true/false)
************************************************************************}
function TTimeSeriesDataTable.ReadChartRecordData(
                   recordData : TStringList;
                   dataCount : integer;
                   yColumns : array of integer;
                   fileType : TTimeSeriesFileType;
                   dataRecord : TTimeSeriesRecord) : boolean;
var
  j,index, iColumn : integer;
  status : boolean;
  myValue : string;
begin
  status := true;  // assume read is successful
  // Extract the active dependent variable data columns
  index := 0;

  try
    SetLength(dataRecord.Value,dataCount+1);
    for j := Low(yColumns) to High(yColumns) do
    begin
      iColumn := yColumns[j];  // default column location
      if(iColumn < 0) then continue;  // inactive column
      if((FChartType = ctSSI) and (FDeviceType = dtSEL3XX)) then
      begin
        if(j > RELAY_3XX_START) then
        begin
          Inc(iColumn,RELAY_3XX_OFFSET);
        end;
      end;
      
      myValue := AnsiReplaceStr(recordData[iColumn],'%','');  // for Harmonic %
//      Assert(myValue = InternalizeData(myValue), 'These should be equivalent as myValue should already be internalized');
      if not TryNonLocaleStrToFloat(myValue, dataRecord.Value[index]) then
      begin
        dataRecord.Value[index] := 0.0;
        status := false;
      end;
      Inc(index);
    end;
    result := status;
  except
    result := false;  // bad numeric format
  end;
end;
{************************************************************************
* NAME:     DetermineDeviceType
* FUNCTIONAL DESCRIPTION:   Determine device type from column count
* INPUT PARAMETERS: fileType : TTimeSeriesFileType  ftReport, ftCSV
*                   columnCount : integer number of columns of data found
*                   lastColumn  : integer  last data column expected
* OUTPUT PARAMETERS: none
* RETURN: result : TTimeSeriesDeviceType (dtSEL7XX, ...)
************************************************************************}
function TTimeSeriesDataTable.DetermineDeviceType(fileType : TTimeSeriesFileType;
                 columnCount, lastColumn : integer): TTimeSeriesDeviceType;
begin
  if((fileType = ftReport) and (FChartType = ctSSI)) then
  begin
    if(isDevice73x(MeterFID)) then
    begin
      FDeviceType := dtSEL7XX;   // standard number of columns (based on 734)
    end else
    if((columnCount-1) > lastColumn) then   // must be a "351 Relay" SSI report
    begin
      FDeviceType := dtSEL3XX;   // more columns in relay SSI report
    end else
    if(columnCount <= SEL351RS_COUNT) then
    begin
      FDeviceType := dtSEL351RS;  // special format for 351 RS
    end else
    if((columnCount-1) < lastColumn) then   // must be a "487 Relay" SSI report
    begin
      FDeviceType := dtSEL4XX;   // Fewer columns in relay SSI report
    end else
    begin
      FDeviceType := dtSEL7XX;   // standard number of columns (based on 734)
    end;
  end else
  begin
    FDeviceType := dtSEL7XX;
  end;
  result := FDeviceType;
end;
{************************************************************************
* NAME:     IsValidDateTime
* FUNCTIONAL DESCRIPTION:   test date/time string for validity
* INPUT PARAMETERS: myDateTime : string  date/time string read from file
* OUTPUT PARAMETERS: none
* RETURN: result : boolean  true if date/time is valid otherwise false
************************************************************************}
function TTimeSeriesDataTable.IsValidDateTime(myDateTime:string): boolean;
begin
  result := true;  // accept any  date/time format for now
end;

{************************************************************************
* NAME:  IsInterruptRecord
* FUNCTIONAL DESCRIPTION:  Identify records where all active phases are interrupt
* INPUT PARAMETERS:  dataRecord : TTimeSeriesRecord   SSI record
* OUTPUT PARAMETERS: none
* RETURN: result : boolean   true if record is interrupt
************************************************************************}
function TTimeSeriesDataTable.IsInterruptRecord(dataRecord : TTimeSeriesRecord) : boolean;
const
  INTERRUPT_EVENT = 'I';
  NOMINAL_EVENT = '.';
var
  j, maxPhase : integer;
  isInterrupt : boolean;
begin
  if(DeviceType = dtSEL351RS) then
  begin
    maxPhase := PHASE_A;
  end else
  begin
    maxPhase := PHASE_C;
  end;

  isInterrupt := true;
  for j := PHASE_A to maxPhase do
  begin
    if((dataRecord.Phase[j] <> INTERRUPT_EVENT) and
       ((dataRecord.Phase[j] <> NOMINAL_EVENT) or
         (dataRecord.Value[VoltageIndices[j-1]] <> 0.0))) then
    begin
      isInterrupt := false;
      break;
    end;
  end;

  result := isInterrupt;
end;

{!******************************************************************************
* MODULE NAME ...  CompareRecords
* SYSTEM NAME ...  Acselerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:  Comparison function used by sort method
*
* INPUT PARAMETERS:
*                 Item1,Item2: Pointer           ptr to data objects being compared
*
* OUTPUT PARAMETERS:
*                    None:
*
* RETURN VALUE:
*                    result : integer -1 if Item1 < Item2
*                                     0 if Item1 = Item2
*                                     1 if Item1 > Item2
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.09.24  Kent Myers        Initial implementation
*******************************************************************************}
function CompareRecords(Item1,Item2: Pointer): Integer;
begin
  result := CompareDateTime(TTimeSeriesRecord(Item1).DateTime,
                            TTimeSeriesRecord(Item2).DateTime);
end;


end.
