{*****************************************************************************
* UNIT DESCRIPTION:  Base class for charting functions
* SPEC REF: (if used)
* DESIGN REF: (if used)
* 
* Copyright (c) 2009 Schweitzer Engineering Laboratories, Pullman, Washington
******************************************************************************}
unit ClassTimeSeriesChart;

interface
uses
  Windows, SysUtils, StrUtils, Classes, DateUtils, Math,
  TeeTools, Chart, TeeComma, TeeEdit, Graphics,
  TimeRoutines, ClassChartInput, ClassTimeSeries, TimeSeriesConstants;

const
  PHASE_A = 1;
  PHASE_C = 3;
type
  TLimits = Record
    limit1 : integer;
    limit2 : integer;
  end;

  TTimeSeriesChart = class(TObject)
  private
    { Private declarations }
    FMouseX,FMouseY         : integer; // last location of mouse after mouse down/move
    FChartVariableCount     : integer; // number of variables in chart (including CBEMA)
    FChartHasEvents         : boolean; // true if event limits are present
    FEventLastSortedColumn  : integer;
    FEventAscending         : boolean;
    FChartXDelta            : double;  // interval between chart labels

    FEventData              : TTimeSeriesDataTable;  // Table containing raw data

    FChartDisplay           : TChart;  // active chart
    FCursorTool             : TCursorTool;  // used for data query with mouse
    FChartRotate            : boolean;  // rotation is active
    FChartTranslate         : boolean;  // translation is active
    FChartZoom              : boolean;  // zoom is active
    FChartThickness         : boolean;  // 3D depth thickness change is active
    FChartActive            : boolean;  // dynamic operation active (rotate,...)
  public
    { Public declarations }
    procedure SaveStringToFile(filePath, aString : string; aEncoding: TEncoding);
    procedure RenameOldFile(filePath : string);
    function GetChartFilename(aFileType : TTimeSeriesFileType; const ExportPath, aID : string) : string;  virtual;
    function TrimFileName(filename : string) : string;
    function AppendCSVHeader(var MeterBuffer : string) : boolean;  virtual;
    function AppendCSVData(var MeterBuffer : string) : boolean;  virtual;
    function AppendReport(var MeterBuffer : string) : boolean;  virtual;
    function GetRecordFromEventTime(aDateTime : TDateTime) : Double;  virtual;
    function GetEventRecordFromRecord(myX : double) : integer;   virtual;
    function EventsInInterval : boolean;    virtual;
    function WhichInterval(iValue : integer; eLimits : array of TLimits) : integer;

    function CBEMAViolation(  indexRecord : TTimeSeriesIndexRecord) : TTimeSeriesCBEMAViolationType; virtual;

    procedure InitChart(fileType : TTimeSeriesFileType);  virtual;
    procedure UpdateEventScale; virtual;

    property ChartRotate : boolean read FChartRotate write FChartRotate;
    property ChartTranslate : boolean read FChartTranslate write FChartTranslate;
    property ChartZoom : boolean read FChartZoom write FChartZoom;
    property ChartThickness : boolean read FChartThickness write FChartThickness;
    property ChartActive : boolean read FChartActive write FChartActive;
    property ChartXDelta : double read FChartXDelta write FChartXDelta;
    property ChartDisplay : TChart read FChartDisplay write FChartDisplay;
    property CursorTool : TCursorTool read FCursorTool write FCursorTool;
    property MouseX : integer read FMouseX write FMouseX;
    property MouseY : integer read FMouseY write FMouseY;
    property ChartVariableCount : integer read FChartVariableCount; // write FChartVariableCount;
    property ChartHasEvents : boolean read FChartHasEvents write FChartHasEvents;
    property EventLastSortedColumn : integer read FEventLastSortedColumn write FEventLastSortedColumn;
    property EventAscending : boolean read FEventAscending write FEventAscending;
    property EventData : TTimeSeriesDataTable read FEventData; // write FEventData;

    Constructor Create(aOwner: TComponent);  virtual;
    Destructor Destroy();  override;
    procedure InitChartVariableCount();  virtual;
    procedure InitChartActiveChannels();   virtual;
    procedure SaveChartActiveChannels(); virtual;
    procedure InitChartActiveChannelsDefault(reset : boolean); virtual;
    procedure InitChartActiveEvents();
    procedure SaveChartActiveEvents();
    procedure InitChartActiveEventsDefault(); virtual;
    procedure InitChartTitles();   virtual;
    function GetChannelPhase(cname : string) : THMIChannelPhase;
    function GetChannelType(cname : string) : THMIChannelType;
    function ChartLineColor(cPhase: THMIChannelPhase; index : integer) : TColor;
    function ChartLineStyle(cType : THMIChannelType) : TPenStyle;
    procedure EraseChartDisplay();
    procedure AutoScaleAxes(reset : boolean);      overload; virtual;
    procedure AutoScaleAxes(var xMin,xMax,yMin,yMax : double);      overload; virtual;
    function FillEventIndexTable(filter : boolean) : boolean; virtual;
    function PhaseToEventType(start : integer; allowMultiEvent : boolean) : TTimeSeriesEventType;   overload;
    function PhaseToEventType(start: integer; previousEventIndex : integer) : TTimeSeriesEventType; overload;
    function PhaseToEventDescription(aEventType : TTimeSeriesEventType) : string;
    function GetMaxPhase() : integer;
    function CBEMAViolationString(Violation: TTimeSeriesCBEMAViolationType) : string;  virtual;

    function NonLocaleFloatToStr(value : double) : String;
    function TryNonLocaleStrToFloat(svalue : String; var value : double) : boolean;
    function NonLocaleStrToFloatDef(value : String; defaultValue : single) : single;

  end;

implementation

{************************************************************************
* NAME:   Create
* FUNCTIONAL DESCRIPTION: class constructor
* INPUT PARAMETERS:  none
* OUTPUT PARAMETERS: none
* RETURN:            none
************************************************************************}
constructor TTimeSeriesChart.Create(aOwner: TComponent);
begin
  inherited Create;
  FEventData := TTimeSeriesDataTable.Create;
  FChartXDelta := 10.0;
  FChartHasEvents := false;
  FEventAscending := true;
  FEventLastSortedColumn := -1;
end;
{************************************************************************
* NAME:   Destroy
* FUNCTIONAL DESCRIPTION: class destructor
* INPUT PARAMETERS:  none
* OUTPUT PARAMETERS: none
* RETURN:            none
************************************************************************}
destructor TTimeSeriesChart.Destroy();
begin
  FreeAndNil(FEventData);
  inherited Destroy();
end;
{************************************************************************
* NAME:   GetChartFilename
* FUNCTIONAL DESCRIPTION: Virtual base class
* INPUT PARAMETERS:  aChartType : THMIChartType (type of chart)
*                    aFileType : TTimeSeriesFileType  (type of file)
*                    ExportPath : default path to file
*                    aID        : ID of meter
* OUTPUT PARAMETERS: none
* RETURN:            filePath : string  path to file that was saved
************************************************************************}
function TTimeSeriesChart.GetChartFilename(aFileType : TTimeSeriesFileType; const ExportPath, aID : string) : string;
begin
  Assert(false,'Must implement this method in sub-class');
  result := ' ';
end;
{************************************************************************
* NAME:   TrimFileName
* FUNCTIONAL DESCRIPTION:  trim filename from the end of a file path name
* INPUT PARAMETERS: filename : string  is the original filename
* OUTPUT PARAMETERS: none
* RETURN:             result : string is the trimmed pathname
************************************************************************}
function TTimeSeriesChart.TrimFileName(filename : string) : string;
var
  newFilename : string;
  myPos, aLength : integer;
begin
  newFilename := filename;
  myPos := LastDelimiter('\',newFilename);   // look for last backslash
  if(myPos > 0) then
  begin
    aLength := Length(filename);
    Delete(newFilename,myPos+1,aLength); // leave a trailing backslash
  end;
  result := newFilename;
end;
{************************************************************************
* NAME:   SaveStringToFile
* FUNCTIONAL DESCRIPTION: Save string to given file name
* INPUT PARAMETERS:
*                    filePath : string  name of file to save string to
*                    aString : string   string to save
*                    eEncoding: encoding Format to save file as (ie ASSCI, Unincode,...)
* OUTPUT PARAMETERS: none
* RETURN:            none
************************************************************************}
procedure TTimeSeriesChart.SaveStringToFile(filePath, aString : string; aEncoding: TEncoding);
begin
  // Save data in file so that it may be read back as individual records
//  SaveTextDataToFile(aString,filePath,aEncoding);
end;
{************************************************************************
* NAME:   RenameOldFile
* FUNCTIONAL DESCRIPTION: Test for old file and add a numeric extension
* INPUT PARAMETERS:
*                    filePath : string  name of file to rename
* OUTPUT PARAMETERS: none
* RETURN:            none
************************************************************************}
procedure TTimeSeriesChart.RenameOldFile(filePath : string);
const
  EXTENSION_LENGTH = 3;
var
  newFileName : string;
  count, ipos : integer;
begin
  if(FileExists(filePath)) then    // test for existing file with same name and rename it
  begin
    count := 0;
    newFileName := filePath;
    while(FileExists(newFileName)) do   // strip off counter extension
    begin
      Inc(count);
      ipos := Pos('.',newFileName);   // find start of first extension
      newFileName := AnsiLeftStr(filePath,ipos+EXTENSION_LENGTH);  // restore original filename
      newFileName := newFileName + Format('.%d',[count]);  // append counter
    end;
    RenameFile(filePath,newFileName);
  end;
end;
{************************************************************************
* NAME:  InitChart
* FUNCTIONAL DESCRIPTION: Initialize chart class members before importing report file
* INPUT PARAMETERS: fileType : TTimeSeriesFileType   type of file (ftReport,...)
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************}
procedure TTimeSeriesChart.InitChart(fileType : TTimeSeriesFileType);
begin
  FEventData.ChartType := ctInvalid;
  FEventData.ChartName := 'Unknown';
  FEventData.ChartVersion := '1.0';
end;
{************************************************************************
* NAME:    AppendCSVHeader
* FUNCTIONAL DESCRIPTION:  Append header records for CSV file to string
* INPUT PARAMETERS: none
* OUTPUT PARAMETERS: MeterBuffer : string  file content
* RETURN: result : boolean   status of append (true = success)
************************************************************************}
function TTimeSeriesChart.AppendCSVHeader(var MeterBuffer : string) : boolean;
const
  CRLF = Chr(13) + Chr(10);     // carriage return line feed
var
  i : Integer;      // indices
  settingRecord : TTimeSeriesSettingRecord;
  myFormat : TFormatSettings;
  myDate,myTime,myScale,myUnits,mySubType : string;
begin

  Assert(FEventData <> nil,'EventData is nil');

  FEventData.InitDateTimeFormat(ftEnglish_US,myFormat);
  myFormat.ShortDateFormat := 'MM/DD/YY';
  myUnits := LDP_UNITS_LIST[integer(FEventData.LDPUnits)+1];
  myScale := LDP_SCALE_LIST[integer(FEventData.LDPScale)+1];
  mySubType := CHART_SUBTYPE_LIST[integer(FEventData.ChartSubType)+1];

  With FEventData do
  begin

    Assert(RawDataLabels <> nil,'RawDataLabels is nil');
    Assert(RawDataTable <> nil,'RawDataTable is nil');
    Assert(IndexTable <> nil,'IndexTable is nil');

    try
      myDate := DateToStr(FEventData.ChartDateTime,myFormat);  // use setting date format
      myTime := FormatDateTime('hh:mm:ss.zzz',FEventData.ChartDateTime);
      MeterBuffer := MeterBuffer +
                     'FileType, Version, Date, Time, ChartField, TimeSource, Recorder, Scale, Units, ChartSubType' + CRLF +
                     ChartName + ', ' + ChartVersion + ', ' +
                     myDate + ', ' + myTime + ', ' +
                     ChartField + ', ' + TimeSource +  ', ' +
                     Format('%d',[LDPRecorder]) + ', ' +
                     myScale + ', ' + myUnits + ', ' + mySubType + CRLF;

      // Export header columns (indep. vars)

      if(Length(HeaderColumnArray) < 5) then
      begin
        raise Exception.Create('Insufficient number of independent variables.');
      end;

      MeterBuffer := MeterBuffer + 'RecNum, Date, Time, Phase, Status' + CRLF;
      for i := 0 to Length(HeaderColumnArray) - 1 do
      begin
        if(i < (Length(HeaderColumnArray) - 1)) then
        begin
          MeterBuffer := MeterBuffer + Format('%d, ',[HeaderColumns[i]]);
        end else
        begin
          MeterBuffer := MeterBuffer + Format('%d',[HeaderColumns[i]]) + CRLF;
        end;
      end;

      // Export data columns

      if(Length(DataColumnArray) < 1) then
      begin
        raise Exception.Create('Insufficient number of dependent variables.');
      end;

      // Channel names
      for i := 0 to RawDataLabels.Count - 1 do
      begin
        if(i < (RawDataLabels.Count-1)) then
        begin
          MeterBuffer := MeterBuffer + Format('%s, ',[RawDataLabels[i]]);
        end else
        begin
          MeterBuffer := MeterBuffer + Format('%s',[RawDataLabels[i]]) + CRLF;
        end;
      end;

      // data column numbers (dependent vars)
      for i := 0 to Length(DataColumnArray) - 1 do
      begin
        if(i < (Length(DataColumnArray)-1)) then
        begin
          MeterBuffer := MeterBuffer + Format('%d, ',[DataColumns[i]]);
        end else
        begin
          MeterBuffer := MeterBuffer + Format('%d',[DataColumns[i]]) + CRLF;
        end;
      end;

      // Table of Device information

      MeterBuffer := MeterBuffer + 'DeviceID, FID, TID';
      for i := 0 to SettingTable.Count - 1 do
      begin
        settingRecord := (SettingTable[i] as TTimeSeriesSettingRecord);
        MeterBuffer := MeterBuffer + ', ' + settingRecord.SettingName;
      end;
      MeterBuffer := MeterBuffer + CRLF;

      MeterBuffer := MeterBuffer + Format('"%s", "%s", "%s"',
              [MeterID,
               MeterFID,
               MeterTID]);
      for i := 0 to SettingTable.Count - 1 do
      begin
        settingRecord := (SettingTable[i] as TTimeSeriesSettingRecord);
        MeterBuffer := MeterBuffer + ', ' + NonLocaleFloatToStr(settingRecord.SettingValue);
      end;
      MeterBuffer := MeterBuffer + CRLF;

    except
      result := false;
      exit;
    end;
  end;
  result := true;
end;

function TTimeSeriesChart.NonLocaleFloatToStr(value : double) : String;
begin
  result := FloatToStr(value);
end;

function TTimeSeriesChart.TryNonLocaleStrToFloat(svalue : String; var value : double) : boolean;
begin
  result := TryStrToFloat(svalue, value);
end;

function TTimeSeriesChart.NonLocaleStrToFloatDef(value : String; defaultValue : single) : single;
begin
  result := StrToFloatDef(value,defaultValue);
end;

{************************************************************************
* NAME:    AppendCSVData
* FUNCTIONAL DESCRIPTION: Virtual base class method
* INPUT PARAMETERS:       none
* OUTPUT PARAMETERS:      MeterBuffer : string  file content
* RETURN:                 result : boolean   true if data appended, otherwise false
************************************************************************}
function TTimeSeriesChart.AppendCSVData(var MeterBuffer : string) : boolean;
begin
  // intentionally empty
  Assert(false,'Must implement this method in sub-class');
  result := false;
end;
{************************************************************************
* NAME:    AppendReport
* FUNCTIONAL DESCRIPTION: Virtual base class method
* INPUT PARAMETERS:    none
* OUTPUT PARAMETERS:   MeterBuffer : string  file content
* RETURN:              result : boolean   true if data appended, otherwise false
************************************************************************}
function TTimeSeriesChart.AppendReport(var MeterBuffer : string) : boolean;
begin
  // empty function
  Assert(false,'Must implement this method in sub-class');
  result := false;
end;
{************************************************************************
* NAME:    InitChartVariableCount
* FUNCTIONAL DESCRIPTION:
* Set the chart variable count based on chart type and event count
* INPUT PARAMETERS:    none
* OUTPUT PARAMETERS:   none
* RETURN:              none
************************************************************************}
procedure TTimeSeriesChart.InitChartVariableCount();
begin
  // update chart variable count to include CBEMA curves if appropriate

  Assert(FEventData <> nil,'EventData is nil');
  Assert(FEventData.RawDataLabels <> nil,'EventData is nil');

  FChartVariableCount := FEventData.RawDataLabels.Count;  // number of channels to display
  FChartHasEvents := false;  // assume event limits not displayed
end;

{************************************************************************
* NAME:    InitChartActiveChannels
* FUNCTIONAL DESCRIPTION: Virtual base class method
* INPUT PARAMETERS:    none
* OUTPUT PARAMETERS:   none
* RETURN:              none
************************************************************************}
procedure TTimeSeriesChart.InitChartActiveChannels();
begin
// intentionally empty
  Assert(false,'Must implement this method in sub-class');
end;
{************************************************************************
* NAME:    InitChartActiveChannelsDefault
* FUNCTIONAL DESCRIPTION:   Virtual base class method
 INPUT PARAMETERS:    reset : boolean  
* OUTPUT PARAMETERS:   none
* RETURN:              none
************************************************************************}
procedure TTimeSeriesChart.InitChartActiveChannelsDefault(reset : boolean);
begin
// intentionally blank
  Assert(false,'Must implement this method in sub-class');
end;
{************************************************************************
* NAME:    SaveChartActiveChannels
* FUNCTIONAL DESCRIPTION:  Virtual base class method
 INPUT PARAMETERS:    none
* OUTPUT PARAMETERS:   none
* RETURN:              none
************************************************************************}
procedure TTimeSeriesChart.SaveChartActiveChannels();
begin
// intentionally empty
  Assert(false,'Must implement this method in sub-class');
end;
{************************************************************************
* NAME:    InitChartActiveEvents
* FUNCTIONAL DESCRIPTION:  Set the active series in the chart
 INPUT PARAMETERS:    none
* OUTPUT PARAMETERS:   none
* RETURN:              none
************************************************************************}
procedure TTimeSeriesChart.InitChartActiveEvents();
var
  i : integer;
  indexRecord : TTimeSeriesIndexRecord;
begin
  Assert(FEventData <> nil,'EventData is nil');
  Assert(FEventData.IndexTable <> nil,'IndexTable is nil');

  for i := 0 to FEventData.IndexTable.Count-1 do
  begin
    indexRecord := (FEventData.IndexTable[i] as TTimeSeriesIndexRecord);
    indexRecord.Checked := FEventData.InputSettings.ActiveEvents[i];
  end;
end;
{************************************************************************
* NAME:    InitChartActiveEventsDefault
* FUNCTIONAL DESCRIPTION: Virtual base class method
 INPUT PARAMETERS:    none
* OUTPUT PARAMETERS:   none
* RETURN:              none
************************************************************************}
procedure TTimeSeriesChart.InitChartActiveEventsDefault();
begin
   // intentionally empty
  Assert(false,'Must implement this method in sub-class');
end;
{************************************************************************
* NAME:    SaveChartActiveEvents
* FUNCTIONAL DESCRIPTION: Save the active series in the chart
 INPUT PARAMETERS:    none
* OUTPUT PARAMETERS:   none
* RETURN:              none
************************************************************************}
procedure TTimeSeriesChart.SaveChartActiveEvents();
var
  i : integer;
  indexRecord : TTimeSeriesIndexRecord;
begin
  Assert(FEventData <> nil,'EventData is nil');
  Assert(FEventData.IndexTable <> nil,'IndexTable is nil');

  for i := 0 to FEventData.IndexTable.Count-1 do
  begin
    indexRecord := (FEventData.IndexTable[i] as TTimeSeriesIndexRecord);
    FEventData.InputSettings.ActiveEvents[i] := indexRecord.Checked;
  end;
end;

{************************************************************************
* NAME:   UpdateEventScale
* FUNCTIONAL DESCRIPTION:  Virtual base class method
* INPUT PARAMETERS: not used
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************}
procedure TTimeSeriesChart.UpdateEventScale;
begin
  // intentionally empty
end;
{************************************************************************
* NAME:   WhichInterval
* FUNCTIONAL DESCRIPTION:  determine which interval the given value is in
* INPUT PARAMETERS:
*                    iValue : integer   value to test
*                    elimits1 : array of integer   lower bound of each interval
*                    elimits2 : array of integer   upper bound of each interval
* OUTPUT PARAMETERS: none
* RETURN:  result : integer   interval number where value is found (-1 if not in any interval)
************************************************************************}
function TTimeSeriesChart.WhichInterval(iValue: integer; eLimits : array of TLimits) : integer;
var
  i,itest : integer;
begin
  itest := -1;

  for i := Low(eLimits) to High(eLimits) do
  begin
    if((iValue >= eLimits[i].limit1) and (iValue <= eLimits[i].limit2)) then  // in this interval
    begin
      itest := i;
      break;
    end;
  end;
  result := itest;
end;
{************************************************************************
* NAME:     EventsInInterval
* FUNCTIONAL DESCRIPTION:  Virtual base class method
* INPUT PARAMETERS: none
* OUTPUT PARAMETERS: none
* RETURN: result : boolean   true if events occur, otherwise false
************************************************************************}
function TTimeSeriesChart.EventsInInterval : boolean;
begin
  // intentionally blank
  Assert(false,'Must implement this method in sub-class');
  result := false;
end;

{************************************************************************
* NAME:   FillEventIndexTable
* FUNCTIONAL DESCRIPTION:  Virtual base class method
* INPUT PARAMETERS:
*                    filter: boolean  use filter controls true/false
* OUTPUT PARAMETERS: none
* RETURN:  result : boolean  true if events found, otherwise false
************************************************************************}
function TTimeSeriesChart.FillEventIndexTable(filter : boolean) : boolean;
begin
  // intentionally blank
  Assert(false,'Must implement this method in sub-class');
  result := false;
end;

{************************************************************************
* NAME:     PhaseToEventType
* FUNCTIONAL DESCRIPTION:   Convert phase string to event type
* INPUT PARAMETERS:
*                    start : integer record number where event starts
*                allowMultiEvent : boolen true if evMulti is a valid return
* OUTPUT PARAMETERS: none
* RETURN:
*                    label : TTimeSeriesEventType  event type
************************************************************************}
function TTimeSeriesChart.PhaseToEventType(start: integer; allowMultiEvent : boolean) : TTimeSeriesEventType;
var
  foundU, foundO, foundI : boolean;
  eventType : TTimeSeriesEventType;
  eventPhase : string;
  i,maxPhase: Integer;
  dataRecord : TTimeSeriesRecord;
begin
  maxPhase := GetMaxPhase();

  foundU := false;      // sag
  foundO := false;      // swell
  foundI := false;      // interrupt

  if(start < FEventData.RawDataTable.Count) then
  begin
    dataRecord := (FEventData.RawDataTable[start] as TTimeSeriesRecord);
    eventPhase := dataRecord.Phase;

    if(FEventData.IsInterruptRecord(dataRecord)) then
    begin
      foundI := true;  // interrupt must occur on all three phases
    end else
    begin
      for i := PHASE_A to maxPhase do
      begin
        if((AnsiCompareText(eventPhase[i],'U') = 0) or
           (AnsiCompareText(eventPhase[i],'I') = 0)) then
        begin
          foundU := true;
        end else
        if(AnsiCompareText(eventPhase[i],'O') = 0) then
        begin
          foundO := true;
        end;
        // In most cases, multi-events are not relevant, just use the first
        // non-nominal phase to indicate the event type.
        if(not allowMultiEvent) then
        begin
          if(foundU or foundO) then
          begin
            break;
          end;
        end;
      end;
    end;

    eventType := evNone;
    if((foundU and foundO) or (foundU and foundI) or (foundO and foundI)) then
    begin
      eventType := evMulti;
    end else
    if(foundU) then
    begin
      eventType := evSag;
    end else
    if(foundO) then
    begin
      eventType := evSwell;
    end else
    if(foundI) then
    begin
      eventType := evInterrupt;
    end;
  end else
  begin
    eventType := evNone;
  end;

  result := eventType;
end;

{************************************************************************
* NAME:     PhaseToEventType
* FUNCTIONAL DESCRIPTION:   Convert phase string to event type
* INPUT PARAMETERS:
*                    start : integer record number where event starts
* OUTPUT PARAMETERS: none
* RETURN:
*                    label : TTimeSeriesEventType  event type
************************************************************************}
function TTimeSeriesChart.PhaseToEventType(start: integer; previousEventIndex : integer) : TTimeSeriesEventType;
var
  eventType : TTimeSeriesEventType;
  indexRecord : TTimeSeriesIndexRecord;
begin
  // test for a multi-event record starting this event
  eventType := PhaseToEventType(start, true);

  {NOTE: This method checks for the special case of a mixed event
  immediately following a previous event (no intermediate records).
  If so, it is necessary to look at previous event to determine type
  and the following rules apply:
    - If it was an interrupt, this must be part of a sag because
      an interrupt cannot be a sub-event of a swell.
    - If it was a swell, this must be a new sag because in order
      for it to be a new event it can have no swell terms.  It
      cannot be an interrupt because that requires all three
      phases to be interrupt.
    - It is not possible for previous to be a sag, because no phases
      can be sag or interrupt in order to leave the sag to start
      this new event.
  }

  if(eventType = evMulti) then
  begin
    Assert(FEventData <> nil,'EventData is nil');
    Assert(FEventData.IndexTable <> nil,'IndexTable is nil');
    if(previousEventIndex >= 0) then
    begin
      indexRecord := (FEventData.IndexTable[previousEventIndex] as TTimeSeriesIndexRecord);
      // see if there was an immediate transition to a new event
      if(indexRecord.LastRecord = start) then
      begin
        eventType := evSag;
      end else
      begin
        eventType := PhaseToEventType(start, false);
      end;
    end else
    begin
      eventType := PhaseToEventType(start, false);
    end;
  end;
  result := eventType;
end;

{************************************************************************
* NAME:     GetMaxPhase
* FUNCTIONAL DESCRIPTION:   Get maximum active phase for device
* INPUT PARAMETERS:  none
* OUTPUT PARAMETERS: none
* RETURN:
*                    result : integer : number of active phases
************************************************************************}
function TTimeSeriesChart.GetMaxPhase() : integer;
var
  maxPhase : integer;
begin
  if(EventData.DeviceType = dtSEL351RS) then
  begin
    maxPhase := PHASE_A;
  end else
  begin
    maxPhase := PHASE_C;
  end;
  result := maxPhase;
end;

{************************************************************************
* NAME:     PhaseToEventDescription
* FUNCTIONAL DESCRIPTION:   Create a descriptive event label for event summary based on phase labels
* INPUT PARAMETERS:
*                    aEventType : THMIEventType
* OUTPUT PARAMETERS: none
* RETURN:
*                    result : string  descriptive label for event
************************************************************************}
function TTimeSeriesChart.PhaseToEventDescription(aEventType : TTimeSeriesEventType) : string;
var
  eventName : string;
begin

  case aEventType of
    evNone:
    begin
      eventName := 'TRIG';
    end;
    evSag:
    begin
      eventName := 'SAG';
    end;
    evSwell:
    begin
      eventName := 'SWELL';
    end;
    evMulti:
    begin
      eventName := 'Multi-Event';
    end;
    evInterrupt:
    begin
      eventName := 'INT';
    end;
    else
    begin
      eventName := 'None';
      Assert(false,'Invalid Event Type');
    end;
  end;

  result := eventName;
end;
{************************************************************************
* NAME:  GetRecordFromEventTime
* FUNCTIONAL DESCRIPTION:  Virtual base class method
* INPUT PARAMETERS: aDateTime : TDateTime   date/time of event
* OUTPUT PARAMETERS: none
* RETURN: result : Double  interpolated record number for event
************************************************************************}
function TTimeSeriesChart.GetRecordFromEventTime(aDateTime : TDateTime) : Double;
begin
  // intentionally empty
  Assert(false,'Must implement this method in sub-class');
  result := 0.0;
end;
{************************************************************************
* NAME:  GetEventRecordFromRecord
* FUNCTIONAL DESCRIPTION: Virtual base class method
* INPUT PARAMETERS: myX : double  interpolated record number of event record
* OUTPUT PARAMETERS: none
* RETURN: result : index of Event record
************************************************************************}
function TTimeSeriesChart.GetEventRecordFromRecord(myX : double) : integer;
begin
  // intentionally empty
  Assert(false,'Must implement this method in sub-class');
  result := 0;
end;
{************************************************************************
* NAME: CBEMAViolation
* FUNCTIONAL DESCRIPTION: Virtual base class method
* INPUT PARAMETERS: indexRecord : TTimeSeriesIndexRecord
* OUTPUT PARAMETERS: none
* RETURN: result : TTimeSeriesCBEMAViolationType
************************************************************************}
function TTimeSeriesChart.CBEMAViolation( indexRecord : TTimeSeriesIndexRecord) : TTimeSeriesCBEMAViolationType;
begin
  // intentionally empty
  Assert(false,'Must implement this method in sub-class');
  result := cvtNone;
end;
{************************************************************************
* NAME: InitChartTitles
* FUNCTIONAL DESCRIPTION: Set chart titles (main, x and y axes)
* INPUT PARAMETERS: none
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************}
procedure TTimeSeriesChart.InitChartTitles();
var
  i  : integer;
begin
  if(FChartDisplay.Title.Text.Count > 1) then    // get rid of old titles
  begin
    for i := 1 to ChartDisplay.Title.Text.Count - 1 do
    begin
      FChartDisplay.Title.Text.Delete(1);  // always delete position 1 (list shrinks)
    end;
  end;
  if(FChartDisplay.Title.Text.Count > 0) then
  begin
    FChartDisplay.Title.Text[0] := 'Click Browse button to read chart data file.';
  end;
end;

{!******************************************************************************
* MODULE NAME ...  GetChannelPhase
* SYSTEM NAME ...  AcSELerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:  Determine channel phase from name
*
* INPUT PARAMETERS:
*                    cName : string   channel name
*
* OUTPUT PARAMETERS:
*                    None:
*
* RETURN VALUE:
*                    result : THMIChannelPhase   channel phase
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.10.1  Kent Myers        Original implementation
*******************************************************************************}
function TTimeSeriesChart.GetChannelPhase(cName : string) : THMIChannelPhase;
const
  PHASE_A : array [0..3] of string = ('a','A','ab','AB');
  PHASE_B : array [0..3] of string = ('b','B','bc','BC');
  PHASE_C : array [0..3] of string = ('c','C','ca','CA');
var
  i: Integer;
  myPhase : THMIChannelPhase;
begin

  myPhase := cpOther;

  for i := Low(PHASE_A) to High(PHASE_A) do
  begin
    if(AnsiEndsStr(PHASE_A[i],cName)) then
    begin
      myPhase := cpA;
      break;
    end;
  end;

  if(myPhase = cpOther) then
  begin
    for i := Low(PHASE_B) to High(PHASE_B) do
    begin
      if(AnsiEndsStr(PHASE_B[i],cName)) then
      begin
        myPhase := cpB;
        break;
      end;
    end;
  end;

  if(myPhase = cpOther) then
  begin
    for i := Low(PHASE_C) to High(PHASE_C) do
    begin
      if(AnsiEndsStr(PHASE_C[i],cName)) then
      begin
        myPhase := cpC;
        break;
      end;
    end;
  end;

  result := myPhase;

end;
{!******************************************************************************
* MODULE NAME ...  GetChannelType
* SYSTEM NAME ...  AcSELerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:  Determine channel type from name
*
* INPUT PARAMETERS:
*                    cName : string   channel name
*
* OUTPUT PARAMETERS:
*                    None:
*
* RETURN VALUE:
*                    result: THMIChannelType   channel type
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.10.1  Kent Myers        Original implementation
*******************************************************************************}
function TTimeSeriesChart.GetChannelType(cName : string) : THMIChannelType;
begin
  if(AnsiStartsStr('I',cName)) then
  begin
    result := ctCurrent;
  end else
  if(AnsiStartsStr('V',cName)) then
  begin
    result := ctVoltage;
  end else
  if(AnsiStartsStr('P',cName)) then
  begin
    result := ctPower;
  end else
  begin
    result := ctOther;
  end;
end;
{!******************************************************************************
* MODULE NAME ...  ChartLineStyle
* SYSTEM NAME ...  AcSELerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:  Return line style based on channel type
*
* INPUT PARAMETERS:
*                    cType : THMIChannelType   channel type
*
* OUTPUT PARAMETERS:
*                    None:
*
* RETURN VALUE:
*                    result: TPenStyle   line style
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.10.1  Kent Myers        Original implementation
*******************************************************************************}
function TTimeSeriesChart.ChartLineStyle(cType : THMIChannelType) : TPenStyle;
const
  MAX_TYPE_STYLE = 4;
  myTypeStyle : array[0..MAX_TYPE_STYLE-1] of TPenStyle =
  (
  psDash, psSolid, psDashDot, psSolid
  );
begin
  result := myTypeStyle[integer(cType)];
end;

{!******************************************************************************
* MODULE NAME ...  ChartLineColor
* SYSTEM NAME ...  AcSELerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:  Return line color based on channel phase
*
* INPUT PARAMETERS:
*                    cPhase : THMIChannelPhase   channel phase
*                    index : integer    channel number
*
* OUTPUT PARAMETERS:
*                    None:
*
* RETURN VALUE:
*                    result: TColor   line color
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.10.1  Kent Myers        Original implementation
*******************************************************************************}
function TTimeSeriesChart.ChartLineColor(cPhase : THMIChannelPhase; index : integer) : TColor;
const
  MAX_COLOR = 12;
  myColorList : array[0..MAX_COLOR-1] of TColor =
  (
    clGreen, clFuchsia, clYellow, clOlive, clTeal, clGray, clAqua, clLtGray, clDkGray, clSkyBlue, clMedGray, clMaroon
  );

  MAX_PHASE_COLOR = 4;
  myPhaseColor : array[0..MAX_PHASE_COLOR-1] of TColor =
  (
  clBlack, clWebFirebrick, clBlue, clGreen
  );
var
  adiv, amod : word;
begin
  if(cPhase = cpOther) then
  begin
    DivMod(word(index),MAX_COLOR,adiv,amod);  // wrap around if not enough colors
    result := myColorList[amod];
  end else
  begin
    result := myPhaseColor[integer(cPhase)]; 
  end;
end;
{************************************************************************
* NAME:     EraseChartDisplay
* FUNCTIONAL DESCRIPTION:  Erase content of chart
* INPUT PARAMETERS: none
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************}
procedure TTimeSeriesChart.EraseChartDisplay();
begin
  FChartDisplay.FreeAllSeries();   // throw away previous chart content
  InitChartVariableCount;
  InitChartTitles;
end;
{************************************************************************
* NAME:     AutoScaleAxes
* FUNCTIONAL DESCRIPTION:  Virtual base class method
* INPUT PARAMETERS: none
* OUTPUT PARAMETERS: none
* RETURN:           none
************************************************************************}
procedure TTimeSeriesChart.AutoScaleAxes(reset : boolean);
begin
  // intentionally empty
  Assert(false,'Must implement this method in sub-class');
end;
{************************************************************************
* NAME:     AutoScaleAxes
* FUNCTIONAL DESCRIPTION:  Virtual base class method
* INPUT PARAMETERS: none
* OUTPUT PARAMETERS: none
* RETURN:           none
************************************************************************}
procedure TTimeSeriesChart.AutoScaleAxes(var xMin,xMax,yMin,yMax : double);
begin
  ChartDisplay.BottomAxis.Automatic := false;
  ChartDisplay.LeftAxis.Automatic := false;
  xMin := ChartDisplay.MinXValue(ChartDisplay.BottomAxis);
  xMax := ChartDisplay.MaxXValue(ChartDisplay.BottomAxis);
  yMin := ChartDisplay.MinYValue(ChartDisplay.LeftAxis);
  yMax := ChartDisplay.MaxYValue(ChartDisplay.LeftAxis);
  ChartDisplay.LeftAxis.Automatic := false;
end;
{************************************************************************
* NAME:     CBEMAViolationString
* FUNCTIONAL DESCRIPTION:  Virtual base class method
* INPUT PARAMETERS: none
* OUTPUT PARAMETERS: none
* RETURN:           none
************************************************************************}
function TTimeSeriesChart.CBEMAViolationString(Violation: TTimeSeriesCBEMAViolationType) : string;
begin
  // intentionally empty
  Assert(false,'Must implement this method in sub-class');
  result:= ' ';
end;

end.
