{*****************************************************************************
* UNIT DESCRIPTION:  Base class for Harmonic charting functions
* SPEC REF: (if used)
* DESIGN REF: (if used)
* 
* Copyright (c) 2009 Schweitzer Engineering Laboratories, Pullman, Washington
******************************************************************************}
unit ClassTimeSeriesChartHRM;

interface
uses
  Windows, SysUtils, StrUtils, Classes, DateUtils, Math,
  TeeTools, Chart, TeeComma, TeeEdit, Graphics, Series,
  TimeRoutines, ClassChartInput, ClassTimeSeries, ClassTimeSeriesChart,
  ClassChartInputHRM, TimeSeriesConstants;
const
  CRLF = Chr(13) + Chr(10);     // carriage return line feed
  CRLF2 = CRLF + CRLF;
type
  TTimeSeriesChartHRM = class(TTimeSeriesChart)
  private
    { Private declarations }
  public
    Constructor Create(aOwner: TComponent); override;
    function GetChartFilename(aFileType : TTimeSeriesFileType; const ExportPath, aID : string) : string;  overload; override;
    procedure InitChart(fileType : TTimeSeriesFileType);  override;
    procedure InitChartTitles(); override;
    procedure InitChartActiveChannels();   override;
    procedure InitChartActiveChannelsDefault(reset : boolean); override;
    procedure AutoScaleAxes(reset : boolean);     overload; override;
    procedure InitChartVariableCount();  override;
    procedure SaveChartActiveChannels(); override;
    function AppendCSVHeader(var MeterBuffer : string) : boolean;  override;
    function AppendCSVData(var MeterBuffer : string) : boolean;  override;
    function AppendReport(var MeterBuffer : string) : boolean; override;
    function FillEventIndexTable(filter : boolean) : boolean; override;
  end;

implementation


{************************************************************************
* NAME:   Create
* FUNCTIONAL DESCRIPTION: class constructor
* INPUT PARAMETERS:  none
* OUTPUT PARAMETERS: none
* RETURN:            none
************************************************************************}
constructor TTimeSeriesChartHRM.Create(aOwner: TComponent);
begin
  inherited Create(aowner);
end;
{************************************************************************
* NAME:   GetChartFilename
* FUNCTIONAL DESCRIPTION: Get filename where chart data is saved by default
* INPUT PARAMETERS:  aFileType : TTimeSeriesFileType  (type of file)
*                    ExportPath : default path to file
*                    aID        : ID of meter
* OUTPUT PARAMETERS: none
* RETURN:            filePath : string  path to file where chart data is saved
************************************************************************}
function TTimeSeriesChartHRM.GetChartFilename(aFileType : TTimeSeriesFileType; const ExportPath, aID : string) : string;
var
  myID : string;
  myPath : string;
begin

 // Choose file names based on Chart type

  myID := Trim(aID);  // If ID is blank, assume it is to be ignored

  if(aFileType = ftCSV) then       // CSV file
  begin
    if(Length(myID) > 0) then
    begin
      myPath := ExportPath + Format('HRM Export %s_%s.csv', [FormatDateTime('yyyy-mmm-dd', Now),myID]);
    end else
    begin
      myPath := ExportPath + Format('HRM Export %s.csv', [FormatDateTime('yyyy-mmm-dd', Now)]);
    end;
  end else
  begin                           // ASCII encoded binary file
    if(Length(myID) > 0) then
    begin
      myPath := ExportPath + Format('HRM Export %s_%s.txt', [FormatDateTime('yyyy-mmm-dd', Now),myID]);
    end else
    begin
      myPath := ExportPath + Format('HRM Export %s.txt', [FormatDateTime('yyyy-mmm-dd', Now)]);
    end;
  end;
  result := myPath;
end;
{************************************************************************
* NAME:  InitChart
* FUNCTIONAL DESCRIPTION: Initialize chart class members before importing
*                         LDP report file (column convention must be followed)
* INPUT PARAMETERS: fileType : TTimeSeriesFileType   type of file (ftReport,...)
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************}
procedure TTimeSeriesChartHRM.InitChart(fileType : TTimeSeriesFileType);
const
  standardHRM : array[0..4] of integer = (0,-1,-1,-1,-1);   // record, date, time, phase, status
var
  i, lastStandard : integer;
begin

  Assert(EventData <> nil,'EventData is nil');
  inherited InitChart(fileType);

  EventData.ChartType := ctHarmonic;
  EventData.ChartName := 'HRM';

  if(EventData.FileType = ftReport) then
  begin
    EventData.SetLengthHeaderColumns(Length(standardHRM));
    EventData.SetLengthDataColumns(EventData.RawDataLabels.Count);
    lastStandard := 0;  // find last standard column used

    // Initialize standard columns.  (Independent variables)
    for i := Low(standardHRM) to High(standardHRM) do
    begin
      EventData.HeaderColumns[i] := standardHRM[i];
      lastStandard := Max(lastStandard,standardHRM[i]);  // find last column used
    end;
    // Initialize data columns (dependent variables)
    for i := 0 to EventData.RawDataLabels.Count - 1 do
    begin
      EventData.DataColumns[i] := lastStandard + i + 1;
    end;
  end;
end;
{************************************************************************
* NAME:    InitChartVariableCount
* FUNCTIONAL DESCRIPTION:
* Set the chart variable count based on chart type and event count
* INPUT PARAMETERS:    none
* OUTPUT PARAMETERS:   none
* RETURN:              none
************************************************************************}
procedure TTimeSeriesChartHRM.InitChartVariableCount();
begin

  Assert(EventData <> nil,'EventData is nil');

  inherited InitChartVariableCount;

//  if(EventData.IndexTable.Count > 0) then     // LDP SER events are active
//  begin
//    ChartHasEvents := true;
//  end;
end;
{************************************************************************
* NAME:    SetChartActiveChannels
* FUNCTIONAL DESCRIPTION:  Set the active series in the chart based on
*                          the user's input selections.
* INPUT PARAMETERS:    none
* OUTPUT PARAMETERS:   none
* RETURN:              none
************************************************************************}
procedure TTimeSeriesChartHRM.InitChartActiveChannels();
var
  i,sc,ac,size : integer;
  status : boolean;
begin
  Assert(EventData <> nil,'EventData is nil');

  sc := ChartDisplay.SeriesCount;
  ac := Length(EventData.InputSettings.ActiveChannelArray);
  size := min(sc,ac);

  {NOTE: sc and ac should never be different.  However, there appears to
   be a memory corruption prior to when the series is freed by
   FChartTimeSeries.ChartDisplay.FreeAllSeries().  After the Free operation,
   some series remain in the object, causing the count to be incorrect.
   This seems to happen only when settings are read into Quickset prior to charting.
   Until this condition is corrected, the following initialization will use
   the minimum of sc and ac.}

  Assert(sc <= ac, Format('Array size mismatch sc (%d) > ac (%d)',[sc,ac]));

  // update chart active channels

  for i := 0 to size-1 do
  begin
    status :=  EventData.InputSettings.ActiveChannels[i];
    ChartDisplay.Series[i].Active := status;
  end;

end;
{************************************************************************
* NAME:    InitChartActiveChannelsDefault
* FUNCTIONAL DESCRIPTION:
* Set the default state of active channels
* INPUT PARAMETERS:    reset : boolean  reset channel status to default if true
* OUTPUT PARAMETERS:   none
* RETURN:              none
************************************************************************}
procedure TTimeSeriesChartHRM.InitChartActiveChannelsDefault(reset : boolean);
var
  i,oldCount,newCount: Integer;
  oldState : array of boolean;
begin
  Assert(EventData <> nil,'EventData is nil');

  {The Active Channel array is a list of booleans that indicates whether
   each channel is selected for display in the chart (active).  There may
   be one additional value in the chart to represent event markers.  These
   markers indicate the occurrence of an LDP event during the time interval.
   If events are present, one additional member is added to the channel array.
  }
  oldCount := Length(EventData.InputSettings.ActiveChannelArray);
  newCount := ChartVariableCount;

  if(ChartHasEvents) then // allocate one more position and disable events by default
  begin
    newCount := ChartVariableCount+1;
  end;

  if(reset or (newCount <> oldCount)) then   // length has changed
  begin
    if(not reset) then     // save active channel state before re-allocating
    begin
      SetLength(oldState,oldCount);
      for i := 0 to oldCount - 1 do   // remember previous active channel state
      begin
        oldState[i] := EventData.InputSettings.ActiveChannels[i];
      end;
    end;

    EventData.InputSettings.SetActiveChannelLength(newCount);
    EventData.InputSettings.ActiveChannels[newCount-1] := false; // disable last channel (event)

    if((EventData.ChartSubType = cstHRM_MAI) or (EventData.ChartSubType = cstHRM_MAV)) then
    begin
//      for i := 0 to ChartVariableCount - 1 do
//      begin
//        EventData.InputSettings.ActiveChannels[i] := ((i mod 2) = 0); // magnitudes active
//      end;
      for i := 0 to ChartVariableCount div 2 - 1 do
      begin
        EventData.InputSettings.ActiveChannels[i] := true; // all active
      end;
    end else
    begin
      for i := 0 to ChartVariableCount - 1 do
      begin
        EventData.InputSettings.ActiveChannels[i] := true; // all active
      end;
    end;

    if(not reset) then   // restore old state up to current length of array
    begin
      oldCount := Min(oldCount,newCount);
      for i := 0 to oldCount-1 do  // restore channel states
      begin
        EventData.InputSettings.ActiveChannels[i] := oldState[i];
      end;
    end;
  end;
end;
{************************************************************************
* NAME:    SaveChartActiveChannels
* FUNCTIONAL DESCRIPTION:
* Save the active series in the chart
 INPUT PARAMETERS:    none
* OUTPUT PARAMETERS:   none
* RETURN:              none
************************************************************************}
procedure TTimeSeriesChartHRM.SaveChartActiveChannels();
var
  i : integer;
  status : boolean;
begin

  Assert(EventData <> nil,'EventData is nil');
  Assert(EventData.InputSettings <> nil,'InputSettings is nil');

  for i := 0 to ChartDisplay.SeriesCount-1 do
  begin
    status := ChartDisplay.Series[i].Active;
    EventData.InputSettings.ActiveChannels[i] := status;
  end;
end;
{************************************************************************
* NAME:     AutoScaleAxes
* FUNCTIONAL DESCRIPTION:  Reset chart axis limits to leave small interval around limits of data
* INPUT PARAMETERS: reset : boolean  true if scaling to fit all data in chart
*                                    false if scaling to fit a subset in chart
* OUTPUT PARAMETERS: none
* RETURN:           none
************************************************************************}
procedure TTimeSeriesChartHRM.AutoScaleAxes(reset : boolean);
var
  xMin, xMax, yMin, yMax, xDelta : double;
  vbase : Double;
  harmonicInput : TTimeSeriesInputSettingsHRM;
begin

{*  TeeChart autoscaling causes an exact fit to the data. Since this sometimes not
*  visually appealing, this scaling function subtracts 5% from the limits of
*  the horizontal and vertical axes so that chart is clearly visible in the chart
*  area.}

  inherited AutoScaleAxes(xMin,xMax,yMin,yMax);
  if(not reset) then
  begin
    harmonicInput :=  EventData.InputSettings as TTimeSeriesInputSettingsHRM;
    if(EventData.ChartSubType = cstHRM_INT) then
    begin
      if(harmonicInput.HarmonicChartDisplay <> hdtInterval) then
      begin
        if(EventData.SettingTable.Count > 0) then
        begin
          vbase := (EventData.SettingTable[0] as TTimeSeriesSettingRecord).SettingValue;
        end else
        begin
          vbase := 5.0;
        end;
        xMin := vbase - 5;
        xMax := xMin + (xMax-xMin)/50;
      end;
    end else
    begin
      if(harmonicInput.HarmonicChartDisplay <> hdtInterval) then
      begin
        xMax := xMin + (xMax-xMin)/2;
      end;
    end;
  end;

  xDelta := Max((xMax - xMin)/20.0,1.0);
  ChartDisplay.BottomAxis.SetMinMax(xMin - xDelta, xMax + xDelta);
  xDelta := (yMax - yMin)/20.0;
  ChartDisplay.LeftAxis.SetMinMax(yMin-xDelta,yMax + xDelta);

end;
{************************************************************************
* NAME: InitChartTitles
* FUNCTIONAL DESCRIPTION: Set chart titles (main, x and y axes)
* INPUT PARAMETERS: none
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************}
procedure TTimeSeriesChartHRM.InitChartTitles();
const
  TITLE_LIST : array[0..7] of string = ('Invalid','Undefined',
  'Harmonic Percentages Current and Voltage',
  'Harmonic Magnitude Current and Voltage',
  'Harmonic Mag/Angle Current',
  'Harmonic Mag/Angle Voltage',
  'Harmonic Power',
  ''
  );
  SUBTITLE_LIST : array[0..4] of string = ('Unknown',
  'Harmonic Spectral Analysis',
  'Harmonic Components',
  'Harmonic Grouping',
  'Inter-harmonic Grouping');
var
  harmonicSettings : TTimeSeriesInputSettingsHRM;
begin
  Assert(EventData <> nil,'EventData is nil');
  inherited InitChartTitles;   // allocates title array

  if(EventData.RawDataTable.Count > 1) then
  begin
    harmonicSettings := EventData.InputSettings as TTimeSeriesInputSettingsHRM;

    ChartDisplay.Title.Text[0] := EventData.MeterID + ' - ' +
                                  TITLE_LIST[integer(EventData.ChartSubType)+1];
    if(EventData.ChartSubType = cstHRM_INT) then
    begin
      ChartDisplay.Title.Text[0] := ChartDisplay.Title.Text[0] +
         SUBTITLE_LIST[integer(harmonicSettings.HarmonicChartType)+1];
    end;

    ChartDisplay.Title.AdjustFrame := true;
    ChartDisplay.LeftAxis.Title.Caption := 'Channel Value';

    ChartDisplay.LeftAxis.Title.Angle := 90;    // vertical orientation

    if(EventData.ChartSubType = cstHRM_INT) then
    begin
      ChartDisplay.BottomAxis.Title.Caption := 'Frequency (Hz)';
    end else
    begin
      ChartDisplay.BottomAxis.Title.Caption := 'Harmonic Number';
    end;
  end;
end;
{************************************************************************
* NAME:    AppendCSVHeader
* FUNCTIONAL DESCRIPTION:  Append header records for CSV file to string
* INPUT PARAMETERS: none
* OUTPUT PARAMETERS: MeterBuffer : string  file content
* RETURN: result : boolean   status of append (true = success)
************************************************************************}
function TTimeSeriesChartHRM.AppendCSVHeader(var MeterBuffer : string) : boolean;
begin
  result := inherited AppendCSVHeader(MeterBuffer);
end;
{************************************************************************
* NAME:    AppendCSVData
* FUNCTIONAL DESCRIPTION: Append CSV data from HMI data stored
*                         in memory resident TimeSeries structure
* INPUT PARAMETERS:       none
* OUTPUT PARAMETERS:      MeterBuffer : string  file content
* RETURN:                 result : boolean   true if data appended, otherwise false
************************************************************************}
function TTimeSeriesChartHRM.AppendCSVData(var MeterBuffer : string) : boolean;
var
  i,j,channelCount,recordCount,pos  : Integer;      // indices
  rValue : Double;
  myValue : string;
  dataRecord : TTimeSeriesRecord;
begin

  Assert(EventData <> nil,'EventData is nil');
  Assert(EventData.RawDataLabels <> nil,'RawDataLabels is nil');
  Assert(EventData.RawDataTable <> nil,'RawDataTable is nil');

  With EventData do
  begin
    channelCount := RawDataLabels.Count;
    recordCount := RawDataTable.Count;

    if((channelCount = 0) or (recordCount = 0)) then   // no data available
    begin
      result := false;
      MeterBuffer := MeterBuffer + '# No Harmonic data available.' + CRLF;
    end else
    begin

      MeterBuffer := MeterBuffer + ChartField + ', ';

      // export channel names
      for i := 0 to channelCount-1 do
      begin
          MeterBuffer := MeterBuffer + RawDataLabels[i];
          if(i < channelCount-1) then
          begin
             MeterBuffer := MeterBuffer + ', '
          end;
      end;

      MeterBuffer := MeterBuffer + CRLF;

      // export channel data
      for i := 0 to recordCount - 1 do
      begin
        dataRecord := (EventData.RawDataTable[i] as TTimeSeriesRecord);
        if(ChartSubType = cstHRM_INT) then
        begin
          MeterBuffer := MeterBuffer + Format('%7.2f',[dataRecord.RecordValue]);
        end else
        begin
          MeterBuffer := MeterBuffer + Format('%d',[dataRecord.RecordNumber]);
        end;

        for j := 0 to channelCount - 1 do
        begin
          rValue := dataRecord.Value[j] + 0.00005;  // account for 4 decimal digit precision error
          myValue := NonLocaleFloatToStr(rValue); // may or may not have decimal pt.
          if(AnsiPos('E',myValue) < 1) then  // make sure no exponent (scientific notation)
          begin
            pos := AnsiPos('.',myValue);  // look for decimal
            if(pos > 0) then
            begin
              myValue := AnsiLeftStr(myValue,pos+4);   // keep only 4 decimal places
            end;
          end else
          begin
            if(dataRecord.Value[j] < 0.0001) then  // if less that 4 decimal places, treat as zero
            begin
              myValue := '0.000';
            end;
          end;
          MeterBuffer := MeterBuffer + ', ' + myValue;
        end;

        MeterBuffer := MeterBuffer + CRLF;
      end;
      result := true;
    end;
  end;
end;
{************************************************************************
* NAME:    AppendReport
* FUNCTIONAL DESCRIPTION: Append Harmonic summary to string
* INPUT PARAMETERS:    none
* OUTPUT PARAMETERS:   MeterBuffer : string  file content
* RETURN:              result : boolean   true if summary appended, otherwise false
************************************************************************}
function TTimeSeriesChartHRM.AppendReport(var MeterBuffer: string) : boolean;
const
  HRM_INT_HEADER = 'Summary of Harmonic and RMS Inter-harmonic Frequencies' + CRLF;
  HRM_INT_FUND = 'Fundamental Frequency = ';
  HRM_NOT_ENOUGH_DATA = 'Not enough rows of data for harmonic report, possible data loss or corruption.';
  HRM_INVALID_DATA = 'Invalid harmonic data';
  TOLERANCE = 0.05;
var
  i,j,start,scount,recCount   : Integer;      // indices
  dataRecord : TTimeSeriesRecord;
  myDataFormat, myDate, myTime : string;
  myFormat : TFormatSettings;
  hbase,nomFreq,binSize,minBin,myValue1,myValue2, binCount : Double;
  sum : array of Double;
  valid : boolean;
begin

  EventData.InitDateTimeFormat(ftInternational,myFormat);
  DateFormat2DigitYear(myFormat);
  myDataFormat := ' %10.3f';
  start := 0;
  hbase := 0.0;
  recCount := 0;
  binSize := 5.0;  // arbitrary default
  binCount := 12;  // arbitrary default
  minBin := 20.0;  // arbitrary default
  valid := true;
  if(EventData.SettingTable.Count > 0) then
  begin
    nomFreq := (EventData.SettingTable[0] as TTimeSeriesSettingRecord).SettingValue;
  end else
  begin
    nomFreq := 0.0;
  end;

  With EventData do
  begin
    if(RawDataTable.Count > 1) then
    begin
      if(ChartSubType = cstHRM_INT) then  // only used for interharmonic calculations
      begin
        dataRecord := (RawDataTable[RawDataTable.Count-1] as TTimeSeriesRecord);
        binSize := dataRecord.RecordValue;   // floating representation of (frequency)
        dataRecord := (RawDataTable[0] as TTimeSeriesRecord);
        minBin := dataRecord.RecordValue;     // smallest frequence bin
        binSize := (binSize - minBin)/RawDataTable.Count;
        if(binsize < TOLERANCE) then    // This can only happen if meter reports data incorrectly
        begin
          binsize := 1.0;
          MeterBuffer := HRM_INVALID_DATA;
          valid := false;
        end;
        binCount := nomFreq / binSize;
        if(binCount < TOLERANCE) then    // This can only happen if meter reports data incorrectly
        begin
          binCount := 1.0;
          MeterBuffer := HRM_INVALID_DATA;
          valid := false;
        end;
      end;

      if(valid and (Length(DataColumnArray) > 0)) then
      begin
        SetLength(sum, Length(DataColumnArray));
        // Write report header information
        myDate := DateToStr(EventData.ChartDateTime,myFormat);  // use setting date format
        myTime := FormatDateTime('hh:mm:ss',EventData.ChartDateTime);
        MeterBuffer := Format('%-60s',[MeterID]) +
                      ' Date: ' + myDate +
                      ' Time: ' + myTime + CRLF;
        MeterBuffer := MeterBuffer + Format('%-60s',[MeterTID]) + ' Time Source: ' + TimeSource + CRLF2;
        if(ChartSubType = cstHRM_INT) then
        begin
          MeterBuffer := MeterBuffer + HRM_INT_FUND + Format(myDataFormat,[nomFreq]) + CRLF2;
          MeterBuffer := MeterBuffer + CRLF + Format('%16s',[' ']) + HRM_INT_HEADER + CRLF;
        end;
        MeterBuffer := MeterBuffer + '      Frequency';

        if(Length(DataColumnArray) > 0) then        // write channel names
        begin
          for i := 0 to Length(DataColumnArray) - 1 do
          begin
            MeterBuffer := MeterBuffer + Format('%11s',[Trim(RawDataLabels[i])]);
          end;
          MeterBuffer := MeterBuffer + CRLF;
        end else
        begin
          MeterBuffer := MeterBuffer + 'No Harmonic records found.' + CRLF;
        end;

        if(ChartSubType = cstHRM_INT) then
        begin
          scount:= 0;    // number of sub-intervals
          for i := start to RawDataTable.Count - 1 do
          begin
            dataRecord := (RawDataTable[i] as TTimeSeriesRecord);
            myValue1 := dataRecord.RecordValue / binCount;
            myValue2 := nomFreq / binCount;

            if((myValue1 > 0.0) and (Abs(Round(myValue1 / myValue2) - (myValue1 / myValue2)) < TOLERANCE)) then
            begin
              inc(recCount);
              if(recCount = 1) then     // first record is special case, harmonics start at 20
              begin
                MeterBuffer := MeterBuffer + Format('%7.2f-%7.2f',[minBin, hbase+nomFreq-binSize]);
              end else
              begin
                MeterBuffer := MeterBuffer + Format('%7.2f-%7.2f',[hbase+binSize, hbase+nomFreq-binSize]);
              end;
              // Display inter-harmonic RMS
              for j := 0 to Length(DataColumnArray) - 1 do  // write channel values
              begin
                MeterBuffer := MeterBuffer + Format(myDataFormat,[Sqrt(sum[j])]);
              end;
              MeterBuffer := MeterBuffer + CRLF;

              // Display fundamental harmonic frequency
              MeterBuffer := MeterBuffer + Format('%15.2f',[dataRecord.RecordValue]);
              for j := 0 to Length(DataColumnArray) - 1 do  // write channel values
              begin
                MeterBuffer := MeterBuffer + Format(myDataFormat,[dataRecord.Value[j]]);
              end;
              MeterBuffer := MeterBuffer + CRLF;
              for j:= 0 to Length(DataColumnArray)-1 do  // initialize sum
              begin
                sum[j] := 0.0;
              end;
              scount := 0;
              hbase := hbase + nomFreq;  // new lower limit of harm. freq. range
            end else   // RMS the inter-harmonics
            begin
              for j := 0 to Length(DataColumnArray) - 1 do
              begin
                sum[j] := sum[j] + dataRecord.Value[j] * dataRecord.Value[j];
              end;
              Inc(scount);
            end;
          end;
          if(scount > 0) then   // left-over sub-intervals
          begin
            MeterBuffer := MeterBuffer + Format('%7.2f-%7.2f',
                                          [hbase+binSize, hbase+nomFreq-binSize]);
            for j := 0 to Length(DataColumnArray) - 1 do  // write channel values
            begin
              MeterBuffer := MeterBuffer + Format(myDataFormat,[Sqrt(sum[j])]);
            end;
            MeterBuffer := MeterBuffer + CRLF;
          end;
          result := true;
        end else
        begin
          // Write table of Harmonic data

          for i := start to RawDataTable.Count - 1 do
          begin
            dataRecord := (RawDataTable[i] as TTimeSeriesRecord);
            MeterBuffer := MeterBuffer + Format('%-9.3f',[dataRecord.RecordValue]) + '  ';

            for j := 0 to Length(DataColumnArray) - 1 do  // write channel values
            begin
              MeterBuffer := MeterBuffer + Format(myDataFormat,[dataRecord.Value[j]]);
            end;
            MeterBuffer := MeterBuffer + CRLF;
          end;
          result := true;
        end;
      end else
      begin
        result := false;     // no data to write
      end;
    end else
    begin
      MeterBuffer := HRM_NOT_ENOUGH_DATA;
      result := false;
    end;
  end;
end;
{************************************************************************
* NAME:   FillEventIndexTable
* FUNCTIONAL DESCRIPTION:  Virtual base class method
* INPUT PARAMETERS:
*                    filter: boolean  use filter controls true/false
* OUTPUT PARAMETERS: none
* RETURN:  result : boolean  true if events found, otherwise false
************************************************************************}
function TTimeSeriesChartHRM.FillEventIndexTable(filter : boolean) : boolean;
begin
  // intentionally blank   (no event index for Harmonic chart)
  result := false;
end;

end.
