{*****************************************************************************
* UNIT DESCRIPTION:  Base class for SSI charting functions
* SPEC REF: (if used)
* DESIGN REF: (if used)
* 
* Copyright (c) 2009 Schweitzer Engineering Laboratories, Pullman, Washington
******************************************************************************}
unit ClassTimeSeriesChartSSI;

interface

uses
  Windows, SysUtils, StrUtils, Classes, DateUtils, Math,
  TeeTools, Chart, TeeComma, TeeEdit, Graphics, Series,
  TimeRoutines, ClassChartInput, ClassTimeSeries, ClassTimeSeriesChart,
  ClassChartInputSSI, TimeSeriesConstants;
const
  CRLF = Chr(13) + Chr(10);     // carriage return line feed
  PHASE_A = 1;
  PHASE_C = 3;
  INTERRUPT_3P = 'III';
  SAG_EVENT = 'U';
  SWELL_EVENT = 'O';
  INTERRUPT_EVENT = 'I';
  NOMINAL_EVENT = '.';
  PRE_DISTURBANCE = 'P';
  POST_DISTURBANCE = 'E';
  INVALID_RECORD = 'X';
  FAST_RECORD = 'F';
  READY_RECORD = 'R';
type
  TTimeSeriesChartSSI = class(TTimeSeriesChart)
  private
    { Private declarations }
  public
    Constructor Create(aOwner: TComponent); override;
    function GetChartFilename(aFileType : TTimeSeriesFileType; const ExportPath,
             aID : string) : string;  override;
    procedure InitChart(aFileType : TTimeSeriesFileType);  override;
    procedure InitChartVariableCount();  override;
    procedure InitChartTitles(); override;
    procedure InitChartActiveChannels();   override;
    procedure InitChartActiveChannelsDefault(reset : boolean); override;
    procedure SaveChartActiveChannels(); override;
    procedure InitChartActiveEventsDefault(); override;
    procedure AutoScaleAxes(reset : boolean); overload; override;
  end;

implementation


{************************************************************************
* NAME:   Create
* FUNCTIONAL DESCRIPTION: class constructor
* INPUT PARAMETERS:  none
* OUTPUT PARAMETERS: none
* RETURN:            none
************************************************************************}
constructor TTimeSeriesChartSSI.Create(aOwner: TComponent);
begin
  inherited Create(aowner);
end;
{************************************************************************
* NAME:   GetChartFilename
* FUNCTIONAL DESCRIPTION: Get filename where chart data is saved by default
* INPUT PARAMETERS:  aFileType : TTimeSeriesFileType  (type of file)
*                    ExportPath : default path to file
*                    aID        : ID of meter
* OUTPUT PARAMETERS: none
* RETURN:            filePath : string  path to file that was saved
************************************************************************}
function TTimeSeriesChartSSI.GetChartFilename(aFileType : TTimeSeriesFileType; const ExportPath, aID : string) : string;
var
  myPath : string;
begin
    myPath := '';
  result := myPath;
end;

{************************************************************************
* NAME:  InitChart
* FUNCTIONAL DESCRIPTION: Initialize chart class members before importing SSI report file
* INPUT PARAMETERS: aFileType : TTimeSeriesFileType  file type (ftReport...)
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************}
procedure TTimeSeriesChartSSI.InitChart(aFileType : TTimeSeriesFileType);
const
  standardSSI_REP : array[0..4] of integer = (-1,1,2,12,13);   // record, date, time, phase, status
  standardSSI_CSV : array[0..4] of integer = (-1,1,2,11,12);   // record, date, time, phase, status
  standardSSI_REP_101 : array[0..4] of integer = (-1,1,2,15,16);   // record, date, time, phase, status
  standardSSI_CSV_101 : array[0..4] of integer = (-1,1,2,14,15);   // record, date, time, phase, status
  standardSSI_CSV_4XX : array[0..4] of integer = (-1,1,2,6,7);   // record, date, time, phase, status
  SIZE_COLUMNS_4XX = 3;    // only 3 voltages (Va, Vb, Vc)
  SIZE_COLUMNS = 8;       // Currents and Voltages (Ia,Ib,Ic,Ig,In,Va,Vb,Vc)
  SIZE_COLUMNS_101 = 11;       // Currents and Voltages (Ia,Ib,Ic,Ig,In,VBaseA,Va,VBaseB,Vb,VBaseC,Vc)
  FIRST_COLUMN = 3;       // First current or voltage
var
  i,size : integer;
  new73x : boolean;
begin
  inherited InitChart(aFileType);
  EventData.ChartType := ctSSI;
  EventData.ChartName := 'VSSI';
  new73x := IsAdvanced73x(EventData.MeterFID);

  if(aFileType = ftReport) then
  begin
    if(new73x) then
    begin
      EventData.SetLengthHeaderColumns(Length(standardSSI_REP_101));
      EventData.SetLengthDataColumns(SIZE_COLUMNS_101);
      // Initialize standard columns
      for i := Low(standardSSI_REP_101) to High(standardSSI_REP_101) do
      begin
        EventData.HeaderColumns[i] := standardSSI_REP[i];
      end;
      size := SIZE_COLUMNS_101;
    end else
    begin
      EventData.SetLengthHeaderColumns(Length(standardSSI_REP));
      EventData.SetLengthDataColumns(SIZE_COLUMNS);
      // Initialize standard columns
      for i := Low(standardSSI_REP) to High(standardSSI_REP) do
      begin
        EventData.HeaderColumns[i] := standardSSI_REP[i];
      end;
      size := SIZE_COLUMNS;
    end;
  end else
  begin
    if(EventData.DeviceType = dtSEL4XX) then
    begin
      EventData.SetLengthHeaderColumns(Length(standardSSI_CSV_4XX));
      EventData.SetLengthDataColumns(SIZE_COLUMNS_4XX);
      // Initialize standard columns
      for i := Low(standardSSI_CSV) to High(standardSSI_CSV_4XX) do
      begin
        EventData.HeaderColumns[i] := standardSSI_CSV_4XX[i];
      end;
      size := SIZE_COLUMNS_4XX;
    end else
    begin
      if(new73x) then
      begin
        EventData.SetLengthHeaderColumns(Length(standardSSI_CSV_101));
        EventData.SetLengthDataColumns(SIZE_COLUMNS_101);
        // Initialize standard columns
        for i := Low(standardSSI_CSV_101) to High(standardSSI_CSV_101) do
        begin
          EventData.HeaderColumns[i] := standardSSI_CSV_101[i];
        end;
        size := SIZE_COLUMNS_101;
      end else
      begin
        EventData.SetLengthHeaderColumns(Length(standardSSI_CSV));
        EventData.SetLengthDataColumns(SIZE_COLUMNS);
        // Initialize standard columns
        for i := Low(standardSSI_CSV) to High(standardSSI_CSV) do
        begin
          EventData.HeaderColumns[i] := standardSSI_CSV[i];
        end;
        size := SIZE_COLUMNS;
      end;
    end;
  end;
  // Initialize data columns
  for i := 0 to size-1 do
  begin
    EventData.DataColumns[i] := FIRST_COLUMN + i;
  end;

end;

{************************************************************************
* NAME:    InitChartVariableCount
* FUNCTIONAL DESCRIPTION:
* Set the chart variable count based on chart type and event count
* INPUT PARAMETERS:    none
* OUTPUT PARAMETERS:   none
* RETURN:              none
************************************************************************}
procedure TTimeSeriesChartSSI.InitChartVariableCount();
begin

  Assert(EventData <> nil,'EventData is nil');

  inherited InitChartVariableCount;  // initializes ChartHasEvents

  // update chart variable count to include CBEMA curves if appropriate

  if(EventData.IndexTable.Count > 0) then     // SSI CBEMA curves are active
  begin
    ChartHasEvents := true;
  end;
end;
{************************************************************************
* NAME:    InitChartActiveChannels
* FUNCTIONAL DESCRIPTION:
* Set the active series in the chart
* INPUT PARAMETERS:    none
* OUTPUT PARAMETERS:   none
* RETURN:              none
************************************************************************}
procedure TTimeSeriesChartSSI.InitChartActiveChannels();
var
  i,sc,ac : integer;
  status : boolean;
begin
  Assert(EventData <> nil,'EventData is nil');

  ac := Length(EventData.InputSettings.ActiveChannelArray);
  sc := ChartDisplay.SeriesCount;

  Assert(sc <= ac, 'Array size mismatch');

  // update chart active channels

  if((EventData.ChartType = ctSSI) and
     TTimeSeriesInputSettingsSSI(EventData.InputSettings).ChartCBEMA) then
  begin
    exit;
  end;

  for i := 0 to ChartDisplay.SeriesCount-1 do
  begin
    status :=  EventData.InputSettings.ActiveChannels[i];
    ChartDisplay.Series[i].Active := status;
  end;

end;
{************************************************************************
* NAME:    InitChartActiveChannelsDefault
* FUNCTIONAL DESCRIPTION:
** Set the default state of active channels
 INPUT PARAMETERS:    reset : boolean  reset channel status to default if true
* OUTPUT PARAMETERS:   none
* RETURN:              none
************************************************************************}
procedure TTimeSeriesChartSSI.InitChartActiveChannelsDefault(reset : boolean);
const
  VOLTAGE_COUNT = 3;
var
  i, currentIndex: Integer;
  oldCount, newCount : integer;
  oldState : array of boolean;
begin

  Assert(EventData <> nil,'EventData is nil');

  oldCount := Length(EventData.InputSettings.ActiveChannelArray);
  newCount := ChartVariableCount;

  if(reset or (newCount <> oldCount)) then   // length has changed
  begin
    if(not reset) then     // save active channel state before re-allocating
    begin
      SetLength(oldState,oldCount);
      for i := 0 to oldCount - 1 do   // remember previous active channel state
      begin
        oldState[i] := EventData.InputSettings.ActiveChannels[i];
      end;
    end;

    EventData.InputSettings.SetActiveChannelLength(newCount);
    EventData.InputSettings.ActiveChannels[newCount-1] := false; // disable last channel (event)

    currentIndex := EventData.RawDataLabels.Count - VOLTAGE_COUNT;

    for i := currentIndex to EventData.RawDataLabels.Count-1 do  // enable voltage channels
    begin
      EventData.InputSettings.ActiveChannels[i] := true;   // voltages
    end;

    for i := 0 to currentIndex-1 do     // disable the current channels
    begin
      EventData.InputSettings.ActiveChannels[i] := false;  // currents
    end;

//    if(not reset) then     // restore old state up to current length of array
//    begin
//      oldCount := Min(oldCount,newCount);
//      for i := 0 to oldCount-1 do  // restore channel states
//      begin
//        EventData.InputSettings.ActiveChannels[i] := oldState[i];
//      end;
//    end;
  end;

end;
{************************************************************************
* NAME: InitChartTitles
* FUNCTIONAL DESCRIPTION: Set chart titles (main, x and y axes)
* INPUT PARAMETERS: none
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************}
procedure TTimeSeriesChartSSI.InitChartTitles();
begin

  inherited InitChartTitles;

end;

{************************************************************************
* NAME:    SaveChartActiveChannels
* FUNCTIONAL DESCRIPTION:
* Save the active series in the chart
 INPUT PARAMETERS:    none
* OUTPUT PARAMETERS:   none
* RETURN:              none
************************************************************************}
procedure TTimeSeriesChartSSI.SaveChartActiveChannels();
var
  i : integer;
  status : boolean;
begin
  Assert(EventData <> nil,'EventData is nil');
  Assert(EventData.InputSettings <> nil,'InputSettings is nil');

  if((EventData.ChartType = ctSSI) and
    TTimeSeriesInputSettingsSSI(EventData.InputSettings).ChartCBEMA) then
  begin
    exit;
  end;

  for i := 0 to ChartDisplay.SeriesCount-1 do
  begin
    status := ChartDisplay.Series[i].Active;
    EventData.InputSettings.ActiveChannels[i] := status;
  end;
end;
{************************************************************************
* NAME:    InitChartActiveEventsDefault
* FUNCTIONAL DESCRIPTION:
* Set the default state of active events
 INPUT PARAMETERS:    none
* OUTPUT PARAMETERS:   none
* RETURN:              none
************************************************************************}
procedure TTimeSeriesChartSSI.InitChartActiveEventsDefault();
begin

end;
{************************************************************************
* NAME:     AutoScaleAxes
* FUNCTIONAL DESCRIPTION:  Reset chart axis limits to leave small interval around limits of data
* INPUT PARAMETERS: reset : boolean  true if scaling to fit all data in chart
*                                    false if scaling to fit a subset in chart
* OUTPUT PARAMETERS: none
* RETURN:           none
************************************************************************}
procedure TTimeSeriesChartSSI.AutoScaleAxes(reset : boolean);
var
  xMin, xMax, yMin, yMax : double;
begin

// NOTE: reset not presently used in this method.  AutoScaleAxes is a polymorphic
// method with different behavior depending on the sub-class. The other
// classes may use the reset argument in their version of this method.

{*  TeeChart autoscaling causes an exact fit to the data. Since this sometimes not
*  visually appealing, this scaling function subtracts 5% from the limits of
*  the horizontal and vertical axes so that chart is clearly visible in the chart
*  area.}

  inherited AutoScaleAxes(xMin,xMax,yMin,yMax);

  {CBEMA chart requires use of a smaller range in the Y direction because the limit
   curves extend far beyond most data in the positive direction (0 to 300).  This
   has the effect of zooming in a small amount in order to see the sag and swell
   regions more clearly.
  }
//  if(TTimeSeriesInputSettingsSSI(EventData.InputSettings).ChartCBEMA) then
//  begin
//    ChartDisplay.BottomAxis.SetMinMax(xMin - (xMax - xMin)/20.0, xMax + (xMax - xMin)/100.0);
//    ChartDisplay.LeftAxis.SetMinMax(yMin - (yMax - yMin)/20,yMax - (yMax - yMin)/4.0);
//  end else
  begin
    ChartDisplay.BottomAxis.SetMinMax(xMin - (xMax - xMin)/20.0, xMax + (xMax - xMin)/20.0);
    ChartDisplay.LeftAxis.SetMinMax(yMin - (yMax - yMin)/20.0,yMax + (yMax - yMin)/20.0);
  end;
end;
end.
