{*****************************************************************************
* UNIT DESCRIPTION:  Date/Time conversion functions used in TOU and charting
* SPEC REF: (if used)
* DESIGN REF: (if used)
*
* Copyright (c) 2009 Schweitzer Engineering Laboratories, Pullman, Washington
******************************************************************************}
unit TimeRoutines;

interface

uses
  Classes,
  SysUtils,
  StrUtils,
  TimeConstants,
  TimeSeriesConstants,
  Controls,
  Windows;

type
  TDateFormatType = (ftEnglish_US=0, ftInternational);

//  function  ShortNameMonth(Month: word): string;

  function  DaysBetween(StartDate, EndDate: TDateTime): integer; overload;

  function  DateToJulian(Year, Month, Day: integer): integer; overload;
  function  DateToJulian(JulianDay: TDateTime): integer; overload;

  function  JulianToDate(Julian, Year: integer): TDateTime; overload;

  function  MinutesToString(Minutes: integer): string;
  function  TenthsMillisecondsToTime(tenthsMS: longword): TDateTime;
  procedure InitTFormatSettings(formatType: TDateFormatType; dateFormat : string; var myFormat : TFormatSettings);
  procedure DateFormat2DigitYear(var myFormat : TFormatSettings);
  procedure DurationToTimeString(var aTimeString : string; duration : TDateTime);
  function  GetDateControlFormat(dateFormat : string) : string;
  function StringInArray(Value: string; Strings : array of String): integer;
function GetHMIVersion(meterFID: string): integer;
function GetHMIDevice(meterFID: string): string;
function IsDevice73x(meterFID: string): boolean;
function IsAdvanced73x(meterFID: string): boolean;

implementation

{!******************************************************************************
* MODULE NAME ...  MinutesToString
* SYSTEM NAME ...  Acselerator
* AUTHOR ........  Bruce Scallorn
* 
* FUNCTIONAL DESCRIPTION:
* Converts a time in minutes since midnight to a 24 hour equivalent.
* 
* INPUT PARAMETERS:
*                 Minutes: integer                The minutes to convert
* 
* OUTPUT PARAMETERS:
*                    None:                        
* 
* RETURN VALUE:
*                  result: string                 the converted time
* 
* ERRORS:
* 
* APPLICATION NOTES & EXAMPLE:
* 
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2003.12.08  Bruce Scallorn    Added Comment Block
*******************************************************************************}
function MinutesToString(Minutes: integer): string;
var
  Hours : integer;
  Min   : integer;
begin
  Hours   := Minutes div MINUTES_IN_HOUR;
  Min     := Minutes mod MINUTES_IN_HOUR;

  result  := Format('%.2d:%.2d', [Hours, Min]);
end;

{!******************************************************************************
* MODULE NAME ...  JulianToDate
* SYSTEM NAME ...  AcSELerator
* AUTHOR ........  Bruce Scallorn
* 
* FUNCTIONAL DESCRIPTION:
* Converts a Julian date to the corresponding TDateTime.
* 
* INPUT PARAMETERS:
*                  Julian: integer                
*                    Year: integer                
* 
* OUTPUT PARAMETERS:
*                    None:                        
* 
* RETURN VALUE:
*                  result: TDateTime              
* 
* ERRORS:
* 
* APPLICATION NOTES & EXAMPLE:
* Unit Test Notes:                                Unit Tested:
* 
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2002.11.08  Bruce Scallorn    Original code
* 2006.01.31  Bruce Scallorn    modified calculation method
* 2008.07.03  Kent Myers        correct date calculation
*******************************************************************************}
function  JulianToDate(Julian, Year: integer): TDateTime;
var
  DayOne  : TDateTime;
begin
  DayOne  := EncodeDate(Year, 1, 1);
  result  := DayOne + Julian - 1;  // subtract one since 1/1 already included
end;


{!******************************************************************************
* MODULE NAME ...  DateToJulian
* SYSTEM NAME ...  AcSELerator
* AUTHOR ........  Bruce Scallorn
* 
* FUNCTIONAL DESCRIPTION:
* Given the Year, Month and Day the Julian Day is returned.
* 
* INPUT PARAMETERS:
*                     Day: integer                
*                   Month: integer                
*                    Year: integer                
* 
* OUTPUT PARAMETERS:
*                    None:                        
* 
* RETURN VALUE:
*                  result: integer                
* 
* ERRORS:
* 
* APPLICATION NOTES & EXAMPLE:
* Unit Test Notes:                                Unit Tested:
* 
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2002.11.4  Bruce Scallorn    Original code
*******************************************************************************}
function DateToJulian(Year, Month, Day: integer): integer;
var
  DayOne      : TDateTime;
  JulianDay   : TDateTIme;
begin
  DayOne      := EncodeDate(Year, 1, 1);
  JulianDay   := EncodeDate(Year, Month, Day);
  result      := Trunc(JulianDay - DayOne) + 1;
end;


{!******************************************************************************
* MODULE NAME ...  DaysBetween
* SYSTEM NAME ...  AcSELerator
* AUTHOR ........  Bruce Scallorn
*
* FUNCTIONAL DESCRIPTION:
* Returns the number of days between the starting date and the ending date
*
* INPUT PARAMETERS:
*                 EndDate: TDateTime
*               StartDate: TDateTime
*
* OUTPUT PARAMETERS:
*                    None:
*
* RETURN VALUE:
*                  result: integer
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
* Unit Test Notes:                                Unit Tested:
* 
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2002.10.31  Bruce Scallorn    Original code
* 2006.01.31  Bruce Scallorn    modified calculation method
*******************************************************************************}
function DaysBetween(StartDate, EndDate: TDateTime): integer;
begin
  result  := abs(Trunc(EndDate - StartDate))
end;

{!******************************************************************************
* MODULE NAME ...  ShortNameMonth
* SYSTEM NAME ...  AcSELerator
* AUTHOR ........  Bruce Scallorn
* 
* FUNCTIONAL DESCRIPTION:
* Returns the short name of the month.
* 
* INPUT PARAMETERS:
*                   Month: word                   
*
* OUTPUT PARAMETERS:
*                    None:
*
* RETURN VALUE:
*                  result: string
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
* Unit Test Notes:                                Unit Tested:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2002.11.19  Bruce Scallorn    Original code
*******************************************************************************}
//function ShortNameMonth(Month: word): string;
//begin
//
//  result := '';
//
//  case Month of
//  1:  result := 'Jan';
//  2:  result := 'Feb';
//  3:  result := 'Mar';
//  4:  result := 'Apr';
//  5:  result := 'May';
//  6:  result := 'Jun';
//  7:  result := 'Jul';
//  8:  result := 'Aug';
//  9:  result := 'Sep';
//  10: result := 'Oct';
//  11: result := 'Nov';
//  12: result := 'Dec';
//  end;
//
//
//end;
//

{!******************************************************************************
* MODULE NAME ...  DateToJulian
* SYSTEM NAME ...  AcSELerator
* AUTHOR ........  Bruce Scallorn
*
* FUNCTIONAL DESCRIPTION:
* Converts the Date at the TDateTime to the Julian day for that date.
* 
* INPUT PARAMETERS:
*               JulianDay: TDateTime              
* 
* OUTPUT PARAMETERS:
*                    None:                        
* 
* RETURN VALUE:
*                  result: integer                
* 
* ERRORS:
* 
* APPLICATION NOTES & EXAMPLE:
* Unit Test Notes:                                Unit Tested:
* 
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2002.11.20  Bruce Scallorn    Original code
* 2006.01.31  Bruce Scallorn    modified calculation method
*******************************************************************************}
function DateToJulian(JulianDay: TDateTime): integer;
var
  Year    : word;
  Month   : word;
  Day     : word;
  DayOne  : TDateTime;
begin
  DecodeDate(JulianDay, Year, Month, Day);

  DayOne  := EncodeDate(Year, 1, 1);
  result  := Trunc(JulianDay - DayOne) + 1;
end;

{**********************************************************************
* MODULE NAME ...   IsNumber
* SYSTEM NAME ...   AcSELerator
* AUTHOR ........
*
* FUNCTIONAL DESCRIPTION:  Check if the input string is a numeric string
*
* INPUT PARAMETERS:   OneStr --- A string to be checked
*
* OUTPUT PARAMETERS:  none
*
* RETURN VALUE:       Boolean
*
* ERRORS:
*
* APPLICATION EXAMPLE & NOTES:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/    REVISION
* --DATE--    IMPLEMENTOR    DESCRIPTION
* 03/04/99    Yuping L.      Original code
* 990609      Jian Chen      Modify the code to return false for non numbers
* 09/10/2001  essgee         Completely rewritten to use Val call.
***********************************************************************}
{$HINTS OFF} //This is done to remove annoying compiler warnings that state that.}
function  IsNumber (OneStr: string): Boolean;
var
  Code: integer;
  TempFloat : double;
begin
  result := False;
  try
    Val(OneStr, TempFloat, Code);
    if Code = 0 then
    begin
      result := True;
    end;
  except
  end;
end;

{************************************************************************
* NAME:    TenthsMillisecondsToTime
* FUNCTIONAL DESCRIPTION:  Convert tenths of milliseconds to TDateTime
* INPUT PARAMETERS: tenthsMS :
* OUTPUT PARAMETERS: none
* RETURN: result : TDateTime  : result of conversion
************************************************************************}
function  TenthsMillisecondsToTime (tenthsMS: longword): TDateTime;
begin
  result := tenthsMS / MSecsPerSec / SecsPerDay / 10;
end;
{************************************************************************
* NAME:       InitTFormatSettings
* FUNCTIONAL DESCRIPTION:    Initialize a TFormatSettings record
* INPUT PARAMETERS: formatType : TDateFormatType  (ftEnglish_US, ftInternational)
*                   dateFormat : string  short date format
*                  myFormat : TFormatSettings
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************}
procedure InitTFormatSettings(formatType: TDateFormatType; dateFormat : string; var myFormat : TFormatSettings);
var
  myDateTime : string;
begin
  myFormat.ShortDateFormat := dateFormat;
  myFormat.ShortTimeFormat := 'hh:mm:ss.zzz';
  myFormat.TwoDigitYearCenturyWindow := 0;

  if(formatType = ftEnglish_US) then   // force US English date format
  begin
    myFormat.DateSeparator := '/';
  end else
  begin
    myDateTime := FormatDateTime('MM/DD/YYYY',Now);
    if(AnsiContainsStr(myDateTime,'-')) then
    begin
      myFormat.DateSeparator := '-';
    end else
    begin
      myFormat.DateSeparator := '/';
    end;
  end;
  myFormat.TimeSeparator := ':';
  myFormat.DecimalSeparator := '.';
  myFormat.ThousandSeparator := '.';
end;
{************************************************************************
* NAME:       DateFormat2DigitYear
* FUNCTIONAL DESCRIPTION:    Convert date format to 2-digit year
* INPUT PARAMETERS:      myFormat : TFormatSettings
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************}
procedure DateFormat2DigitYear(var myFormat : TFormatSettings);
begin
  if(myFormat.ShortDateFormat[1] = 'm') then
  begin
    myFormat.ShortDateFormat := 'mm/dd/yy';
  end else
  if(myFormat.ShortDateFormat[1] = 'd') then
  begin
    myFormat.ShortDateFormat := 'dd/mm/yy';
  end else
  begin
    myFormat.ShortDateFormat := 'yy/mm/dd';
  end;
end;
{************************************************************************
* NAME:       DurationToTimeString
* FUNCTIONAL DESCRIPTION:    Convert date/time duration to time string (hhh:mm:ss.zzz)
* INPUT PARAMETERS:      aTimeString : string   time string in hhh:mm:ss.zzz
*                        duration : TDateTime  duration
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************}
procedure DurationToTimeString(var aTimeString : string; duration : TDateTime);
var
  myDateTime : string;
  myString : string;
  year,month,day,hour : integer;
  baseJulian, thisJulian : integer;
  baseDate : TDateTime;
begin

  myDateTime := FormatDateTime('MM/DD/YYYY hh:mm:ss.zzz',duration);
  myString := AnsiMidStr(myDateTime,1,2);
  month := StrToInt(myString);
  myString := AnsiMidStr(myDateTime,4,2);
  day := StrToInt(myString);
  myString := AnsiMidStr(myDateTime,7,4);
  year := StrToInt(myString);
  myString := AnsiMidStr(mydateTime,12,2);
  hour := StrToInt(myString);
  myString := AnsiMidStr(mydateTime,15,9);

  if(year < 1900) then
  begin
    if(day > 30) then
    begin
      thisJulian := 1;
    end else
    begin
      thisJulian := 0;
    end;
  end else
  begin
    thisJulian := DateToJulian(year,month,day) + 1;
  end;

  aTimeString := Format('%.3d:',[thisJulian*24 + hour]) + myString;
end;
{************************************************************************
* NAME:       GetDateControlFormat
* FUNCTIONAL DESCRIPTION:  convert English format string to international format
* INPUT PARAMETERS:      dateFormat : string  English format string
* OUTPUT PARAMETERS: none
* RETURN: result : string  updated format string for date time picker control
************************************************************************}
function GetDateControlFormat(dateFormat : string) : string;
var
  myFormat : string;
  myDate : string;
begin
  // First test for abreviated date format and convert to standard format
  if(dateFormat = 'MDY') then
  begin
    dateFormat := 'MM/dd/yyyy';
  end else
  if(dateFormat = 'DMY') then
  begin
    dateFormat := 'dd/MM/yyyy';
  end else
  if(dateFormat = 'YMD') then
  begin
    dateFormat := 'yyyy/MM/dd';
  end;

  // change month to upper case for date time picker convention
  myFormat := AnsiReplaceStr(dateFormat,'m','M');

  // test default date to determine standard date separator
  myDate := FormatDateTime('MM/DD/YYYY',Now);
  if(AnsiContainsStr(myDate,'-')) then    // international date
  begin
    myFormat := AnsiReplaceStr(myFormat,'/','-');
  end;
  result := myFormat;
end;
{************************************************************************
* NAME:     StringInArray
* FUNCTIONAL DESCRIPTION:   find matching string in array of strings
* INPUT PARAMETERS: Value : string  to find
*                   Strings: array of strings to search
* OUTPUT PARAMETERS: none
* RETURN: result : integer position of match in list (-1 if not found)
************************************************************************}
function StringInArray(Value: string; Strings : array of String): integer;
var
  i : integer;
begin
  for i := Low(Strings) to High(Strings) do
  begin
    if(AnsiCompareText(Value,Strings[i]) = 0) then
    begin
      result := i;
      exit;
    end;
  end;
  result := -1;  // not found;
end;
{************************************************************************
* NAME:     GetHMIVersion
* FUNCTIONAL DESCRIPTION:   get HMI Z-number from FID
* INPUT PARAMETERS: meterFID : string  meter FID string
* OUTPUT PARAMETERS: none
* RETURN: result : integer HMI z-number
************************************************************************}
function GetHMIVersion(meterFID: string): integer;
const
  SKIP_SETTINGS_ZNUMBER = 5;
var
  myString : string;
  HMIVersion : integer;
begin
  myString := AnsiMidStr(meterFID,Pos('-Z',meterFID)+SKIP_SETTINGS_ZNUMBER,3);
  if(TryStrToInt(myString,HMIVersion)) then
  begin
    result := HMIVersion;
  end else
  begin
    result := 0;
  end;
end;
{************************************************************************
* NAME:     GetHMIDevice
* FUNCTIONAL DESCRIPTION:   get device ID string from FID
* INPUT PARAMETERS: meterFID : string  meter FID string
* OUTPUT PARAMETERS: none
* RETURN: result : string device ID
************************************************************************}
function GetHMIDevice(meterFID: string): string;
var
  myString : string;
  HMIVersion : integer;
begin
  if(Pos('-',meterFID) > 0) then
  begin
    result := AnsiMidStr(meterFID,Pos('-',meterFID)+1,3);
  end else
  begin
    result := 'XXX';  // default ID when not found
  end;
end;

{************************************************************************
* NAME:     IsAdvanced73x
* FUNCTIONAL DESCRIPTION:   return true if device is an advanced 73x
* INPUT PARAMETERS: meterFID : string  meter FID string
* OUTPUT PARAMETERS: none
* RETURN: result : boolean true if advanced 73x, otherwise false
************************************************************************}
function IsAdvanced73x(meterFID: string): boolean;
var
  deviceID : string;
begin
  deviceID := GetHMIDevice(meterFID);
  result := ((Pos('735',deviceID) > 0) or
             ((Pos('734',deviceID) > 0) and (GetHMIVersion(MeterFID) >= 101)));
end;

{************************************************************************
* NAME:     IsDevice73x
* FUNCTIONAL DESCRIPTION:   return true if device is a 73x
* INPUT PARAMETERS: meterFID : string  meter FID string
* OUTPUT PARAMETERS: none
* RETURN: result : boolean true if advanced 73x, otherwise false
************************************************************************}
function IsDevice73x(meterFID: string): boolean;
var
  deviceID : string;
begin
  deviceID := GetHMIDevice(meterFID);
  result := ((Pos('735',deviceID) > 0) or (Pos('734',deviceID) > 0));
end;
{$HINTS ON}
end.
